-- Affine analysis

module AA where

import FiniteMap
import IOExts

-- Affine forms.  Constant term plus linear combination of noise symbols
-- represented by integers.
data AForm a = AForm a (LPart a)

-- Linear part
type LPart a = FiniteMap Int a

addLP :: Num a => LPart a -> LPart a -> LPart a
addLP alin blin = optLPart (plusFM_C (+) alin blin)

-- Remove zeros
optLPart :: Num a => LPart a -> LPart a
optLPart = filterFM (const (/= 0))

instance Show a => Show (AForm a) where
  show (AForm a0 lin) =
    -- "AForm " ++ show a0 ++ " " ++ show (fmToList lin)
    show a0 ++ concat [ " + " ++ show ai ++ "*eps" ++ show i
                      | (i,ai) <- fmToList lin ]

magLPart :: Num a => LPart a -> a
magLPart lin = sum [abs a | (_,a) <- fmToList lin]

-- Or, equivalently,
--   magLPart = sum . map (abs.snd) . fmToList

addNoise :: Num a => LPart a -> a -> LPart a
addNoise lin a = if a == 0 then lin else
                  addToFM lin (newNoiseSymbol()) a

-- Noise symbol generation, via the usual gensym hack.  Use only in
-- addNoise.
newNoiseSymbol :: () -> Int
newNoiseSymbol = unsafePerformIO $ do
  counter <- newIORef 1
  return $ \ () -> unsafePerformIO $ do
                     n <- readIORef counter
                     writeIORef counter (n+1)
                     return n


---- Conversion between intervals and affine forms

-- Interval described by center and half-width
type MidInterval a = (a,a)

toAF :: Fractional a => MidInterval a -> AForm a
toAF (mid,halfWidth) =
  AForm mid (addNoise emptyFM halfWidth)

fromAF :: Num a => AForm a -> MidInterval a
fromAF (AForm a0 lin) = (a0,magLPart lin)

{-
-- Or traditional intervals.  These guys make for more complicated
-- formulas, and currently (2002-07-12), expression simplification doesn't
-- simplify them well enough.
-- TODO: Revisit.
type Interval a = (a,a)

toAF :: Fractional a => Interval a -> AForm a
toAF (lo,hi) =
  AForm ((hi+lo)/2) (addNoise emptyFM ((hi-lo)/2))

fromAF :: Num a => AForm a -> Interval a
fromAF (AForm a0 lin) = (a0-mag,a0+mag) where mag = magLPart lin
-}

---- Math operations

noOvAForm meth = error $ "No AForm overloading for " ++ meth

-- Eq & Show are prerequisites for Num, so they need to be faked here
instance Eq (AForm a) where
  (==) = noOvAForm "(==)"
  (/=) = noOvAForm "(/=)"

instance Ord a => Ord (AForm a) where
  min = noOvAForm "min"
  max = noOvAForm "max"

instance Num a => Num (AForm a) where
  AForm a0 alin + AForm b0 blin =
    AForm (a0+b0) (addLP alin blin)
  AForm a0 alin * AForm b0 blin =
    AForm (a0*b0) (addNoise (addLP (mult b0 alin) (mult a0 blin))
                            (magLPart alin * magLPart blin))
   where
      mult x0 = mapFM_elt (* x0)

  negate (AForm a0 alin) = AForm (negate a0) (mapFM_elt negate alin)
  fromInteger a0 = AForm (fromInteger a0) emptyFM

instance Fractional a => Fractional (AForm a) where
  recip           = noOvAForm "recip"  -- TODO!
  fromRational a0 = AForm (fromRational a0) emptyFM

{-
instance Floating a => Floating (AForm a) where
  sqrt = 
  exp  = 
  log  = 
-}

{-
-- Univariate Chebyshev affine approximations.  See "Self-Validated
-- Numerical Methods and Applications" (1997) by Jorge Stolfi & Luiz
-- Henrique de Figueiredo, section 3.8.2.

-- Given functions f and dfInv such that deriv f . dfInv = id, and an
-- interval (lo,hi), yield (alpha,zeta) representing the minimax affine
-- approximation \ x -> alpha*x+zeta to f over (lo,hi).
uniCheby :: Num a => (a -> a) -> (a -> a) -> (a,a) -> (a,a)
uniCheby f dfInv (a,b) = (alpha,zeta)
 where
   alpha = (f b - f a) / (b - a)
   zeta  = (f u + r u) / 2 - alpha * u
   r 
-}

---- A few finite map utilities

-- Simpler map, where function ignores key
mapFM_elt :: (a -> b) -> FiniteMap key a -> FiniteMap key b
mapFM_elt = mapFM . const


---- Testing

a1 = toAF (0,2*pi)
a2 = toAF (5,1)

a3 = a1 + a2
a4 = a1 * a2
a5 = a3*a4

a6 = a2-a2
a7 = a2 * (10-a2)

a8 = 7 * a2
a9 = a8 - a2

-- Example to do: x^2+sin x.  Compare with "applied interval analysis"
-- book.
