-- Vertex shader assembly code 

module Asm where

import List (intersperse)
import Types
import DExp (Op(..),opName)

-- All instructions are op applications (even Mov) and contain a comment.
data Instr = PrimOp Op Dest [Source] String deriving Show

type Asm = [Instr]

-- Register classes
data RegClass = RegIn | RegConst | RegTemp | RegAddr | RegOut
  deriving (Eq,Ord,Show)

-- A register has a register class and index and a friendly name
data Reg = Reg RegClass Int String  deriving (Show,Eq)

-- Source registers may be swizzled and negated.  Not output reg.
data Source = Source NegSwiz Reg  deriving Show

-- Destination registers have layouts, which also serve as masks.
-- Not input reg.
data Dest = Dest Reg Layout  deriving Show


type SourceEnv = Env Source


---- Misc

-- Extract a single-float dest
destPart :: Index -> Dest -> Dest

destPart i dest@(Dest (Reg cls ix name) layout) =
  Dest (Reg cls ix partName) [layout !! i]
 where
   partName = takeWhile (/= '.') name ++ showMask [i]

swizzleSrc :: Swizzle -> Source -> Source
swizzleSrc swiz (Source ns reg) = Source (swizzleNS swiz ns) reg

---- Pretty printing

showRegClass :: RegClass -> Char
showRegClass RegIn    = 'v'
showRegClass RegConst = 'c'
showRegClass RegTemp  = 'r'
showRegClass RegAddr  = 'a'
showRegClass RegOut   = 'o'

showReg :: Reg -> String

showReg (Reg RegOut n _) = showRegOut n  -- output looks special
showReg (Reg regc   n _) = showRegClass regc : show n

showRegOut :: Int -> String
showRegOut 0 = "oD0"
showRegOut 1 = "oD1"
showRegOut 2 = "oFog"
showRegOut 3 = "oPos"
showRegOut 4 = "oPts"
showRegOut n = "oTex" ++ show (n-5)

showSource :: Source -> String
showSource (Source (NegSwiz neg swiz) reg) =
  showNegate neg ++ showReg reg ++ showSwizzle True swiz
 where
   showNegate True  = "-"
   showNegate False = ""

showDest :: Dest -> String
showDest (Dest reg layout) = showReg reg ++ showMask layout

showInstr :: Instr -> String
showInstr (PrimOp op dest srcs comment) = 
  padRightTo commentColumn ' '
            ( opName op ++ " " ++ showDest dest
              ++ concatMap showSourceC srcs )
  ++ "; " ++ comment

commentColumn = 36 :: Int

padLeftTo, padRightTo :: Int -> a -> [a] -> [a]

padLeftTo  n x xs = replicate (n - length xs) x ++ xs
padRightTo n x xs = xs ++ replicate (n - length xs) x


-- prepend comma
showSourceC src = ", " ++ showSource src

showAsm :: Asm -> String
showAsm = unlines . map showInstr
          -- concat . intersperse "\n" . map showAsm  -- omits final newline


-- Is dest:=src a no-op?
nopMov :: Dest -> Source -> Bool
nopMov (Dest destR layout) (Source (NegSwiz neg swiz) srcR) =
  destR == srcR && not neg && layout == swiz
