-- Conversion to "codegen normal form" (CNF).  To use after CSE, since
-- it doesn't try to preserve sharing, and after literal extraction
-- (ExtractLiterals), since it doesn't handle literals.

module CNF where

import Monad
import Maybe (catMaybes,isJust,fromJust)
import List (nub,elemIndex)

import Types
import DExp
import SmartDExp
import Asm

-- The normal form is designed to reflect what the processor can
-- do, so code generation will be easy, handling not much more than
-- register allocation and argument reordering.

-- CNF is a subset of the DExp.Exp type such that:
-- 
-- + There are no literals (see ExtractLiterals).
-- 
-- + Operators other than MkVec may only be applied to only to "args",
--   which are swizzled and possibly negated variables.  Actually, to
--   expressions that result from "fromArg arg" for some arg :: Arg.  The
--   toArg function gets back the readily-apparent structure.
--   
-- + Swizzling, negation, and variables show up *only* in these args.  (If
--   necessary, insert a Mov (identity) op application.

-- TODO: handle restriction of no more than one constant register and no
-- more than one vertex register used per op.

-- To distinguish between ordinary exps and CNF ones in type signatures
-- below, use the following:

type CNF = Exp

-- Variables will correspond to readable registers, possibly swizzled for
-- layout.  Swizzling and negation get rewritten away whenever possible,
-- by using distributive properties and pushing them into argument
-- position.  When we can't push them into an existing operator (possibly
-- by changing the operator, e.g., Min), introduce a new identity operator
-- (Mov) to push it into, which will cost an additional instruction.


---- The "argument" type, as supported directly by the instruction set.

data Arg = Arg NegSwiz Id Type deriving Show

-- Exp/Arg conversions

fromArg :: Arg -> CNF
toArg   :: CNF -> Arg

fromArg (Arg ns var ty) = applyNS ns ty (Var var ty)

toArg (Var var ty) = Arg (idNegSwiz ty) var ty

toArg (Apply (Swizzle swiz) [Var var ty]) =
  Arg (NegSwiz False swiz) var ty

toArg (Apply Negate [Var var ty]) =
  Arg (NegSwiz True (idSwizzle ty)) var ty

toArg (Apply Negate [Apply (Swizzle swiz) [Var var ty]]) =
  Arg (NegSwiz True swiz) var ty


-- Convenient creation & dissection of applications in CNF.

apply :: Op -> [Arg] -> CNF
apply op args = Apply op (map fromArg args)

dsApply :: CNF -> (Op,[Arg])
dsApply (Apply op exps) = (op, map toArg exps)

appArg1 :: Op -> Arg -> CNF
appArg1 = appL1 . apply

{-  -- Unused
appArg2 :: Op -> Arg -> Arg               -> CNF
appArg3 :: Op -> Arg -> Arg -> Arg        -> CNF
appArg4 :: Op -> Arg -> Arg -> Arg -> Arg -> CNF

appArg2 = appL2 . apply
appArg3 = appL3 . apply
appArg4 = appL4 . apply
-}

argToCNF :: Arg -> CNF

argToCNF (Arg ns var ty) | ns == idNegSwiz ty = Var var ty

argToCNF arg = appArg1 Mov arg

-- argToCNF wraps up its arg in a Mov (identity op), to absorb any
-- negation or swizzling.  If the arg is really being used as an argument,
-- the Mov application will stripped off, and the negation & swizzling
-- will be absorbed before we get to code generation.

---- Conversion to CNF

toCNF :: SourceEnv -> Exp -> CNF
toCNF sourceEnv = normalize
 where
   normalize :: Exp -> CNF

   -- MkVec & Out are special (and similar), because they don't correspond to
   -- a machine op.
   normalize (Apply MkVec exps)      = Apply MkVec (map normalize exps)
   normalize (Apply op@(Out _) exps) = Apply op (map normalize exps)

   -- Introduce multiply-add op.  Turn m*m'+a or a+m*m' into mad m m' a
--    normalize (Apply Add [Apply Mul [m,m'], a]) = normalize (Apply Mad [m,m',a])
--    normalize (Apply Add [a, Apply Mul [m,m']]) = normalize (Apply Mad [m,m',a])

   normalize (Apply Add [a,b]) | isJust mb = fromJust mb a where mb = tryMad b
   normalize (Apply Add [a,b]) | isJust mb = fromJust mb b where mb = tryMad a

   normalize (Apply Negate [exp]) | isJust opt = normalize (fromJust opt)
    where
      opt = negateOpt exp

   normalize (Apply (Swizzle swiz) [exp]) | isJust opt = normalize (fromJust opt)
    where
      opt = swizzleOpt swiz exp

   -- Fall-back for variables, Negate & Swizzle: insert a Mov and hope it
   -- gets absorbed in argument position.

   normalize e | needsMov = --ptrace ("add Mov to " ++ show e) $
                            normalize (apply1 Mov e)
    where
      needsMov = case e of Var _ _             -> True
                           Apply Negate _      -> True
                           Apply (Swizzle _) _ -> True
                           _                   -> False

   -- For other applications, replace complex argument expressions with
   -- introduced args and generate the corresponding let bindings.

   normalize (Apply op exps) = mkLet (catMaybes mbBinds) (apply op args)
    where
     (args,mbBinds) =
       unzip (zipWith3_3 mkArg [0..] (inits' sources) (map asNegSwiz exps))
     sources = map (\ (Arg _ v _) -> lookupVC v) args

   -- Note tricky circularity between args & sources.  The extra laziness
   -- of inits' & zipWith3_3 over inits & zipWith3 is crucial for
   -- termination.

   normalize (Let dbinds dbody) = Let (zip vars (map normalize exps))
                                      (normalize dbody)
    where
      (vars,exps) = unzip dbinds

   normalize Undefined = Undefined

   -- For each argument, generate an op arg and possibly a binding, given the
   -- argument number, previous vars, and a NegSwiz's exp.  The previous vars
   -- are given to help avoid using two vertex or two constant registers in
   -- the same application, which the HW architecture cannot handle.
   mkArg :: Int -> [Maybe Source] -> (NegSwiz, Exp) -> (Arg, Maybe (Id,CNF))

   mkArg _ prevs (ns, Var v ty) | okVar prevs v = (Arg ns v ty, Nothing)

   mkArg i _ (ns, exp)      = (Arg ns v (typeOf exp), Just (v, normalize exp))
     where v  = "p_" ++ show i

   -- See if the given var conflicts with any of the previous ones.
   okVar :: [Maybe Source] -> Id -> Bool
   okVar prevs var = not (any (conflict (lookupVC var)) prevs)

   conflict :: Maybe Source -> Maybe Source -> Bool
   conflict Nothing _ = False
   conflict _ Nothing = False
   conflict (Just (Source _ (Reg cls  n  _)))
            (Just (Source _ (Reg cls' n' _))) = cls == cls' && n /= n'

   lookupVC :: Id -> Maybe Source
   lookupVC = flip lookup sourceEnv

   tryMad :: Exp -> Maybe (Exp -> CNF)
   tryMad = try . stripLets
    where
      try (bindss,Apply Mul [m,m']) =
        Just (\ a -> applyLets (bindss, normalize (Apply Mad [m,m',a])))
      try (bindss,Apply Negate [Apply Mul [m,m']]) =
        Just (\ a -> applyLets (bindss, normalize (Apply Mad [Apply Negate [m],m',a])))
      try _ = Nothing


stripLets :: Exp -> ([[(Id,Exp)]],Exp)
stripLets (Let binds body) = (binds:bindss,body')
                              where (bindss,body') = stripLets body
stripLets e                = ([],e)

applyLets :: ([[(Id,Exp)]],Exp) -> Exp
applyLets (bindss,body) = foldr Let body bindss



-- Dissect expression as if NegateSwizzle
asNegSwiz :: Exp -> (NegSwiz, Exp)
asNegSwiz exp = --ptrace ("asNegSwiz " ++ show exp) $
                gather False (idSwizzle (typeOf exp)) exp
 where
 gather negB swizB exp =
   case exp of
   Apply Mov [e]             -> gather negB swizB e
   Apply Negate [e]          -> gather (not negB) swizB e
   Apply (Swizzle swizA) [e] -> gather negB (swizB `compSwiz` swizA) e
   _                         -> (NegSwiz negB swizB, exp)


-- Rewrite away the negation where possible.
negateOpt :: Exp -> Maybe Exp

negateOpt (Apply Negate [e]) = return $ e
negateOpt (Apply Add exps)   = return $ Apply Add (negateAll exps)
negateOpt (Apply Max exps)   = return $ Apply Min (negateAll exps)
negateOpt (Apply Min exps)   = return $ Apply Max (negateAll exps)
negateOpt (Apply Mad exps)   = return $ -- 3/5/3 I just found this bug.
                                        -- Where does it come up??
                                        error "tripped negateOpt Mad bug"
                                        -- Apply Max (negateTail exps)
                                        -- Should be:
                                        -- Apply Mad (negateTail exps)
negateOpt (Apply op exps)
  | op `elem` [Mul,Rcp,Dp3,Dp4] = return $ Apply op (negateHead exps)

negateOpt (Apply MkVec exps) = return $ Apply MkVec (negateAll exps)

negateOpt (Apply op@(Swizzle _) exps) = fmap (Apply op)
                                             (mapM negateOpt exps)

-- For a let, push the negation into the body if it can be then optimized
-- away.  Otherwise, it can be harmful to do so, because more registers
-- will be tied up when the negation is performed.  (Actually, I think it
-- can only be harmful if the result register is an output reg.)
negateOpt (Let binds body) = fmap (Let binds) (negateOpt body)

negateOpt _ = fail ""

-- Helpers
negateAll, negateHead, negateTail :: [Exp] -> [Exp]
negateAll = map negate
negateHead (exp:exps) = negate exp : exps
negateTail (exp:exps) = exp : negateAll exps


-- Similarly for swizzling.  For simd ops, distribute over args.  For
-- scalar-producing ops, swizzling is a no-op in the implementation, since
-- all elements are equal.  If we simply removed it here, the result would
-- have the wrong type, so insert a coercion operator to be ignored during
-- code generation.  Oh, rats, that won't work either, because it looks
-- like an opportunity for free negation.  For now, leave out the
-- coercion.  TODO: resolve the type error.

swizzleOpt :: Swizzle -> Exp -> Maybe Exp
swizzleOpt swiz app@(Apply op exps)
 | op `elem` simdOps = return $ Apply op (map (apply1 (Swizzle swiz)) exps)
 | op `elem` scalarResultOps = return $ app

swizzleOpt swiz (Apply MkVec exps) =
  return $ Apply MkVec (swizzleVals swiz exps)

swizzleOpt swizB (Apply op@(Swizzle swizA) exps) =
  return $ Apply (Swizzle (swizB `compSwiz` swizA)) exps

swizzleOpt swiz (Apply Negate exps) = negateOpt (Apply (Swizzle swiz) exps)

swizzleOpt swiz (Let binds body) = fmap (Let binds) (swizzleOpt swiz body)

swizzleOpt _ _ = fail ""

-- TODO: consider other ops when they get added, particularly messy ones
-- like Expp, Logp, Frc, MatMul.


-- inits in Hugs's prelude isn't as lazy as it might be.  It always yields
-- [] first, but it looks at its argument before doing so.  Here's a
-- lazier version:

inits' :: [a] -> [[a]]
inits' xs = [] : initsRest xs
 where
  initsRest []     = []
  initsRest (x:xs) = map (x:) (inits' xs)

-- Lazy on all but third argument
zipWith3_3 z ~(a:as) ~(b:bs) (c:cs) = z a b c : zipWith3_3 z as bs cs
zipWith3_3 _ _ _ _          = []



---- Testing

-- Try these with normalize:
t0 = Var "a" (Vec 2)
t1 = apply1 Negate t0    -- generates a Mov
t2 = apply2 Slt (Var "a" (Vec 2)) (Var "b" (Vec 2))
t3 = apply2 Slt t1 (Var "b" (Vec 2))    -- absorbs t1's Mov
