-- Common sub-expression elimination

module CSE (cse) where

import List (intersect,union,(\\),nub,elemIndex,find)
import Maybe (isJust,fromMaybe,catMaybes)
import Memo
import FiniteMap
import IOExts (unsafePerformIO)
import IORef


import Types
import DExp
import SimpleGraph
import Namer

---- CSE

cse :: [Id] -> Exp -> Exp
cse avoidVars exp = graphToExp avoidVars egn (abstractions (fst egn))
 where
   egn = expToGraph (uniquifyExp exp)


-- Replace equal expressions with eq ones.  Assume no lets.
uniquifyExp :: Exp -> Exp
uniquifyExp exp = uExp exp
 where
  uOp   = smemo copyOp
  uExp  = smemo copyExp
  uList = smemo copyList
  uType = smemo copyType

  copyOp (Swizzle swiz) = uSwizzle (uList swiz)
  copyOp (Out     regs) = uOut (uList regs)
  copyOp op             = op

  copyExp (LitVec vec)    = uLitVec (uList (map uFloat vec))
  copyExp (Var v ty)      = uVar v (uType ty)
  copyExp (Apply op exps) = uApply (uOp op) (uList (map uExp exps))
  copyExp (Let _ _)       = error "uniquifyExp: found Let"
  copyExp Undefined       = Undefined

  copyType (Vec n) = uVec n

  copyList []    = []
  copyList (h:t) = uCons h (uList t)

  uLitVec  = smemo  LitVec
  uVec     = smemo  Vec
  uVar     = smemo2 Var
  uApply   = smemo2 Apply
  uCons    = smemo2 (:)
  uSwizzle = smemo  Swizzle
  uOut     = smemo  Out

  -- Floats need care, since equal floats are often not pointer-equal
  uFloat = hashEqual ()

-- Strict memo function
smemo f = ((memo f) $!)

-- Simple memoizing identity function on Eq types.  When equal values are
-- put in, pointer-equal values come out.  Use it, for instance, with
-- ints, chars and floats, e.g., "memo factorial . hashEqual ()"

-- The first argument is just there to keep the
-- internal state from being a CAF.

{-# NOINLINE hashEqual #-}
hashEqual :: Eq a => () -> (a -> a)
hashEqual () = hash
 where
   valsRef = unsafePerformIO (newIORef [])
   hash x = unsafePerformIO $ do
            xs <- readIORef valsRef
            case find (== x) xs of
              Nothing -> do writeIORef valsRef (x:xs)
                            return x
              Just x' -> return x'


-- Our graphs are labeled with operators or atomic expressions (literal,
-- var, undefined).
data Label = Op Op | Simple Exp deriving (Eq,Show)


-- TODO: Out constructor.  Perhaps make it an Apply or vector??

type EGraph = Graph Label

-- Graph with distinguished node
type GraphN a = (Graph a, Node)

type EGraphN = (EGraph, Node) -- equivalently, GraphN Label

expToGraph :: Exp -> EGraphN
expToGraph = mkG emptyFM
 where
   mkG g exp = (if alreadyInG then g else extG, node)
    where
      node                 = ptrToNode exp
      alreadyInG           = isJust (lookupFM g node)
      (label,exps)         = dsExp exp
      (gWithArgs,argNodes) = loop g exps
      extG                 = addToFM gWithArgs node (argNodes,label)
      
      loop g [] = (g,[])
      loop g (arg:args) = (g'',argNode:argNodes)
       where
         (g' ,argNode ) = mkG g arg
         (g'',argNodes) = loop g' args


-- We want to put the abstraction as close as possible above all of the
-- occurrences of an expression, which is the lowest node d (furthest from
-- start) such that every path from the root to the shared expression.
-- This corresponds with the notion of immediate dominator.  This
-- mapping of nodes to their immediate dominator may be thought of as a
-- tree with links going upward to the root.

-- Note that (a) every node dominates itself, and (b) for distinct nodes d
-- and n, where n is not the root, d dominates n iff d dominates every
-- parent of n.

-- Compute a path of dominators leading from the root to the given node.
dominators :: Graph a -> Graph a -> Node -> Path
dominators g gConv = mdom
 where
   mdom = smemo dom
   dom node | null parents = [node]
            | otherwise = foldl1 commonPrefix (map mdom parents) ++ [node]
    where
      parents = children gConv node  -- predecessors

isNegOrSwiz (Op Negate)      = True
isNegOrSwiz (Op (Swizzle _)) = True
isNegOrSwiz _                = False

-- Intersect paths, yielding common prefix
commonPrefix :: Eq a => [a] -> [a] -> [a]
commonPrefix xs ys = map fst (takeWhile (uncurry (==)) (zip xs ys))

-- Immediate dominator as finite map
dominatorFM :: -- (Show a, Show b) =>
               (a -> Bool) -> Graph a -> Graph a -> [Node] -> NodeMap Node
dominatorFM trivial g gConv nodes = functionFM nodes domF
 where
   domF = pick
        . reverse
        . filter (not . trivial . lab g)
        . dominators g gConv
   -- Pick immediate dominator out of path from node up toward root
   pick (_:idom:_) = idom
   pick doms       = error ("dominatorFM: path too short.  " ++ show doms)

type NodeMap a = FiniteMap Node a

-- For each node n, which nodes get abstracted at n?
-- TODO: how to represent the set?  Some choices: list, characteristic
-- function, subgraph.
abstractions :: EGraph -> NodeMap [Node]
abstractions egraph =
  -- Form dominators; subset for shared nodes; invert site map
  invertFM (dominatorFM isNegOrSwiz egraph egConv
            (filter (curry (shouldAbstract egConv) egraph) (keysFM egraph))) 
 where
   egConv = converse egraph


-- We should abstract a graph node iff it's an operator application and
-- shared.  Exception: never abstract negations or swizzles, but sometimes
-- abstract an unshared expression if it's "virtually shared", meaning
-- that its unique parent is a negation or swizzle that is either shared
-- or virtually shared.  (Note: negations & swizzles must be rejected as
-- dominators.)

shouldAbstract :: EGraph -> EGraphN -> Bool
shouldAbstract gConv (g,n) = isOp label && not (isNegOrSwiz label) &&
                             (shared n || virtuallyShared n)
                             -- nontrivial n && length (children gConv n) > 1
 where
   label = lab g n
   shared n = length (children gConv n) > 1
   virtuallyShared n = length (children gConv n) == 1 &&
                       isNegOrSwiz (lab g parent) &&
                       (shared parent || virtuallyShared parent)
    where
      parent = head (children gConv n)

   isOp (Op _) = True
   isOp _      = False

-- Place abstractions implied by node map
graphToExp :: [Id] -> EGraphN -> NodeMap [Node] -> Exp
graphToExp avoidVars (egraph,n0) absMap =
  runNamer (allIds \\ avoidVars)
           (fromG emptyFM n0)
 where
   fromG :: NodeMap Exp -> Node -> Namer Exp
   fromG env node =
     -- If this node is already abstracted as a variable, use that
     -- variable.  Otherwise, recursively convert children, make an
     -- expression, and wrap with bindings if any (if node is an
     -- abstraction site).
     maybe (mkLets (stratify egraph absNodes) env)
           return
           (lookupFM env node)
    where
      Just (children,label) = lookupFM egraph node
      -- Topologically sorted set of nodes to be abstracted here
      absNodes = sortF topoNum (fromMaybe [] (lookupFM absMap node))
      
      mkLets :: [NGroup] -> NodeMap Exp -> Namer Exp
      mkLets [] env = fmap (mkExp label) (mapM (fromG env) children)
      mkLets (group1:groups) env = do
        exps <- mapM (fromG env) group1
        names <- getNames (length exps)
        let vars = zipWith Var names (map typeOf exps)
            env' = addListToFM env (zip group1 vars)
        body <- mkLets groups env'
        return $
          Let (zip names exps) body
   -- Position in topological ordering of all nodes.
   topoNum :: Node -> Int
   topoNum = smemo (negate .  -- why
                    fromMaybe (error "graphToExp: node missing in topsort") .
                    flip elemIndex (topsort egraph))

type NGroup = [Node]

-- Stratify a given top-sorted set of nodes to be abstracted into a
-- sequence of node sets, in which elements of later sets may depend on
-- elements of earlier sets.
stratify :: Graph a -> NGroup -> [NGroup]
stratify g = strat [] []
 where
  strat :: [NGroup] -> NGroup -> NGroup -> [NGroup]
  strat groups group [] = reverse $ ifF (not (null group)) (group :) groups
  strat groups group nodes@(node:nodes')
    | node `dependsOn` group = strat (group:groups) [] nodes
    | otherwise              = strat groups (node:group) nodes'
  
  dependsOn :: Node -> NGroup -> Bool
  node `dependsOn` group = --ptrace ("dependsOn " ++ show (node,group)) $
                           not (null (depthFirst g node `intersect` group))



---- Misc

mkExp                 :: Label -> [Exp] -> Exp
mkExp (Op op) exps    = Apply op exps
mkExp (Simple exp) [] = exp

dsExp                 :: Exp -> (Label, [Exp])
dsExp (Apply op exps) = (Op op, exps)
dsExp exp             = (Simple exp, [])


-- Memoize on more than one argument
smemo2 :: (a -> b -> c) -> (a -> b -> c)
-- etc

smemo2 f = smemo (smemo  . f)
smemo3 f = smemo (smemo2 . f)
smemo4 f = smemo (smemo3 . f)
              

{-
---- Testing

g1 = mkUGraph [1,2] [(1,2)]

t1 = map (dominators g1) [1,2]
t2 = dominatorFM g1 [2]

-}