-- Code generation from CNF (codegen normal form) expressions.  Handles
-- and register allocation.  Uses dynamic programming to minimize register
-- use by reordering operator arguments.

module CodeGen where

import List (sortBy,elemIndex,isPrefixOf,partition,intersperse)
import Maybe (fromMaybe,fromJust)
import IOExts (trace)

import Types
import CNF
import Asm
import Pool
import DExp
import PPDExp (showExp)

-- A code generator says how much free register space needed (in floats,
-- e.g., three per Vec3) and how to generate code.

type CodeGen = (Int, Gen)

-- Generate code for destination register, extra swizzle to accommodate
-- the destination layout, mapping from variables to sources, free temp
-- pool.
type Gen = Dest -> Swizzle -> SourceEnv -> Pool -> Asm

codegen :: CNF -> CodeGen

codegen (Apply MkVec         exps) = codegenMkVec exps
codegen (Apply (Out regNums) exps) = codegenOut regNums exps

codegen exp@(Apply op argExps) = (0, gen)
 where
   gen dest swiz env _ =
     [PrimOp op dest (map (getSource tweakNS . toArg) argExps) comment]
    where
      tweakNS | op `elem` simdOps = swizzleNS swiz  -- See comment below
              | ignoreSwizzle     = id
              | otherwise         = error $ "swizzle puzzle " ++ show (op,swiz)
      -- Ok to ignore swizzle?
      ignoreSwizzle =    op `elem` scalarResultOps  -- components identical
                      || isPrefixOf swiz [0..]      -- identity swizzle
      getSource :: (NegSwiz -> NegSwiz) -> Arg -> Source
      getSource tweakNS (Arg nsB v _) = Source (tweakNS (nsB `composeNS` nsA)) r
       where
         Source nsA r = lookupEnv env v
      
      comment = destName dest ++ " = " ++ showExp (tweakExp exp)
      destName (Dest (Reg _ _ name) _) = name
      tweakExp (Apply Mov [e])     = e  -- optional prettification
      tweakExp (Apply Mad [a,b,c]) = Apply Add [Apply Mul [a,b],c]
      tweakExp e                   = e


-- In the Apply case above we need tweakNS due to a tricky point arising
-- from layout.  Variables smaller than Vec4 may require swizzling on
-- write, which is not supported in general by the processor architecture.
-- However, for *almost all* operations, a write swizzle can be correctly
-- simulated by a combination of write masking and argument swizzling.
-- For simd ops (add, mul, mad, min, etc), just swizzle each argument
-- correspondingly.  For scalar-producing ops (dp4, rcp, rsq, etc), the
-- same scalar result is written to all components of the output, so write
-- swizzling is just write masking.  An exception is frc, which insists on
-- writing to x & y.  That means we have to allocate the result in x&y.
-- However, frc is a macro, and we just don't use it.
-- 
-- Other odd ones: dst, expp, lit, logp.  Oh, they all write to all four
-- components, so don't pose a problem, unless I used a non-obvious
-- layout.  We handle this concern by having Pool.allocReg guarantee to
-- produce the identity layout for four-float allocations, so that
-- unpredictable layout swizzling cannot happen.  Explicit swizzling still
-- may, and so may require an explicit instruction.  Hmm... seems just
-- lucky and hence not robust.  For now, if we reach a case we can't
-- handle, raise "swizzle puzzle" error.

codegen Undefined = (0, \ _ _ _ _ -> [])

-- For an n-ary Let, there are n+1 stages of evaluation: one for each rhs,
-- and one for the body.  The register use will be the maximum of the
-- register uses over the n+1 stages.  At each stage, we have to preserve
-- the registers used to hold the results of previous stages.  Since later
-- stages have the added burden of preserving earlier results, we
-- rearrange the bindings to put the less register-intensive rhs's later,
-- thus minimizing the maximum register usage over the stages.  We cannot
-- move the body, since it depends on all the bindings.  (Assume useless
-- bindings are eliminated in an earlier stage of computation.)

codegen (Let bindings body) = (nr,gen)
 where
   (vars,types,cgs) = unzip3 (reorder (zip3 vars0
                                            (map typeOf exps0)
                                            (map codegen exps0)))
                      where (vars0,exps0) = unzip bindings
   (nrs,gens) = unzip cgs
   (nrb,genb) = codegen body

   -- Total space tied up at each stage.
   savedRs = scanl1 (+) (0 : map (ifF saveDest (+ bsize) . type2size) types)
   nr = maximum (zipWith (+) (nrs ++ [nrb]) savedRs)
   gen dest swiz env pool = asm ++ genb dest swiz env' pool'
    where
      (asm,env',pool') = genBindings vars types gens env poolRhs
      poolRhs = ifF (saveDest && destClass == RegTemp)
                (removeFromPool destNum destMask)
                pool
      Dest (Reg destClass destNum _) destMask = dest
   -- If the body is a MkVec, possibly wrapped in let bindings, then we
   -- may have to preserve the dest register while computing the RHSs.
   -- Otherwise, a rhs register could get overwritten in the middle of the
   -- sequence of instructions that write the vector.  This problem can
   -- only arise with MkVec, since it's the only operator that yields more
   -- than one instruction.
   bsize = type2size (typeOf body)
   saveDest = isLetsMkVec body
    where
      isLetsMkVec (Apply MkVec _) = True
      isLetsMkVec (Let _ body)    = isLetsMkVec body
      isLetsMkVec _               = False

   -- Sort by decreasing register use
   reorder :: [(Id,Type,CodeGen)] -> [(Id,Type,CodeGen)]
   reorder = sortF (negate . \ (_,_,(nr,_)) -> nr)


negSwizLookup :: SourceEnv -> NegSwiz -> Id -> Source
negSwizLookup env nsB v =
  Source (nsB `composeNS` nsA) r
 where
    Source nsA r = lookupEnv env v


-- Generate code for bindings.  Allocate one register per binding,
-- removing it from the pool for later bindings and for the body of the
-- enclosing Let.  Returns the code, extended environment, and diminished
-- pool.

genBindings :: [Id] -> [Type] -> [Gen] -> SourceEnv -> Pool
            -> (Asm,SourceEnv,Pool)
genBindings [] [] [] env pool = ([],env,pool)

genBindings (var1:vars) (type1:types) (gen1:gens) env pool =
  (asm1 ++ asm, extendEnv env' [(var1,src1)], pool')
 where
   (regNum1,layout1,pool1) = allocReg (type2size type1) pool
   reg1                    = Reg RegTemp regNum1 var1
   asm1                    = gen1 dest1 swizzle1 env pool
   (asm,env',pool')        = genBindings vars types gens env pool1
   dest1                   = Dest reg1 layout1
   src1                    = Source (NegSwiz False layout1) reg1
   swizzle1                = layoutWriteSwizzle layout1



-- Output.  Use embedded output register.  There's no need to preserve
-- registers, since no temp registers are not being written.  For that
-- reason, the order of evaluation doesn't matter.
codegenOut regNums exps = (nr, gen)
 where
 (nrs,gens) = unzip (map codegen exps)
 nr         = maximum nrs
 gen _ _ env pool = concat (zipWith3 genPart regNums gens (map typeOf exps))
  where
    genPart regNum gen ty = gen dest mask env pool
     where
       mask = idSwizzle ty
       dest = Dest (Reg RegOut regNum (showDest dest)) mask


-- The MkVec case is slightly different from the general Apply case,
-- because its arguments are written directly into a vector's float
-- components.  As such, they're in CNF rather than argument form.

codegenMkVec exps = (nr, gen)
 where
 -- Register use is max of the uses of the exps.  Use the float
 -- subregisters rather than allocating new ones.  Since we don't have
 -- to preserve registers, the evaluation order doesn't matter.
 -- Oops!  That's not true.  We do have to preserve the registers that
 -- hold each component while computing the others.
 -- TODO: revisit and fix well.  Remove Pool.removeFromPool when I do this
 -- better.
 (nrs,gens) = unzip (map codegen exps)
 nr         = maximum nrs
 gen dest swiz env pool = concatMkVecAsms (genParts gens [0..] layout pool)
  where
    -- Generate for each exp, but put in swizzled positions.
    genPart gen part pool =
      gen (destPart part dest) (idSwizzle scalar) env pool -- Why idSwizzle?
    -- genParts gens dests = zipWith genPart gens dests
    genParts [] _ _ pool = []
    genParts (gen:gens) (valPart:valParts) (regPart:regParts) pool =
      genPart gen valPart pool :
      genParts gens valParts regParts (removeFromPool regNum [regPart] pool)
    genParts _ _ _ _ = error "genParts"
    Dest (Reg _ regNum _) layout = dest

-- TODO: reorder MkVec arguments and calculate nr as in Let.  The
-- reordering in concatMkVecAsms shouldn't interfere, since it keeps all
-- but the moves in front, and moves don't use any registers.
-- 
-- Idea: in CNF, don't allow a let in a MkVec argument.  Floating it out
-- would require renaming.  The MkVec args would then have to be operators
-- applied to possibly negated & swizzled variables.  That would sure
-- simplify things, at a small cost in optimality of register usage.  The
-- MkVec case would then never need to allocate space.  It would just
-- evaluate its simple arguments right into the allocated vector.  I like
-- this idea.

-- Perform a MkVec-specific optimization over simple concatenation.
-- When components of the register are copied into different components of
-- the result, coalesce the writes.
concatMkVecAsms :: [Asm] -> Asm
concatMkVecAsms asms = concat nonMoves ++ coalesce (map head moves)
 where
   (moves,nonMoves)          = partition isMove asms
   isMove [PrimOp Mov _ _ _] = True
   isMove _                  = False
   -- Of the moves, group by source register and coalesce within each group
   coalesce :: [Instr] -> Asm
   coalesce = map (merge . sortF destMaskIx . snd) .
              gather .
              withF sourceNegReg
   -- Extract negation, register class & index, for sorting on
   sourceNegReg :: Instr -> (Bool,RegClass,Int)
   sourceNegReg (PrimOp Mov _ [(Source (NegSwiz neg _) (Reg cls ix _))] _) =
     (neg,cls,ix)
   destMaskIx (PrimOp Mov (Dest _ [maskIx]) _ _) = maskIx
   -- Merge multiple moves into a single, given that all have the same
   -- negation, register class & index.  We've already sorted the moves by
   -- mask index so we can simply concat the mask indices with no
   -- rearrangements.
   merge :: [Instr] -> Instr
   merge moves = PrimOp Mov dest [source] comment
    where
      dest    = Dest destReg mask
      source  = Source (NegSwiz neg (padSwiz mask swiz)) sourceReg
      comment = concat (intersperse ", " comments)
      PrimOp Mov (Dest destReg _)
             [Source (NegSwiz neg _) sourceReg] _ = head moves
      (mask,swiz,comments) = unzip3 (map maskSwizComment moves)
      maskSwizComment :: Instr -> (Int,Int,String)
      maskSwizComment (PrimOp Mov (Dest _ [maskIx])
                              [Source (NegSwiz _ [swizIx]) _] comment) =
         (maskIx,swizIx,comment)
      -- Pad swizzle for components missing in the mask.  Also, prettify
      -- in two ways.  First, if the swizzle components are all the same,
      -- just use it as a single-component swizzle.  Second, don't bother
      -- adding swizzle components after the max mask part.
      padSwiz :: Mask -> Swizzle -> Swizzle
      padSwiz mask swiz
       | all (== head swiz) (tail swiz) = [head swiz]
       | otherwise                      = map sw [0 .. maximum mask]
       where
         sw part = fromMaybe part (lookup part msAssoc)
         msAssoc = zip mask swiz


