-- Test compiler phases.  Use with the examples from TExp.

module Compile where

import Maybe (isJust,fromJust,catMaybes,fromMaybe)
import List (intersperse,nub,partition)
import Monad (when)
import Char (isSpace)
import qualified System

import PP
import Types
import qualified DExp as D
import CSE
import Pool
import TExp hiding (cse)
import ExtractLiterals
import Approx
import CNF (toCNF)
import Asm
import CTrans
import CsPPDExp
import CodeGen
import SampleExp

-- To help make examples on the fly:
import Vector
import SwizzleFuns
import NameStruct

-- Whether the compiler shows intermediate stages
verboseCompile = False

-- Whether to run a demo after compiling.
runCompiled = True

-- Directory for demos.  The Vertigo Viewer.dll is presumed to be there.
demoDir = "..\\Demos"


-- How much to exaggerate about registers.  Set to more than one to see
-- how close we came to running out.
exaggerate = 2 :: Int

-- Number of registers of each type.  Perhaps move to Asm.hs.
numConsts = 96 * exaggerate
numVertex = 16 * exaggerate
numTemp   = 12 * exaggerate

compile :: (AsVar uN uniform, AsVar vN varying)
        => String -> (Int,Int) -> uN -> vN
        -> [Range]  -- For some scalar uniform vars
        -> (uniform -> varying -> Exp o) -> IO ()
compile name tess uN vN uranges f =
  compileOut name tess uniformEnv varyingEnv uranges (wrapOut dexp)
 where
   (E dexp,uniformEnv,varyingEnv) = staged uN vN f

-- Simple tester.
compileC :: Exp a -> IO ()
compileC (E dexp) = compileOut "" (100,100) [] (map unVar vertexVars) [] (wrapOut dexp)
 where
   unVar (D.Var name ty) = (name,ty)

-- Wrap dexp as an output if it's not already
wrapOut :: D.Exp -> D.Exp
wrapOut dexp = ifF (D.typeOf dexp /= Output) (D.apply1 (D.Out [0])) dexp

compileOut :: String -> (Int,Int) -> TypeEnv -> TypeEnv -> [Range] -> D.Exp -> IO ()
compileOut name tess uniformVars varyingVars uranges dexp = 
  do when verboseCompile $
       putStr $ ""
                ++ "Uniform: "                  ++ showTyE uniformNM
                ++ "\nVarying: "                ++ showTyE varyingVars
                ++ "\nBefore approx:\n"         ++ show' (E beforeApprox)
--                 ++ "\nC# normal form:\n"        ++ show' (E csnf)
--                 ++ "\nC# code:\n"               ++ csCodeStr
                ++ "\nAfter approx:\n"          ++ show' (E postCse)
                ++ "\nConstant registers:\n"    ++ show' litRegs'
                ++ "\nWithout literals:\n"      ++ show' (E noLits)
                ++ "\nAfter CNF conversion:\n"  ++ show' (E cnf)
                ++ "\nGenerated code:\n"        ++ shownAsm
                ++ "\nStatistics:  "            ++ stats
                ++ "\nConstant setting code:\n" ++ showPPDoc' constCode 0
     when (name /= "") $ do
       writeFile (srcName ++ ".cs") csCode
       putStrLn $ "Wrote " ++ srcName ++ ".cs"
       syscall (cscCompile demoDir name)
       when runCompiled $ shellOpen srcName
       return ()
 where
   beforeApprox      = cse (map fst env1) dexp  -- just for show
   csnf              = csnorm beforeApprox
   csCodeStr         = showPPDoc' (ppExpForCs beforeApprox) 0
   approxExp         = approx dexp
   postCse           = cse (map fst env1) approxExp
   (noLits,litRegs)  = extractLiterals numConsts postCse
   (matRs,uniformNM) = removeMatrices uniformVars
   varyingEnv        = runAlloc (fullPool numVertex) $
                         mkSourceEnv RegIn varyingVars
   uniformEnv        = runAlloc (constantPool litRegs) $
                         mkSourceEnv RegConst uniformNM
   env1              = varyingEnv ++ uniformEnv ++ literalEnv numConsts
   cnf               = toCNF env1 noLits
   litRegs'          = takeWhile (not.null) litRegs
   (nr,gen)          = codegen cnf
   asm               = -- Generate code, using bogus dest & mask, since the
                       -- Output will provide one.
                       gen (bogus "dest") (bogus "mask") env1 (fullPool numTemp)
   prelude           = "vs.1.1\n"
                       ++ concat (zipWith constDef [0..] litRegs')
                       ++ prefixLines "; "
                           (   showSourceEnv "uniform" uniformEnv
                            ++ showSourceEnv "varying" varyingEnv
                           )
                       ++ unlines (map vertexDecl (vertexRegsUsed varyingEnv))
   showAsm'          = indentLines 4 . (prelude ++) . showAsm
   shownAsm          = showAsm' asm
   showPPDoc' x p    = indentLines 4 (showPPDoc x p)
   bogus str         = error $ "compileOut: bogus " ++ str
   constCode         = constSetter litRegs'
                         (restoreMatrices matRs uniformEnv) uranges
   stats             = show nr ++ " float temporaries, " ++
                       show numConstFloats ++ " float constants, and " ++
                       show (length asm) ++ " instructions."
   numConstFloats    = sum (map length litRegs') + sum (map (type2size.snd) uniformNM)
   csCode            = wrapCsCode name $
                       ""
                       ++ csComment ("Vertex shader \"" ++ name ++ "\".  "
                                             ++ "Uses " ++ stats)
                       ++ "public " ++ name ++ "() : base(@\"\n"
                       ++ shownAsm ++ "\"){ }\n"
                       ++ showPPDoc (constCode
                                     $$ uiCode name tess uranges
                                     $$ mainCode
                                    ) 0
   srcName = demoDir ++ "\\" ++ name

vertexDecl :: Int -> String
vertexDecl i = "dcl_texcoord" ++ si ++ " v" ++ si  where si = show i

vertexRegsUsed :: SourceEnv -> [Int]
vertexRegsUsed = nub . map (vregNum . snd)
 where
   vregNum (Source _ (Reg RegIn i _)) = i

removeMatrices :: TypeEnv -> (Env [Id], TypeEnv)
removeMatrices typings = (catMaybes mbs, concat typings')
 where
   (mbs,typings') = unzip (map deMatrix typings)

deMatrix :: (Id,Type) -> (Maybe (Id,[Id]), TypeEnv)
deMatrix (name, Matrix cols rows) =
  (Just (name, rowNames), map (`pair` (Vec cols)) rowNames)
 where
   rowNames = map (matVarName name) [1 .. rows]
deMatrix nty = (Nothing,[nty])


restoreMatrices :: Env [Id] -> SourceEnv -> SourceEnv
restoreMatrices matEnv = id

-- WORKING HERE: use the matEnv to re-introduce matrices.


showSourceEnv :: String -> SourceEnv -> String
showSourceEnv label env =
  label ++ ": " ++
  concat (intersperse ", " (map showBinding env))
  ++ "\n"
 where
   showBinding (var,src) = var ++ " = " ++ showSource src
     

-- Generate a constant def, padding vector with zeros as necessary.
constDef :: Int -> Vector -> String
constDef n vec =
  "; c" ++ show n ++ " = {" ++
  concat (intersperse ", " (map show (take 4 (vec ++ repeat 0))))
  ++ "}\n"

showTyE :: TypeEnv -> String
showTyE = concat . intersperse ", " . map showTyping
 where
   showTyping (v,ty) = v ++ "::" ++ show ty
          

-- Pool remaining after literal extraction
constantPool :: [Vector] -> Pool
constantPool = catMaybes . zipWith remaining [0..]
 where
   remaining i vec | size == 4 = Nothing
                   | otherwise = Just (i,[size..3])
     where size = length vec

mkSourceEnv :: RegClass -> TypeEnv -> Alloc SourceEnv
mkSourceEnv cls = mapM mkBinding
 where
   mkBinding (name,ty) = do (regNum,layout) <- alloc (type2size ty)
                            return $
                              (name, Source (NegSwiz False layout)
                                            (Reg cls regNum name))


literalEnv :: Int -> SourceEnv
literalEnv n = [ (v, Source (idNegSwiz (Vec 4)) (Reg RegConst i v))
               | i <- [0..n-1], let v = constVarName i
               ]

setVertexConst = "device.SetVertexShaderConstant"

constSetter :: [Vector] -> SourceEnv -> [Range] -> PPDoc
constSetter constVecs uniformEnv uranges =
  ppMethod "public override void" "SetConsts" decls assigns
 where
   -- We filter out float-valued parameters and make them into sliders.
   decls = text "Device device" :
           map ppUnifDecl noRanges
           ++ [text "float[] floats"]
   assigns = [ text ("float " ++ var ++ " = floats[" ++ show i ++ "]")
             | ((var,_),i) <- zip hasRanges [0..] ]
             ++
             setterCode constVecs uniformEnv
   
   (hasRanges,noRanges) = partition hasRange uniformEnv
   hasRange (var, _) = var `elem` rangedVars
   rangedVars = [ var | Range var _ _ _ <- uranges ]

ppUnifDecl :: (Id,Source) -> PPDoc
ppUnifDecl (var, Source (NegSwiz _ layout) _) = text ty <+> text var
 where
   ty = if size == 1 then "float" else "Vector" ++ show size
   size = length layout


-- To make the setter code, first turn every literal and uniform variable
-- into a list of (register #, (part #, FloatSource)), where FloatSource
-- is a simple data type of either a float literal or variable name and
-- part #. Then regroup by reg #, and merge the (part#,FloatSource) pairs
-- into a single piece of code that constructs a float array.  As an
-- optimization, may construct a simple Vertex4 instead.

data FloatSource = FSLit Float          -- float literal
                 | FSVar Id Type Int    -- var, type & part
  deriving Show

type ConstPart   = (Int, FloatSource)   -- register part & float src
type ConstAssign = (Int, ConstPart)     -- register # & const part
type VecAssign   = (Int, [FloatSource]) -- register # & all sources

setterCode :: [Vector] -> SourceEnv -> [PPDoc]
setterCode constVecs uniformEnv =
  map setConst $
  combineFS  $
    -- Earlier register claims take precedence over later ones (due to
    -- lookup in extractFSs), so provide uniform vars before literals.   
    concatMap uniformFS uniformEnv
    ++
    concat (zipWith literalFS [0..] constVecs)

literalFS :: Int -> Vector -> [ConstAssign]
literalFS regNum vec = withReg regNum (map FSLit vec)
 where
   withReg :: Int -> [FloatSource] -> [ConstAssign]
   withReg regNum = map (regNum `pair`) . zip [0..]

uniformFS :: (Id,Source) -> [ConstAssign]
uniformFS (var, Source (NegSwiz _ layout) (Reg RegConst regNum _)) =
  zipWith assign layout [0..]
 where
   assign regPart varPart = (regNum, (regPart, FSVar var ty varPart))
   ty = Vec (length layout)

-- ufs1 = uniformFS ("e", Source (NegSwiz False [3]) (Reg RegConst 6 "e"))

combineFS :: [ConstAssign] -> [VecAssign]
combineFS = map (onSnd extractFSs) . gather

extractFSs :: [(Int,FloatSource)] -> [FloatSource]
extractFSs assigns = map (fromMaybe (FSLit 0) . flip lookup assigns) [0..3]

-- TODO: does extractFSs do the right thing

setConst :: VecAssign -> PPDoc
setConst (regNum, fss) = setVC regNum $
                         if isJust asSing then
                            newArr "Vector4" [text (fromJust asSing)]
                         else newArr "float" (map ppFS fss)
 where
   asSing = asSingleVar4 fss

asSingleVar4 :: [FloatSource] -> Maybe Id
asSingleVar4 fss = do (var:vars) <- mapM asFSVar fss
                      if all (== var) vars then
                         return var
                       else fail "" 
 where
   asFSVar (FSVar var _ _) = Just var
   asFSVar _               = Nothing

ppFS :: FloatSource -> PPDoc
ppFS (FSLit x)              = ppFloat x
ppFS (FSVar v (Vec 1) part) = text v
ppFS (FSVar v ty part)      = getPart ty part (text v)

getPart :: Type -> Int -> PPDoc -> PPDoc
getPart (Vec 1) 0 exp = exp
getPart ty   part exp = exp <> char '.' <> char ("XYZW"!!part)

setVC :: Int -> PPDoc -> PPDoc
setVC i vec = ppCall setVertexConst [ppInt i, vec]


-- Generate UI code

uiCode :: String -> (Int,Int) -> [Range] -> PPDoc
uiCode name (uverts,vverts) uranges =
  ppMethod "public static Form" "MakeDemo" []
           [ text "return" <+>
             ppCall "new Demo" [ text (quote name)
                               , ppInt uverts, ppInt vverts
                               , ppCall ("new " ++ name) []
                               , text ("new VSlider[]")
                                 $$ stmtBlock 2
                                      (vsepWith comma (map mkSlider uranges))
                               ]
           ]
 where
   mkSlider :: Range -> PPDoc
   mkSlider (Range name min max init) =
     ppCall "new VSlider" (map text [quote name, f min, f max, f init])
    where
      f e = show e ++ "f"


mainCode :: PPDoc
mainCode = text "[STAThread]" $$
           ppMethod "static void" "Main" []
                    [ ppCall "Application.Run"
                       [ ppCall "MakeDemo" [] ] ]



-- Wrap up the generated C# code in a complete program module
wrapCsCode :: String -> String -> String
wrapCsCode name =   surround (unlines prelude) (unlines postlude)
                  . wrapNameSpace
                  . wrapClass
 where
   prelude           = [ "// Vertex shader \"" ++ name
                         ++ "\", generated by Vertigo compiler."
                       , ""
                       , concatMap (surround "using " ";\n") usedNameSpaces
                       ]
   postlude          = [ ]
   usedNameSpaces    = ["System", "Microsoft.DirectX"
                       , "Microsoft.DirectX.Direct3D"
                       -- , "Microsoft.DirectX.Direct3DX"
                       , "System.Windows.Forms"
                       ]
   wrapClass         =   (("\npublic class " ++ name ++ " : VShader ") ++)
                       . wrapBraces
                       . indentLines 4
   wrapNameSpace str = "namespace Vertigo" ++ " " ++ wrapBraces str


cscdir = "c:/WINDOWS/Microsoft.NET/Framework/v1.0.3705"
dxdir = "c:/WINDOWS/Microsoft.NET/Managed DirectX/v4.09.00.0900"

cscCompile :: String -> String -> String
cscCompile dir name =
           "cd " ++ dir ++ "\n" ++
           cscdir ++ "/csc.exe /target:winexe /reference:Viewer.dll"
           ++ concatMap ((',':) . quote . surround (dxdir ++ "/Microsoft.DirectX") ".dll") dxlibs
           ++ " " ++ name ++ ".cs"
 where
   dxlibs = ["",".Direct3D",".Direct3DX"]

testcc = cscCompile "BasicLightingNH"

-- The "system" command won't accept long strings
syscall str = do writeFile "foo.cmd" str
                 System.system "foo.cmd"
                 return ()

shellOpen exe = do System.system ("start " ++ exe)
                   return ()

quote = surround "\"" "\""
