-- Pretty printing of expressions for C#.  Handles operator precedence and
-- some simple syntactic sugaring.  Sync changes with PPDExp.

module CsPPDExp (ppExpForCs) where

import Maybe (fromMaybe,fromJust)
import DExp
import Types
import PP
import CTrans (csnorm)

ppExpForCs = ppExp . csnorm

ppExp :: Exp -> PPDoc
ppExp (LitVec xs)      = ppNewVec (map (text.show) xs) 
ppExp (Var n ty)       = text n
ppExp (Apply op args)  = ppApply op args
ppExp (Let binds body) = vsep (map ppBind binds) $$
                         ppExp body
ppExp e                = error ("CsPPDExp.ppExp: no case for " ++ show e)


ppNewVec :: [PPDoc] -> PPDoc
ppNewVec [x] = x
ppNewVec xs  = ppCall ("new Vector" ++ show (length xs)) xs
                      
ppBind (v,e) = text v <+> text "=" <+> ppExp e <> semi

ppApply op = pp op
 where
   pp Add [e1, LitVec [x]]
     | x < 0                      = ppOp (ppBinOp "-" e1 (LitVec [-x]))
   pp Add [e1, Apply Negate [e2]] = ppOp (ppBinOp "-" e1 e2)
   pp Add [e1, e2]                = ppOp (ppBinOp "+" e1 e2)
   pp Mul [e1, Apply Rcp [e2]]    = ppOp (ppBinOp "/" e1 e2)
   pp Mul [e1, e2]                = ppOp (ppBinOp "*" e1 e2)
   pp Slt [e1, e2]                = ppOp (ppBinOp "<" e1 e2)
   pp Sge [e1, e2]                = ppOp (ppBinOp ">=" e1 e2)
   pp Negate [e1]                 = ppOperator (apPrec-1,LAssoc)
                                      (text "-" <> onRight (ppExp e1))
   pp MkVec args                  = ppNewVec (map ppExp args)
   pp (Swizzle ixs) [e]           = ppSwiz ixs e
   pp op args                     = ppCall (opNameCs op) (map ppExp args)
--                                     ppOperator (apPrec,LAssoc) $
--                                       ppCall (opNameCs op) (map ppExp' args)

   ppOp = ppOperator (opInfo op)
   ppExp' = ppOperator (0,NAssoc) . ppExp


opNameCs :: Op -> String
opNameCs Dp3 = vmath "dot"
opNameCs Dp4 = vmath "dot"
opNameCs Cos = math  "Cos"
opNameCs Sin = math  "Sin"
opNameCs Log = math  "LogTwo"  -- Needs log_2 x.  Revisit
opNameCs Exp = math  "ExpTwo"  -- Needs 2^x.  Revisit
opNameCs op  = opName op

math, vmath :: String -> String
math  = ("Math."  ++)
vmath = ("VMath." ++)

ppSwiz [i] e = ppOperator (apPrec,NAssoc) $
                ppExp e <> char '.' <> char ("XYZW" !! i)

-- OOPS!! This def makes repetitions of e.
ppSwiz ixs e = ppNewVec [ppSwiz [i] e | i <- ixs]

-- ppSwiz ixs e = ppOperator (apPrec,NAssoc) $
--                (ppExp e <> text (showSwizzle False ixs))
-- 
-- ppSwiz' ixs e = ppOperator (apPrec,RAssoc) $
--                 text (swizF ixs) <+> onRight (ppExp e)
-- 
-- swizF ixs = "swiz" ++ map (assocJ (zip [0..] "XYZW")) ixs

-- ppSwiz ixs e = ppOperator (apPrec+1,LAssoc) $
--                ppExp e <> (text ('.' : ixls))
--  where
--    ixls = map (assocJ (zip [0..] "xyzw")) ixs

assocJ keyVals key = fromJust (lookup key keyVals)

ppBinOp :: String -> Exp -> Exp -> PPDoc
ppBinOp op e1 e2 = sep [ppExp e1, text op, onRight (ppExp e2)]


apPrec :: Int
apPrec  = 10

-- operator precedence levels
opPrec :: Op -> Maybe Int
opPrec Mul = Just 7
opPrec Add = Just 6
opPrec Slt = Just 4
opPrec Sge = Just 4
opPrec _   = Nothing

opAssoc :: Op -> Assoc
opAssoc op = LAssoc  -- Currently all left-associative

opInfo :: Op -> (Int, Assoc)
opInfo op = (fromMaybe 0 (opPrec op), opAssoc op)
