-- Simple expression type See also Exp, which provides statically typed
-- wrappers, and SmartDExp, which contains optimizing "smart
-- constructors".

module DExp where

import Maybe (fromMaybe)
import List (lookup)

import Char (toLower)
import Types

data Op = Mov | Add | Mad | Max | Min | Mul | Sge | Slt  -- simd
        | Rcp | Rsq | Log | Exp                          -- scalar
        | Dp3 | Dp4                                      -- vector -> scalar
        | Expp | Logp | Frc
        -- Negation and swizzling are integrated into each generated
        -- instruction but are logically separate at this level.
        | Negate | Swizzle Swizzle | MkVec
        | Frac
        | Cos | Sin  -- See note about approximation
        | Out [Int]  -- output with register #s
  deriving (Eq,Ord,Show)

-- "Mov" is the identity operator and sometimes gets introduced during
-- normalization for code generation (NormalExp).

-- Sin & Cos are introduced but replaced later by approximations.  It's
-- important to extract derivatives before approximation rather than
-- after.  Instead of having operators like Sin & Cos, we could have a
-- power series operator containing the infinite sequence of coefficients.
-- More elegant & general but harder to compare head-to-head with Brian's
-- approach, and less amenable to generating host code if we choose to.


-- Name of operator, for assembly code.  Just change first char of
-- constructor to lower case.
opName :: Op -> String
opName = (\ (c:cs) -> toLower c : cs) . show

data Exp = LitVec Vector
         | Var Id Type
         | Apply Op [Exp]
         | Let [(Id, Exp)] Exp
         | Undefined
  deriving (Eq,Show)


-- Often convenient than Apply
apply1 :: Op -> Exp                      -> Exp
apply2 :: Op -> Exp -> Exp               -> Exp
apply3 :: Op -> Exp -> Exp -> Exp        -> Exp
apply4 :: Op -> Exp -> Exp -> Exp -> Exp -> Exp

apply1 = appL1 . Apply
apply2 = appL2 . Apply
apply3 = appL3 . Apply
apply4 = appL4 . Apply

-- Construct a Let
mkLet :: [(Id,Exp)] -> Exp -> Exp
mkLet [] = id
mkLet bs = Let bs


---- Typing

typeOf :: Exp -> Type

typeOf (LitVec vals)              = Vec (length vals)
typeOf (Var _ ty)                 = ty
typeOf (Apply MkVec exps)
  | all ((== scalar).typeOf) exps = size2type (length exps)

typeOf e@(Apply MkVec exps)       = error $ "typeOf: " ++
                                            show (e,map typeOf exps)
typeOf (Apply Negate [e])         = typeOf e
typeOf (Apply (Swizzle swiz) _)   = size2type (length swiz)
typeOf (Apply op args)
  | op `elem` scalarResultOps     = scalar
  | op `elem` simdOps             = typeOf (head args)
typeOf (Apply Expp _)             = Vec 4
typeOf (Apply (Out _) _)          = Output
typeOf (Let binds body)           = typeOf body

simdOps = [Mov,Add,Mad,Max,Min,Mul,Sge,Slt,Negate]
scalarResultOps = [Rcp,Rsq,Log,Exp,Dp3,Dp4,Sin,Cos,Frac]

-- TODO: add consistency checks to typeOf



---- Misc

applyNS :: NegSwiz -> Type -> (Exp -> Exp)
applyNS (NegSwiz neg swiz) ty = ifF neg (apply1 Negate) .
                                ifF (swiz /= idSwizzle ty) (apply1 (Swizzle swiz))
