-- Symbolic derivatives of expressions untyped & typed and of typed
-- functions.

module Derivative ( DDeriv,dderiv
                  , Derivative,derivative
                  ) where

import Types
import DExp
import SmartDExp
import qualified TExp as T
import TExp (FloatE)
import Memo
import Namer

-- Compute derivative expression.  To be used before CSE, because
-- differentiation often generates common sub-expressions.  For that
-- reason, differentiation is memoized and doesn't handle Let
-- expressions.

ederiv :: Id -> Exp -> Exp
ederiv var exp = d exp
 where
   d = memo nd   -- nd is the non-memoized d

   nd e@(LitVec _)               = zero (typeOf e)
   nd (Var var' ty)              = (if var == var' then one else zero) ty
   nd (Apply Add [u,v])          = d u + d v
   nd (Apply Mul [u,v])          = u * d v + v * d u
   nd (Apply Rcp [v])            = - d v / (v^2)
   nd (Apply Sin [u])            = cos u * d u
   nd (Apply Cos [u])            = - sin u * d u
   nd (Apply Rsq [u])            = - d u * rsqrt u / (twoF * u)
   nd e@(Apply Exp [u])          = e * d u * logTwoF
   nd (Apply Log [u])            = recip u * logTwoF * d u
   nd (Apply Negate [u])         = negate (d u)
   nd (Apply MkVec es)           = vecL (map d es)
   nd (Apply (Swizzle swiz) [u]) = swizzle swiz (d u)
   nd (Apply Frac [u])           = d u
   nd (Apply Dp3 [u,v])          = dp3 u (d v) + dp3 v (d u)
   nd (Apply Dp4 [u,v])          = dp4 u (d v) + dp4 v (d u)
   nd (Apply Slt [u,v])          = zero (typeOf u)
   nd (Apply Sge [u,v])          = zero (typeOf u)
   nd (Apply Max [u,v])          = ifE (u >* v) (d u) (d v)
   nd (Apply Min [u,v])          = ifE (u <* v) (d u) (d v)
   nd u                          = noDeriv u

   noDeriv u = error ("ederiv: unhandled expression: " ++ show u )

floatE :: Float -> Exp
floatE x = LitVec [x]

zeroF, twoF, logTwoF :: Exp
[zeroF, twoF,logTwoF] = map floatE [0, 2, log 2]


{- Rsq case:

d (1 / sqrt u) == d (u ** (-1/2))
               == -1/2 * (u ** (-3/2)) * d u
               == - 1/2 * (u ** (-1/2)) * (u ** (-1)) * d u
               == - d u * (1 / sqrt u) / (2 * u)

-}

-- Statically typed version.  Generalized to work on tuples, etc.

class DDeriv a where
  dderiv :: Id -> a -> a

-- We're playing loose here: the derivative of a point-valued function
-- should really be a vector-valued function.  Could generalize with
-- MPTC, but for now points and vectors are both just pairs.

instance Floating a => DDeriv (T.Exp a) where
  dderiv var = T.typ1 (ederiv var)

-- Differentiate tuples
instance (DDeriv a, DDeriv b) => DDeriv (a,b) where
  dderiv var (a,b) = (dderiv var a, dderiv var b)

instance (DDeriv a, DDeriv b, DDeriv c) => DDeriv (a,b,c) where
  dderiv var (a,b,c) = (dderiv var a, dderiv var b, dderiv var c)

-- Differentiate *w.r.t.* a pair (i.e., partial derivatives)
-- dderivPair :: DDeriv a => (Id,Id) -> a -> (a,a)
-- dderivPair (x,y) u = (dderiv x u, dderiv y u)


---- (Partial) derivatives of *functions*.

-- lmapAB represents a linear map from a to b.
class Derivative a b lmapAB | a b -> lmapAB where
 derivative :: (a -> b) -> (a -> lmapAB)

instance (DDeriv b, Substable b) => Derivative FloatE b b where
  derivative f = subD "u"
    where body = f (T.var "u")
          subD w = \ (T.E a) -> subst [("u",a)] d
                   where d = dderiv w body

instance (DDeriv b, Substable b) =>
         Derivative (FloatE,FloatE) b (b,b) where
  derivative f = \ z -> (subD "u" z, subD "v" z)
    where body = f (T.var "u", T.var "v")
          subD w = \ (T.E a,T.E b) -> subst [("u",a),("v",b)] d
                   where d = dderiv w body

-- and so on for higher dimensional domain types

-- TODO: Perhaps generalize from FloatE above.  If so, use a Namer.
