-- Some experiments in an embedded shader language.  Patterned after the
-- concepts in RenderMan's shader language.

module EPShader where

import TExp
import Vector
import Function
import Namer


type Dir3E = Vec3E  -- Normalized direction vector

-- Info about how one light affects a given point.  The surface shader
-- decides what to do with the light info.  Attenuation and relation of
-- light position (if finitely distant) to surface position are already
-- accounted for.
type LightInfo = (ColorE,Dir3E) -- Color and direction

-- A light is something that provides light info to every point in space
-- (though to some points it provides blackness), independent of
-- obstructions.  Should probably also take an atmosphere argument.
type Light = Vec3E -> LightInfo

-- Viewing environment.
type View = (ColorE, Vec3E, [Light]) -- ambient, eye, lights

-- Info about a surface at a point: position, tangent, binnormal, color
type SurfPt = (Vec3E, Dir3E, Dir3E, ColorE)

-- Shaders are simple functions from contextual info to values.
type Shader  a = (View, SurfPt) -> a
type LShader a = LightInfo -> Shader a

-- Extractors

ca     :: Shader  ColorE ; ca     ((c,_,_),_) = c
eye    :: Shader  Vec3E  ; eye    ((_,e,_),_) = e
lights :: Shader [Light] ; lights ((_,_,l),_) = l

pobj :: Shader  Vec3E  ; pobj (_,(p,_,_,_)) = p
dpdu :: Shader  Dir3E  ; dpdu (_,(_,t,_,_)) = t
dpdv :: Shader  Dir3E  ; dpdv (_,(_,_,b,_)) = b
cs   :: Shader  ColorE ; cs   (_,(_,_,_,c)) = c

cl :: LShader ColorE ; cl (c,_) _ = c
l  :: LShader Vec3E  ; l  (_,d) _ = d

illuminance :: (Num a, VectorOf FloatE a) => LShader a -> Shader a
illuminance lshader vs@((_,_,ls),(p,_,_,_)) =
   sum [ ({-bsign (lift n <.> l >* 0) *^-} lshader) (light p) vs | light <- ls ]

-- Does this bsign multiplier really help?  The vertex engine clamps to
-- [0,1] anyway.  Oh, with multiple light sources, negative contributions
-- would subtract from positive ones.

---- Now some shaders

-- Just the intrinsic color, ignoring all lighting
intrinsic :: Shader ColorE
intrinsic = cs

-- Pure ambient (ignoring intrinsic color)
ambient :: Shader ColorE
ambient = ca

-- Normal vector
n :: Shader Dir3E
n = cross (pairF dpdu dpdv)

-- Pure diffuse
diffuse :: Shader ColorE
diffuse = illuminance ((lift n <.> l) *^ cl)

-- Weighted combination of ambient and diffuse
ambDiff :: (ColorE, ColorE) -> Shader ColorE
ambDiff (ka,kd) = intrinsic * (lift ka * ambient + lift kd * diffuse)

-- The Stanford rtsl version, with ambient and weights:
-- 
-- surface float4
-- lightmodel_diffuse (float4 ka, float4 kd)
-- {
--     perlight float diffuse = dot(N,L);
--     perlight float4 fr = kd * select(diffuse > 0, diffuse, 0);
--     return ka * Ca + integrate(fr * Cl);
-- }

-- Direction from surface point to eye
eyeDir :: Shader Dir3E 
eyeDir = normalize (eye - pobj)

-- Eye/light vector average (CGPP p 731)
h :: LShader Dir3E
h = normalize (l + lift eyeDir)

-- Reflection vector (CGPP p 730)
refl :: LShader Dir3E
refl = (2 * (n' <.> l)) *^ n' - l  where n' = lift n

-- Pure specular.  Ignores intrinsic surface color.  There are different
-- ways to compute the power base.
specularG :: LShader FloatE -> FloatE -> Shader ColorE
specularG base sh = illuminance ((max 0 base ** lift sh) *^ cl)

{-
-- Phong's specular model
specularPh :: FloatE -> Shader ColorE
specularPh = specularG (refl <.> lift eyeDir)

-- Or the N.H model:
specularNH :: FloatE -> Shader ColorE
specularNH = specularG (lift n <.> h)
-}

-- surface floatv
-- lightmodel_specular (floatv s, floatv e, float sh)
-- {
--     perlight float diffuse = dot(N,L);
--     perlight float specular = pow(max(dot(N,H),0),sh);
--     perlight floatv fr = select(diffuse > 0, s * specular, Zero);
--     return integrate(fr * Cl) + e;
-- }

type BasicSh = (ColorE,ColorE,ColorE,FloatE) -> Shader ColorE

-- Combine intrinsic, ambient, diffuse and specular, with weightings
basic :: LShader FloatE -> BasicSh
basic base (ka,kd,ks,sh) = ambDiff (ka,kd) + lift ks * specularG base sh

basicRV, basicNH :: BasicSh
basicRV = basic (refl <.> lift eyeDir)  -- The R.V model
basicNH = basic (lift n <.> h)          -- The N.H model

-- surface float4
-- lightmodel (float4 a, float4 d, float4 s, float4 e, float sh)
-- {
--     perlight float diffuse = dot(N,L);
--     perlight float specular = pow(max(dot(N,H),0),sh);
--     perlight float4 fr = d * max(diffuse, 0) +
--                          s * select(diffuse > 0, specular, 0);
--     return a * Ca + integrate(fr * Cl) + e;
-- }


-- Anisotropic lighting example.  See section 6.2 of "A
-- User-Programmable Vertex Engine" in Siggraph 01, adopted from reference
-- [12] in that paper, by Heidrich & Seidel.

anisotropic1 :: (Shader Vec3E -> Shader Vec3E -> Shader Vec3E) -> BasicSh
anisotropic1 tang (ka,kd,ks,sh) = lift ka * cs + illuminance (perlight * cl)
 where
   perlight = lift kd ^* ltsq + lift ks ^* spec ** lift sh
   spec     = ltsq * cosToSin edotT - ldotT * edotT
   t        = tang dpdu dpdv

   ltsq, ldotT, edotT :: LShader FloatE
   ltsq  = cosToSin ldotT
   ldotT = l <.> lift t
   edotT = lift (eyeDir <.> t)

   cosToSin :: Floating c => c -> c
   cosToSin c = sqrt (1 - c^2)

-- surface float4
-- lightmodel_anisotropic_u (float4 a, float4 d, float4 s, float4 e, float sh)
-- {
--     float EdotT = dot(E,T);
--     perlight float LdotT = dot(L,T);
--     perlight float diff = sqrt(1 - LdotT * LdotT);
--     perlight float spec = max(diff * sqrt(1 - EdotT*EdotT) - LdotT*EdotT, 0);
--     perlight float4 fr = max(dot(N,L),0) * (d * diff + s * pow(spec,sh));
--     return a * Ca + integrate(fr * Cl) + e;
-- }


-- Translated:
anisotropic2 :: (Shader Vec3E -> Shader Vec3E -> Shader Vec3E) -> BasicSh
anisotropic2 tang (ka,kd,ks,sh) = lift ka * cs + illuminance (fr * cl)
 where
  ldotT, edotT, diff :: LShader FloatE
  edotT = lift (eyeDir <.> t)
  ldotT = l <.> lift t
  diff  = sqrt (1 - ldotT * ldotT)

  t     = tang dpdu dpdv

  spec :: LShader FloatE
  spec  = max (diff * sqrt(1 - edotT*edotT) - ldotT*edotT) 0
  fr :: LShader ColorE
  fr    = {- max (lift n <.> l) 0 *^-} (lift kd ^* diff + lift ks ^* spec ** lift sh)


-- TODO: get working; use intrinsic; factor common stuff with Phong
-- lighting.

-- TODO: atmospheric attenuation

-- Some refactoring to be emulate basic.
anisotropic3 :: (Shader Vec3E -> Shader Vec3E -> Shader Vec3E) -> BasicSh
anisotropic3 tang (ka,kd,ks,sh) =
  intrinsic * (lift ka * ambient + lift kd ^* illuminance ldotN)
  + lift ks ^* illuminance (vdotR ** lift sh)
 where
  ldotN, vdotR :: LShader FloatE
  ldotN = sqrt (1 - ldotT^2)
  vdotR = ldotT * sqrt (1 - vdotT^2) - ldotT * vdotT

  ldotT, vdotT :: LShader FloatE
  vdotT = lift (eyeDir <.> t)
  ldotT = l <.> lift t

  t :: Shader Vec3E
  t = tang dpdu dpdv

-- Factoring of basic to accomodate isotropic & anisotropic

{-
-- Phong model, taking normal as parameter
-- WORKING HERE.  Coefficients not yet right.

phongN :: LShader DirE -> Shader ColorE
phongN n = 
  cs * lift ka * ambient + cs * lift kd ^* illuminance ldotN
  + lift ks ^* illuminance (vdotR ** lift sh)
-}