-- Extract literals from an expression, producing a literal-free
-- expression and a constant register file.

-- Context: The assembly language has no literals, so instead, the
-- would-be literals go into a register file.  We build that file with as
-- much sharing as possible while still having each literal fit into a
-- single register.  At a last resort, we may break literals into pieces
-- of multiple registers.  (Not done yet.)

-- TODO: Introduce variables and build up an environment mapping them to
-- swizzled constant registers, to make the code a little friendlier.


module ExtractLiterals ( extractLiterals
                       , constVarName, constVar, asConstVar
                       ) where

import List (elemIndex,nub,union,intersect,isPrefixOf)
import Maybe (fromMaybe)
import Monad (mplus,msum)
import Types
import DExp

-- The values contained in a register file.  Each register may hold up to
-- four floats.  Fewer than four means that there are still some float
-- slots free.
type RegFile = [Vector]

-- The main function
extractLiterals :: Int -> Exp -> (Exp,RegFile)

-- Proceed in three phases: *discover* the literals, *pack* into a constant
-- register file, and *replace* the literals with variables (possibly
-- swizzled and negated).

extractLiterals numRegs exp = (replace regs exp, regs)
 where
   regs = pack numRegs (discover exp)

-- TODO: Consider not providing numRegs argument, and use as much as we
-- need.  Pro: when there aren't enough constant registers, we can tell
-- the author how many were needed, so s/he can try simplifying.  Con: we
-- won't know when to try last-resort measures like breaking literals
-- across registers.


---- Discover the literals

discover :: Exp -> [Vector]
discover exp = disc exp []
 where
   disc :: Exp -> [Vector] -> [Vector]
   disc (LitVec vec)     sofar = vec : sofar
   disc (Apply _ exps)   sofar = foldr disc sofar exps
   disc (Let binds body) sofar = disc body (foldr disc sofar
                                                  (map snd binds))
   disc _                sofar = sofar



---- Pack vectors into a constant register file

-- Build a constant register file that accommodates the literals, doing as
-- much sharing as possible.  Trim (nub) each literal so it contains no
-- repeated float.  Negate if all non-positive, to get more sharing.  Add
-- trimmed vectors in order of decreasing size, sharing as much as
-- possible, while still fitting.

pack :: Int -> [Vector] -> RegFile
pack numRegs vecs = foldr addVec
                          (initRegs numRegs)
                          (sortF (negate.length) (map (mbneg.nub) vecs))
 where
   mbneg vec | all (<= 0) vec = map negate vec
             | otherwise      = vec

-- TODO: Oh -- I didn't do the most general thing here.  I might be able
-- to fit the negated vector into an existing register, but now I only
-- negate if all of the components are non-negative.

-- TODO: avoid repeated computation of keys during sorting in pack.  Could
-- fix in sortF, but I think it's better to keep sortF simple and define
-- the key-caching version on top of sortF.  (Hint: pair keys with vals
-- and sortF with fst.)

-- TODO: break literals into pieces of multiple registers if necessary.

-- Initially empty register file, given max size.
initRegs :: Int -> RegFile
initRegs numRegs = take numRegs (repeat [])

-- Add a single vector.  Use the register that already shares the most
-- floats with the given vector while having enough free space to add the
-- unshared floats.
addVec :: Vector -> RegFile -> RegFile
addVec vec regs = update (best regs)
 where
   -- Index of most suitable register and its contents
   best :: RegFile -> (Int,Vector)
   best = hd' .                            -- take best
          filter ((<= 4).length.snd) .     -- eliminate register overflows
          map (onSnd (union vec)) .        -- merge vec with each
          sortF (onSnd (negate.overlap)) . -- sort by descending overlap
          zip [0..]                        -- tag with reg index

   overlap = length . intersect vec

   hd' []    = error "addVec: couldn't add vector"  -- bummer :(
   hd' (a:_) = a

   update :: (Int,Vector) -> RegFile
   update (ix,newVec) = befores ++ [newVec] ++ afters
     where (befores,(_:afters)) = splitAt ix regs


-- Question: is the sort algorithm friendly for lazy use in this way?
-- That is, if I only want to examine a few values near the beginning,
-- does it leave most of the tail-end sorting work undone?  I think
-- quicksort and insertion sort are friendly in this way.  Hugs uses
-- insertion sort.


---- Replace the literals with variables (possibly swizzled and negated)

-- Given a constant register file and an expression, replace the literals
-- with references to variables denoting constant registers, possibly
-- negated and possibly swizzled.

replace :: RegFile -> Exp -> Exp
replace regs = repl
 where
   repl (LitVec vec)     = fitReg vec regs
   repl (Apply op exps)  = Apply op (map repl exps)
   repl (Let binds body) = Let (zip vars (map repl exps)) (repl body)
                           where (vars,exps) = unzip binds
   repl exp              = exp

-- Find a register that can accommodate the given vector.  Produce an
-- expression that extracts the vector using a variable that denotes the
-- constant register, possibly using swizzling and/or negation.
fitReg :: Vector -> RegFile -> Exp
fitReg vec = mkExp               .   -- build exp
             fromMaybe err       .   -- make sure we found one
             msum                .   -- take first success, if any
             zipWith tagMb [0..] .   -- remember reg #
             map (tryNS vec)         -- try each reg
 where
   err = error "fitReg: Couldn't fit!"  -- shouldn't happen

   -- Try to realize the vector as a negswiz of the register
   tryNS :: Vector -> Vector -> Maybe NegSwiz
   tryNS vec reg = try False id `mplus` try True (map negate)
    where
      try neg f = fmap (NegSwiz neg) $
                  sequence (map (flip elemIndex reg) (f vec))

   tagMb :: a -> Maybe b -> Maybe (a,b)
   tagMb a = fmap (\ b -> (a,b))  -- or fmap . (,)

   mkExp :: (Int,NegSwiz) -> Exp
   mkExp (regNum,ns) = applyNS ns (Vec 4) (constVar regNum)

constVarPrefix = "c_"

constVarName :: Int -> String
constVarName regNum = constVarPrefix ++ show regNum

constVar :: Int -> Exp
constVar regNum = Var (constVarName regNum) (Vec 4)

asConstVar :: Exp -> Maybe Int
asConstVar (Var name _)
  | isConst = Just n
 where isConst = isPrefixOf constVarPrefix name
       n       = read (drop (length constVarPrefix) name)

