-- Some experiments in an embedded shader language.  Patterned after the
-- concepts in RenderMan's shader language.  This version uses implicit
-- parameters, nested in a way that requires rank 2 polymorphism, so it
-- requires GHC in version at least 5.03.

module Shader where

import Vertigo

-- Info about how one light affects a given point.  The surface shader
-- decides what to do with the light info.  Attenuation and relation of
-- light position (if finitely distant) to surface position are already
-- accounted for.
data LightInfo = LightInfo ColorE Vec3E -- ColorE & normalized direction

-- A light is something that provides light info to every point in space
-- (though to some points it provides blackness), independent of
-- obstructions.  Should probably also take an atmosphere argument.
type Light = Point3E -> LightInfo

-- Value depending on view parameters: ambient ColorE, eye location, and
-- collection of lights.
type ViewDep a = (?ca :: ColorE, ?eye :: Point3E, ?lights :: [Light]) => a

-- Dependent on surface point parameters: ColorE, position, and normal:
type SurfDep a = (?cs :: ColorE, ?pobj :: Point3E, ?n :: Vec3E) => a

-- Dependent on light info: ColorE and normalized direction vector
type LightDep a = (?cl :: ColorE, ?l :: Vec3E) => a

-- "Surface shader".
type Shader  a = ViewDep (SurfDep a)

-- "Light shader" (perhaps a misnomer) adds in light info dependence
type LShader a = Shader (LightDep a)

-- We'll want to sum per-light shading results.  illuminance gets its
-- lights and surface point implicitly, and binds ?linfo iteratively.
-- Discard lights when the surface faces away from them.
illuminance :: LShader ColorE -> Shader ColorE
illuminance ia = sum [ let LightInfo cl l = light pobj in
                         ifForward ia with {?cl = cl; ?l = l}
                     | light <- ?lights
                     ]
-- Use given ColorE if the surface is roughly facing the light
ifForward :: ColorE -> LShader ColorE
ifForward col | n <.> l > 0 = col
              | otherwise = black

-- Some helpful computations.  Note how easily we can make up new ones for
-- convenience.  They don't need to be built in.

-- Eye/light vector average
h :: LShader Vec3E
h = (norm (eye - pobj) + l) ^/ 2

-- As formulated, there's a constant syntactic reminder of which shaders
-- are simply accessors, like "cs", and which are utilities, like "h".
-- (Like most OO languages syntactically distinguish between methods and
-- utility functions.)  If we want to hide this distinction, we can define
-- a bunch of functions like the following:

-- To hide the syntactic evidence of between implicit parameters &
-- utilities (e.g., "n" vs "h"), use the following:

ca   :: ViewDep  ColorE  ; ca   = ?ca
eye  :: ViewDep  Point3E ; eye  = ?eye
cs   :: SurfDep  ColorE  ; cs   = ?cs

pobj :: SurfDep  Point3E ; pobj = ?pobj
n    :: SurfDep  Vec3E   ; n    = ?n

cl   :: LightDep ColorE  ; cl   = ?cl
l    :: LightDep Vec3E   ; l    = ?l


---- Misc

type Point3E = Vec3E


---- Sample shaders

-- The simplest two surface shaders are pure intrinsic and pure ambient.
-- They're just the accessors "cs" and "ca".

-- Pure diffuse shader.  Ignores surface color.
diffuse :: Shader ColorE
diffuse = illuminance ((n <.> l) *^ cl)

-- Pure specular.  Ignores cs surface color.  (Do I still need to test
-- sign of n<.>l?)
specular :: FloatE -> Shader ColorE
specular sh = illuminance ((max 0 (n <.> h) ** sh) *^ cl)

-- Weighted combination of ambient & diffuse
ambDiff :: FloatE -> FloatE -> Shader ColorE
ambDiff ka kd = ka *^ ca + kd *^ diffuse

-- Combine ambient, intrinsic, diffuse & specular
basic :: FloatE -> FloatE -> FloatE -> FloatE -> Shader ColorE
basic ka kd ks sh = ambDiff ka kd * cs + ks *^ specular sh

