-- This module uses memoized copying functions to maximize sharing in a
-- tree/dag.  It works under Hugs but not GHCi 5.03.

module MemoTest where

import IOExts (trace)
import Memo

data Tree = Node Int [Tree] deriving Show

-- Strict memo function
smemo f = ((memo f) $!)

-- Simple tree mapper.
mapTree :: (Int -> Int) -> Tree -> Tree
mapTree f = mapT
 where
   mapT (Node a trees) = Node (f a) (map mapT trees)

-- Memoizing version
mapTree' :: (Int -> Int) -> Tree -> Tree
mapTree' f = mapT'
 where
   mapT' = smemo mapT
   mapT (Node a trees) = Node (f a) (map mapT' trees)

-- Maximize sharing by performing recursive memoized copying.  Replace
-- equal expressions with pointer-equal ones.  Since I'm using memo
-- functions here, it's vital not to eta-contract uniquifyTree.
-- Otherwise, the memo tables would be CAFs and so not get collected.
uniquifyTree :: Tree -> Tree
uniquifyTree tree = length (show tree) `seq` 
                    uTree tree
 where
  uTree = smemo copyTree
  uList = smemo copyList
  uCons = smemo2 (:)
  uNode = -- smemoInt (smemo . Node)
          smemo2 Node

  copyTree (Node label trees) = uNode label (uList (map uTree trees))

  copyList []    = nilTree
  copyList (h:t) = uCons h (uList t)

  smemo2 :: (Show a, Show b) => (a -> b -> c) -> (a -> b -> c)
  smemo2 f = smemo (smemo  . f)

nilTree :: [Tree]
nilTree = []


---- Int-based memoization

-- smemoInt :: (Int -> a) -> (Int -> a)
-- smemoInt = mkMemo (\ a b -> return (a==b)) return 40

---- Examples

[n1,n2,n3,n4] = [1..4] :: [Int]

-- These ones have a lot of sharing
t1 = Node n1 []      :: Tree
t2 = Node n2 [t1,t1]
t3 = Node n3 [t2,t2]
t4 = Node n4 [t3,t3]

-- These ones have no sharing
t1'() = Node n1 []      :: Tree
t2'() = Node n2 [t1'(),t1'()]
t3'() = Node n3 [t2'(),t2'()]
t4'() = Node n4 [t3'(),t3'()]

-- Simple tracing increment function.
incr :: Int -> Int
incr n = n `seq` trace ("(incr " ++ show n ++ ")") (n+1)

t2'' = uniquifyTree (t2'())

m1() = mapTree' incr t2
m2() = mapTree' incr (t2'())
m3() = mapTree' incr (uniquifyTree (t2'()))
m4() = mapTree' incr t2''

{- Results

As expected mapTree' finds memo table hits with shared trees and no hits
with unshared trees.  For instance, here's ghci output:

  MemoTest> mapTree' incr t2
  (incr 2)
  (incr 1)
  Node 3 [Node 2 [],Node 2 []]

  MemoTest> mapTree' incr (t2'())
  (incr 2)
  (incr 1)
  (incr 1)
  Node 3 [Node 2 [],Node 2 []]

Now use unshared trees and apply uniquifyTree.  Hugs gets the memo table
hits:

  MemoTest> mapTree' incr (uniquifyTree (t2'()))
  Node (incr 2)3 [Node (incr 1)2 [],Node 2 []]

But ghci 5.03 doesn't:

  MemoTest> mapTree' incr (uniquifyTree (t2'()))
  (incr 2)
  (incr 1)
  (incr 1)
  Node 3 [Node 2 [],Node 2 []]

Strangely, under ghc, uniquifyTree even loses the sharing already
present.

  MemoTest> mapTree' incr (uniquifyTree t2)
  (incr 2)
  (incr 1)
  (incr 1)
  Node 3 [Node 2 [],Node 2 []]

-}