-- Procedural geometry with computed normals

-- I'm tinkering with using host language tuples instead of Vertigo vector
-- types here.  Simplifies notation.

module NormalGeom where

import Vertigo
import qualified NumOps
import Function
import Lighting 
import NameStruct

type R  = FloatE
type R1 = R
type R2 = (FloatE,FloatE)
type R3 = (FloatE,FloatE,FloatE)

type Image a     = R2 -> a
type Texture     = Image ColorE
type Region      = Image BoolE
type HeightField = Image R
type Surf        = Image R3

type Curve2 = R1 -> R2
type Curve3 = R1 -> R3

type Transform1 = R1 -> R1
type Transform2 = R2 -> R2
type Transform3 = R3 -> R3

mul2pi :: R -> R
mul2pi = (* (2*pi))

-- "Normalized" trig functions with unit period.
cosNU, sinNU, cosU, sinU :: R -> R
cosNU = cosN . mul2pi
sinNU = sinN . mul2pi
cosU  = cos  . mul2pi
sinU  = sin  . mul2pi

normal :: (Derivative c vec vecs, Cross vecs vec, VectorOf s vec)
       => (c -> vec) -> (c -> vec)

normal = normalize . cross . derivative

-- Inference comes up with a slightly more general type:
-- normal :: (Derivative c vec vecs, Cross vecs vec', VectorOf s vec')
--        => (c -> vec) -> (c -> vec')

-- Apply a displacement map.  For instance, v is Surf or Curve2.
displace v s = v + s *^ normal v

-- This guy has a pretty hairy type (inferred):
displace :: ( Num (a -> vec), Cross vecs vec, Derivative a vec vecs
            , VectorOf d vec, VectorOf (a -> s) (a -> vec) )
         => (a -> vec) -> (a -> s) -> (a -> vec)



---- Examples

hfSurf :: HeightField -> Surf
hfSurf field = \ (u,v) -> (u, v, field (u,v))

-- Similarly for curve construction
fcurve :: Transform1 -> Curve2
fcurve f = \ u -> (u, f u)


rippleH :: HeightField
rippleH = sinU . magnitude

-- Alter frequency and magnitude of a periodic function.  More generally,
-- domain-compress and range-magnify a function.
freqMag :: (VectorOf si vi, VectorOf so vo)
        => (vi->vo) -> (si,so) -> (vi->vo)
freqMag f (freq,mag) = (mag *^) . f . (freq *^)

wavy :: (R,R) -> Transform1
wavy = freqMag sinU

rippleH' :: (R,R) -> HeightField
rippleH' = freqMag rippleH

-- rippleH' (freq,mag) = (mag *) . rippleH . (freq *^)

ripple :: Surf
ripple = onXY' (3.5 *^) (hfSurf rippleH)

ripple' :: (R,R) -> Surf
ripple' frm = onXY' (3.5 *^) (hfSurf (rippleH' frm))
                   -- (mag *^) . ripple . (freq *^)

-- Unit circle.  For efficiency, it comes in two flavors: one with domain
-- [-1/2,1/2], and one with arbitrary domain.
-- TODO: Use some simple interval analysis to determine which cos & sin to
-- use.
circleN, circle :: Curve2
circleN = cosNU `pairF` sinNU

circle = circleN . wrap (-1/2,1/2)

-- Torus given radius of sweep circle and cross section
torus :: R -> R -> Surf
torus sr cr = revolve (const (sr,0) + const cr *^ circle)

-- Surface of revolution, formed by rotation around Z axis.  The curve is
-- parameterized by u, and the rotation by v.  In this generalized
-- version, we have not a single curve, but a function from v to curves.
revolveG :: (R -> Curve2) -> Surf
revolveG curveF = \ (u,v) -> onXY (rotate (-2*pi*v)) (addY (curveF v) u)

revolve :: Curve2 -> Surf
revolve curve = revolveG (const curve)


-- A sphere is a revolved semi-circle
sphere = revolve semiCircle

-- Shrink domain to [-1/4,1/4], corresponding to theta in [-pi/2,pi/2]
semiCircle = circleN . (/ 2)

cylinder :: R -> Surf
cylinder = cylinder1

-- A cylinder is a revolved line
cylinder1 h = onZ (h *) . revolve (\ y -> (1,y))

-- or a swept circle
cylinder2 h = sweep (addXY (h *)) (addY circleN)

cylinder3 h (u,v) = (x,y,h*u) where (x,y) = circleN v

wavyCylinder :: (R,R) -> R -> Surf

-- revolve a wavy line
wavyCylinder1 frm h = revolve (\ u -> (1 + wavy frm u, h*u))

-- Or scale a cylinder
wavyCylinder2 frm h (u,v) = onXY ((1 + wavy frm u) *^) (cylinder h (u,v))

-- Expressed as a product of functions:

wavyCylinder frm h = scaleF * cylinder h
 where
   scaleF (u,v) = (s,s,1) where s = 1 + wavy frm u

-- More explicitly, for the presentation
{-
wavyCylinder (f,m) h = scaleF * cylinder h
 where
   scaleF (u,v) = (s,s,1) 
     where
       s = 1 + m * sin(2*pi*f*u)
-}

-- Torus given sweep circle radius and varying surface
torusG :: R -> (R -> Curve2) -> Surf
torusG sr curveF = revolveG (\ u -> const (sr,0) + curveF u)

wavyTorus frm sr cr =
  torusG sr (\ u -> const (cr * (1 + wavy frm u)) *^ circleN)

-- const (cr u) *^ circleN

torus1 sr cr = torusG sr (const cr)

torus2 = wavyTorus (0,0)


phase rate f = \ u -> f . (+ 2*pi*rate*u)

licorice (freq,mag,cr,twistRate) =
 phase twistRate $
 const cr *^ (
  displace circle $
  wavy (freq,mag))

helicalLicorice (hfreq,zrate,sr,cr,freq,mag,twistRate) =
 helicalSweep hfreq zrate $
 torusG sr $
 licorice (freq,mag,cr,twistRate)


-- Profile product
profile :: Curve2 -> Curve2 -> Surf
profile curve profile (u,v) = (cx*px,cy*px,py)
 where
   (cx,cy) = curve   u
   (px,py) = profile v

-- More spheres
sphere2 = profile circle semiCircle
sphere3 = profile semiCircle circle

-- John S's thesis points out that surfaces of revolution are profile
-- products with circular profiles.

-- Helix curve, given XY radius, Z rate
helix :: R -> R -> Curve3
helix r zrate = \ u -> onZ (+ zrate*u) (addZ (lift r *^ circle) u)

-- sweep scurve basis (u,v) = scurve u + basis v

-- Given a combining op and two curves, make a surface.  A sort of
-- Cartesian product with combination.
cartF :: (a -> b -> c) -> (u -> a) -> (v -> b) -> ((u,v) -> c)
cartF op f g = \ (u,v) -> f u `op` g v

-- Sweep a basis curve by a sweep curve
sweep :: Curve3 -> Curve3 -> Surf
sweep = cartF (+)

helicalTube1 :: R -> R -> R -> R -> Surf
helicalTube1 freq zrate r cr =  sweep (helix r zrate . (* freq))
                                      (addY (const cr *^ circle))

-- The definition above doesn't work.  I'd need to reorient the circular
-- cross section so its normal points in the direction of the helix's
-- derivative vector.  That's what sweep should do.  TODO: make it so.

-- This version builds on the torus instead.  It doesn't achieve a truly
-- circular cross section but is close.
helicalTube2 :: R -> R -> R -> R -> Surf
helicalTube2 freq zrate r cr (u,v) = onZ (+ zrate*v') (torus r cr (u,v'))
 where
   v' = v * freq

-- Factor out the helical sweep

helicalTube :: R -> R -> R -> R -> Surf
helicalTube freq zrate r cr = helicalSweep freq zrate (torus r cr)

helicalSweep freq zrate surf (u,v) = onZ (+ zrate*v') (surf (u,v'))
 where
   v' = v * freq


-- One period, unit height eggcrate
eggcrate :: HeightField
eggcrate = cartF (*) cosU sinU

-- We can tweak the frequency and height from the outside.
eggcrate' :: (R,R) -> HeightField
eggcrate' = freqMag eggcrate

grooveTube :: (R,R) -> R -> R -> Surf
grooveTube frm sr cr =
  displace (torus sr cr) (1 + wavy frm . fst)

helicalGrooveTube :: R -> R -> (R,R) -> R -> R -> Surf
helicalGrooveTube freq zrate (wiggles,mag) sr cr =
  helicalSweep freq zrate (grooveTube (wiggles,mag) sr cr)


-- Add phase to frequency and magnitude

-- Alter frequency, magnitude, and phase of a periodic function.  More
-- generally, domain -compress and -shift and range-magnify a function.
freqMagPhase :: (VectorOf si vi, VectorOf so vo, Num vi)
                => (vi->vo) -> ((si,so),vi) -> (vi->vo)
freqMagPhase f (fm,phase) = freqMag f fm . (phase +)

-- freqMagPhase f ((freq,mag),phase) = (mag *^) . f . (freq *^) . (phase +)

type FMP = ((R,R),R)

wavyP :: FMP -> Transform1
wavyP = freqMagPhase sinU

-- Another ripple.  This one is a revolved sine.  It's darker than ripple',
-- but it looks much smoother.  I think because the mesh follows the
-- shape.  Interesting.  The subtraction gets the normal going the right way.
rippleR :: FMP -> Surf
rippleR fmp = onXY' (2 *^) (
               revolve (const (0.5,0) - fcurve (wavyP fmp)))

---- Dimples

-- First a single hemispherical dimple with unit radius

dimple :: HeightField
dimple p = ifE (msq >=* 1) 0 (sqrt (1 - msq)) where msq = magnitudeSq p

-- Next, for lowering the dimple
cutoffHF :: R -> HeightField -> HeightField
cutoffHF minZ hf = max (hf - const minZ) 0

-- For tiling

class Tile a where tileF :: a -> a -> a

tile :: Tile a => a -> (a -> b) -> (a -> b)
tile wid = (. tileF wid)

-- Could generalize this next one if I loosen the type of wrap
instance Tile R where tileF width = wrap (-width/2,width/2)

instance (Tile a, Tile b) => Tile (a,b) where
  tileF (wa,wb) = cartF (,) (tileF wa) (tileF wb)

tiledDimples :: R -> R -> HeightField
tiledDimples minZ spacing = tile (spacing,spacing) (cutoffHF minZ dimple)

-- simpler, without cutoff
tiledDimples' spacing = tile (spacing,spacing) dimple


-- tileF (wa,wb) (a,b) = (tileF wa a, tileF wb b)

---- Misc

rotate :: R -> Transform2
rotate theta = \ (x,y) -> (x * c - y * s, y * c + x * s)
 where c = cos theta
       s = sin theta

addX, addY, addZ :: (a -> R2) -> (a -> R3)
addX = lift1 (\ (y,z) -> (0,y,z))
addY = lift1 (\ (x,z) -> (x,0,z))
addZ = lift1 (\ (x,y) -> (x,y,0))

addYZ,addXZ,addXY :: (a -> R1) -> (a -> R3)
addYZ = lift1 (\ x -> (x,0,0))
addXZ = lift1 (\ y -> (0,y,0))
addXY = lift1 (\ z -> (0,0,z))

onX,onY,onZ :: Transform1 -> Transform3
onX f (x,y,z) = (f x, y, z)
onY f (x,y,z) = (x, f y, z)
onZ f (x,y,z) = (x, y, f z)

onXY,onYZ,onXZ :: Transform2 -> Transform3
onXY f (x,y,z) = (x',y',z ) where (x',y') = f (x,y)
onXZ f (x,y,z) = (x',y ,z') where (x',z') = f (x,z)
onYZ f (x,y,z) = (x ,y',z') where (y',z') = f (y,z)

onX',onY',onZ' :: Transform1 -> (a -> R3) -> (a -> R3)
[onX',onY',onZ'] = map on' [onX,onY,onZ]
 where
   on' on f surf a = on f (surf a)

onXY',onXZ',onYZ' :: Transform2 -> (a -> R3) -> (a -> R3)
[onXY',onXZ',onYZ'] = map on' [onXY,onXZ,onYZ]
 where
   on' on f surf a = on f (surf a)

-- Twist about the X axis at a given rate of radians per x.
twist :: R -> Transform3
twist rate p@(x,_,_) = onYZ (rotate (rate*x)) p

-- Useful esp for height fields, to squeeze more into view while keeping
-- proportions.
squeezeF :: (VectorOf a b, VectorOf a c) => a -> (b -> c) -> (b -> c)
squeezeF sq f = (^/ sq) . f . (sq *^)

twister :: Surf -> Anim Surf
twister surf t = twist t . surf

---- Testing

count = NumOps.analyzeOps . unE

uv :: R2
uv@(u,v) = (var "u", var "v")

-- t :: R
-- t = var "t"

cseUV :: CSE a => a -> a
cseUV = cse ["u","v","t"]

testD :: Surf -> (R3,R3)
testD f = cseUV (derivative f uv)

testF :: Surf -> Vec3E
testF f = cseUV (fromCoords (f uv))

-- For testing, try testF (cylinder 2) & sphere, and with normal applied
-- to the same.  Thrown in "count" to measure work.

uncurry3 f (a,b,c)   = f a b c
uncurry4 f (a,b,c,d) = f a b c d


-- Procedural parametric surface.  Maps two-space to point, normal and
-- color.
type SurfVP = Vec2E -> SurfPt

{-
mkSurfVP' :: Texture -> Surf -> SurfVP
mkSurfVP' texture surf = \ p ->
  (cver surf p, cver norm p, texture (coords p))
 where
   norm   = normal surf
   cver s = fromCoords . s . coords
-}

mkSurfVP' :: Texture -> Surf -> SurfVP
mkSurfVP' texture surf p =
  ( fromCoords (surf p')
  , fromCN du, fromCN dv
  , texture p'
  )
 where
   p'      = coords p
   (du,dv) = derivative surf p'
   fromCN  = normalize . fromCoords


mkSurfVP :: ColorE -> Surf -> SurfVP
mkSurfVP = mkSurfVP' . const

type ColSurf = Vec2E -> (Vec3E, ColorE)

mkColSurf :: Shader ColorE -> View -> SurfVP -> ColSurf
mkColSurf shader view surf p = (p', shader (view,surfPt))
 where
   surfPt@(p',_,_,_) = surf p


cwrap' :: (AsRange nrtup ns, AsVar ns b)
       => String -> (Int,Int) -> Texture -> nrtup
       -> (b -> Shader ColorE) -> R3 -> (b -> Surf) -> IO ()

cwrap' name tess texture nrtup shaderF lightPos f = 
  compile name tess ("world","eyePos", vars) ("u","v") ranges $
  wrapVecDiffuse4 $ \ eyePos z ->
   (mkColSurf (shaderF z) (view eyePos) $ mkSurfVP' texture $ f z)
 where
   (vars, ranges) = asRange nrtup
   view :: Vec3E -> View
   view ep  = (white, ep, [light])
   light = dirLight white (normalize (fromCoords lightPos))
           -- pointLight white lightPos

defaultTess = (125,125) :: (Int,Int)

-- cwrap name col = cwrap' name defaultTess (const col)

-- Oddly, the version above causes type errors, but the version below doesn't

cwrap name = cwrap' name defaultTess . const

initEye :: R3
initEye = (-1,-4,2)  -- Coordinate with m_Eye in VShaderViewer.cs

-- Two useful lights

-- Use this one for roughly horizontal surfaces, e.g., from hfSurf.  The
-- light is past the surface, looking from the viewer's initial eye point
lightPast = onXY negate initEye

-- Use this one for roughly vertical surfaces.  The
-- light is past the surface, looking from the viewer's initial eye point
lightBack = onX negate initEye


basicC = const basicStd


adsp = ( Range "ambient"  0 1 0.1
       , Range "diffuse"  0 1 0.3
       , Range "specular" 0 1 0.6
       , Range "power"    0 100 15
       )

---- Testing.  Some (as marked) exceed instruction limit

cBasicG :: String -> BasicSh -> Surf -> IO ()
cBasicG name shaderF surf = 
  cwrap name green adsp shaderF' lightPast
  (const surf) -- (const sphere3)
 where
   shaderF' (a,d,s,p) = shaderF (gray a, gray d, gray s, p)

cBasicLightingNH = cBasicG "BasicLightingNH" basicNH (ripple' (3,0.2))
cBasicLightingRV = cBasicG "BasicLightingRV" basicRV (ripple' (3,0.2))

cAnisotropic = cBasicG "Anisotropic" (anisotropic1 (curry snd)) sphere3

fmRanges f0 m0 = ( Range "frequency" 0 10 f0
                 , Range "magnitude" 0 1 m0
                 )

cSphere = cwrap' "Sphere" (125,250) (const blue)
          () basicC lightBack (\ () -> 2.2 *^ sphere)

gold = rgb 1 (215/255) 0 :: ColorE

-- cSphereLit = cwrap' "SphereLit" (125,250) (const blue)

cRippleH = cwrap "RippleH" yellow (fmRanges 5 0.1) basicC lightPast
           ripple'


fmpRanges f0 m0 p0 = ( fmRanges f0 m0
                     , Range "phase" (-1/2) (1/2) p0
                     )

cRipple = cwrap "Ripple" yellow (fmpRanges 3 0.08 0) basicC lightPast
          rippleR

cRippleTextured = cwrap' "RippleTextured" defaultTess (checkerI . (7 *^))
                  (fmpRanges 1.5 0.1 0) basicC lightPast
                  rippleR

checkerI = cond checker (const yellow) (const blue)

checker :: Region
checker = cartF (==*) stripe stripe

-- checker (x,y) = stripe x ==* stripe y

stripe :: R -> BoolE
stripe x = frac x <=* 1/2

cond :: (FromScalar a b, GCond b c) => (d -> a) -> (d -> c) -> (d -> c) -> d -> c
cond bim = lift3 ifE (fromScalar . bim)


cRippleHTextured = cwrap' "RippleHTextured"
                   defaultTess (checkerI . (7 *^))
                   (fmRanges 5 0.1) basicC lightPast
           ripple'


torusR ((sr,cr),(sf,cf))
       = ( ( Range "sweepRad" 0 3 sr
           , Range "crossRad" 0 3 cr)
         , ( Range "sweepFrac" 0 1 sf
           , Range "crossFrac" 0 1 cf)
         )

torusF ((sr,cr),(sf,cf)) = torus sr cr . (* (cf,sf))

-- Torus with fractions of u & v
cTorus = cwrap "Torus" blue (torusR ((1.5,1/2),(1,1))) basicC lightPast torusF
         
-- Wrap a checker texture and add a circulation of u to show toroidal
-- energy flow.
cTorusTextured = cwrap' "TorusTextured"
                   defaultTess (checkerI . ((15,20) *))
                   ( torusR ((1.2,1),(1,1))
                   , Range "circulate" 0 2 0
                   )
                   basicC lightPast
                   (\ (tpars,cir) -> torusF tpars . ((cir,0) + ))


cHelicalTube = cwrap' "HelicalTube" (75,200) (const red)
               ( Range "periods" 0 10 5
               , Range "zrate" 0 2 (8/10)
               , Range "sweepRad" 0 3 0.7
               , Range "crossRad" 0 3 (1/4)
               ) basicC lightBack $
               uncurry4 helicalTube

cEggcrate = cwrap "Eggcrate" yellow (fmRanges 3 0.2) basicC lightPast $
            onXY' (5 *^) . hfSurf . eggcrate'

hR = Range "height"    0 10 3

fmh = ( Range "frequency" 0 8 3
      , Range "magnitude" 0 1 0.15
      , hR
      )

cWavyCylinder = cwrap "WavyCylinder" green fmh basicC lightBack $ \ (a,b,c) ->
                wavyCylinder (a,b) c

cEggcrateCylinder = cwrap "EggcrateCylinder" red fmh basicC lightBack $ \ (f,m,h) ->
                    (3/4) *^ displace (cylinder h) (eggcrate' (f,m))


cDimple = cwrap "Dimple" green () basicC lightPast $ \ () ->
           hfSurf dimple . (3 *^)

minZR = Range "minZ" 0 1

cDimpleCutoff = cwrap "DimpleCutoff" green ( minZR 0
                                           ) basicC lightPast $ \ minZ ->
                hfSurf (cutoffHF minZ dimple) . (3 *^)

ssR = ( Range "spacing" 0 5 2.5
      , Range "squeeze" 0 15 3
      )

zssR = (minZR 0.2, ssR)

dimples (minZ,(spacing,squeeze)) = squeezeF squeeze (tiledDimples minZ spacing)

dimples' (spacing,squeeze) = squeezeF squeeze (tiledDimples' spacing)

cDimpleTiled = cwrap "DimpleTiled" green zssR basicC lightPast $ \ zss  ->
                hfSurf (dimples zss) . (3 *^)


cDimpledCylinder' = cwrap "DimpledCylinder" green (hR,zssR)
                    basicC lightBack $ \ (h,zss) ->
                      (3/4) *^ displace (cylinder h) (dimples zss)

-- Too many instructions, so simplify:

cDimpledCylinder = cwrap "DimpledCylinder" green ssR
                   basicC lightBack $ \ ss ->
                     (3/4) *^ displace (cylinder 5)
                                (3 *^ dimples' ss . (* (2,3)))

-- Too many instructions (141)
cTorusTwist = cwrap "TorusTwist" blue ( Range "sweepRad" 0 3 2
                                      , Range "crossRad" 0 3 (1/2)
                                      , Range "twistRate" 0 3 1
                                      ) basicC lightPast $
         \ (sr,cr,rate) -> twist rate . torus sr cr

-- Too many instructions (141)
cPlaneTwist = cwrap "PlaneTwist" blue ( Range "twistRate" 0 10 5
                                      ) basicC lightBack $
         \ rate -> twist rate . onX' (4 *) (hfSurf (const 0))

plane = hfSurf (const 0)

cPlane = cwrap "Plane" blue () basicC lightPast $ \ () -> plane . (3 *^)

main = sequence [ cBasicLightingNH, cBasicLightingRV
                , cRipple, cRippleTextured, cTorus, cTorusTextured
                , cHelicalTube
                , cEggcrate, cWavyCylinder, cEggcrateCylinder
                , cDimple, cDimpleCutoff, cDimpleTiled
                , cPlaneTwist
                ]
