module PP where

import Char (isSpace)
import qualified Pretty as P
import qualified PrettyUtils as PU  -- sigh, Hugs98 doesn't allow two modules
				    -- to be mapped to the same 'as' 

import List ( intersperse )

infixl 6 <+>
infixl 5 $$

------ Pretty printing

type PPDoc = Context -> P.Doc

data Assoc = LAssoc | NAssoc | RAssoc deriving ( Eq )

type Context
 = ( Bool  -- True => on the right of a binary operator.
   , Assoc -- associativity of operator we're appearing as expr of.
   , Int   -- current precedence
   )


-- TODO: Rewrite most of these in terms of lifting combinators, e.g.,
--  (<+>) = lift2 (P.<+>)


-- Lifting the pretty printing combinators needed here.
text x = const (P.text x)

parens x = \ p -> P.parens (x p)

a <+> b  = \ x -> (P.<+>) (a x) (b x)
a <> b   = \ x -> (P.<>) (a x) (b x)
($$) a b = \ x -> (P.$$) (a x) (b x)
($+$) a b = \ x -> (P.$+$) (a x) (b x)

nest x d = \ p -> P.nest x (d p)

hang d1 x d2 = \ p -> P.hang (d1 p) x (d2 p)
char c = const (P.char c)

hsep xs = \ p -> P.hsep (map (flip ($) p) xs)
hcat xs = \ p -> P.hcat (map (flip ($) p) xs)
vcat xs = \ p -> P.vcat (map (flip ($) p) xs)
sep xs  = \ p -> P.sep  (map (flip ($) p) xs)

comma  = \ _ -> P.comma
equals = \ _ -> P.equals
semi   = \ _ -> P.semi

empty  = \ _ -> P.empty

onRight :: PPDoc -> PPDoc
onRight pd = \ (_,as,prec) -> pd (True,as,prec)

ppOperator :: (Int    -- precedence
              ,Assoc) -- associativity
           -> PPDoc
           -> PPDoc
ppOperator (p2,as2) pd (onR,as1,p1)
  | needParens = parens pd new_context
  | otherwise  = pd new_context
 where
  new_context = (False,as2,p2)
  
  needParens  =
    p1 > p2   ||                -- if a lower-precedence operator is appearing
    		                -- as an arg. to a higher prec. one, we do need parens.
    (p1 == p2 &&                -- same precedence ops will need parens, iff:
     ((isRA_op1 /= isRA_op2)  || --   * they have opposing associativity.
      (onR && not (isRA_op1)) || --   * the new operator appears on the right of a l. assoc operator.
      (not onR && isRA_op1)))    --   * the new op appears on the left of a r. assoc operator.
     
  isRA_op1 = as1 == RAssoc
  isRA_op2 = as2 == RAssoc


showPPDoc :: PPDoc -> Int -> String
showPPDoc x p = show (x (False,NAssoc,p))

-- The default line width is 100.  We could change to 72 via the
-- following, but Hugs is missing renderStyle and Style
-- 
-- showPPDoc x = P.renderStyle (P.Style 72 1.5 P.PageMode) (x 9)

ppInt   x = text (show x)
ppFloat x = text (show x) <> char 'f'
ppBool  x = text (show x)

isEmpty :: PPDoc -> Bool
isEmpty p = let x = p (False,NAssoc,0) in P.isEmpty x

vsepWith d ds  = \ p -> PU.vsepWith (d p) (map (flip ($) p) ds)
vsep ls        = \ p -> PU.vsep (map (flip ($) p) ls)
ppDecls ls     = \ p -> PU.ppDecls (map (flip ($) p) ls)
ppTuple ls     = \ p -> PU.ppTuple (map (flip ($) p) ls)
ppTupleVert ls = \ p -> PU.ppTupleVert (map (flip ($) p) ls)

-- TODO: refactor the definitions above.

stmtBlock n ds = char '{' $+$ nest n ds $$ char '}'


{-
ppConstructor :: String -> PPDoc
ppConstructor name =
  ppMethod "public" name [text "Device device"] [ text "m_device = device" ]
  $$
  text "Device m_device;"
-}

ppMethod :: String -> String -> [PPDoc] -> [PPDoc] -> PPDoc
ppMethod resultTy methName paramDecls statements =
  text resultTy <+> text methName <> ppTuple paramDecls
  $$
  stmtBlock 4 (ppDecls statements)

newArr :: String -> [PPDoc] -> PPDoc
newArr ty vec = text ("new " ++ ty ++ "[]{") <>
                hcat (intersperse comma vec) <>
                text "}"

ppCall :: String -> [PPDoc] -> PPDoc
ppCall fun args = text fun <> ppTuple args


---- Other stuff for generating C# code.  Probably should abstract into
---- PPDoc functions.

commentLineMax = 70 :: Int

-- Make a C# comment with line breaks as necessary
csComment :: String -> String
csComment =   prefixLines "// " . layout
 where
   layout :: String -> String
   layout "" = ""
   layout str = line ++ "\n" ++ layout rest
                where (line,rest) = divide str
   divide str | null rest = (trunc,"")
              | otherwise = (beforeSpace, afterSpace)
    where (trunc,rest) = splitAt commentLineMax str
          (revExtra,revBefore) = break isSpace (reverse trunc)
          beforeSpace          = reverse (dropWhile isSpace revBefore)
          afterSpace           = dropWhile isSpace (reverse revExtra ++ rest)


-- TODO: A simple & clearly correct reimplementation of csComment, as follows:

-- csComment = unlines . map ("// " ++) . map unwords . layout 70 . words
--  where
--    -- Break into groups of that will fit onto one line
--    layout :: [String] -> [[String]]
--    layout = ...

-- Like show, but indent each line
show' :: Show a => a -> String
show' = indentLines 4 . show

prefixLines :: String -> String -> String
prefixLines prefix = untabify . unlines . map (prefix ++) . lines

indentLines :: Int -> String -> String
indentLines indent = prefixLines (replicate indent ' ')

-- pretty printing inserts tabs, and indentation interferes with them
untabify :: String -> String
untabify = concat . map replace
 where
   replace '\t' = replicate 8 ' '
   replace ch   = [ch]

surround :: String -> String -> (String -> String)
surround before after = (before ++) . (++ after)

wrapBraces :: String -> String
wrapBraces = surround "{\n" "}\n"

{-
-- Wrap up generated C# code in a complete program module
ppCsModule :: String
              String -> String -> String
ppCsModule namespace using className 
            =   surround (unlines prelude) (unlines postlude)
                  . wrapNameSpace
                  . wrapClass
 where
   prelude           = [ "// Vertex shader \"" ++ name
                         ++ "\", generated by Vertigo compiler."
                       , ""
                       , concatMap (surround "using " ";\n") usedNameSpaces
                       ]
   postlude          = [ ]
   usedNameSpaces    = ["System", "Microsoft.DirectX",
                        "Microsoft.DirectX.Direct3D", "Microsoft.DirectX.Direct3DX"]
   wrapClass         = (("\npublic class " ++ name ++ " ") ++) . wrapBraces . indentLines 4
   wrapNameSpace str = "namespace Vertigo." ++ name ++ " " ++ wrapBraces str
-}