-- Pretty printing of expressions.  Handles operator precedence and some
-- simple syntactic sugaring.  Sync changes with PPDExp.

module PPDExp where

import Maybe (fromMaybe,fromJust)
import DExp
import Types
import PP

-- We could define the following Show instance, but let's not, so we can
-- see the structure when necessary, by means of the derived Show instance.

-- instance Show Exp where
--   showsPrec = showsPrecDExp

showsPrecExp p e = (showPPDoc (ppExp e) p ++)

-- Sometimes more convenient
showExp e = showsPrecExp 0 e ""

ppExp :: Exp -> PPDoc
ppExp (LitVec [x])     = text (show x)
ppExp (LitVec xs)      = ppTuple (map (text.show) xs)
ppExp (Var n ty)       = text n
ppExp (Apply op args)  = ppApply op args
ppExp (Let binds body) = ppOperator (0,NAssoc) $
                           text "let" $+$
                           nest 4 (vsep (map ppBind binds)) $$
                           text "in " $+$
                           nest 1 (ppExp body)
ppExp e                = error ("PPDExp.ppExp: no case for " ++ show e)

ppBind (v,e) = text v <+> text "=" <+> ppExp e


ppApply op = pp op
 where
   pp Add [e1, LitVec [x]]
     | x < 0                      = ppOp (ppBinOp "-" e1 (LitVec [-x]))
   pp Add [e1, Apply Negate [e2]] = ppOp (ppBinOp "-" e1 e2)
   pp Add [e1, e2]                = ppOp (ppBinOp "+" e1 e2)
   pp Mul [e1, Apply Rcp [e2]]    = ppOp (ppBinOp "/" e1 e2)
   pp Mul [e1, e2]                = ppOp (ppBinOp "*" e1 e2)
   pp Slt [e1, e2]                = ppOp (ppBinOp "<" e1 e2)
   pp Dp3 [e1, e2]                = ppOp (ppBinOp "<.>" e1 e2)
   pp Dp4 [e1, e2]                = ppOp (ppBinOp "<.>" e1 e2)
   pp Sge [e1, e2]                = ppOp (ppBinOp ">=" e1 e2)
   pp Negate [e1]                 = ppOperator (apPrec-1,LAssoc)
                                      (text "-" <> onRight (ppExp e1))
   pp (Swizzle ixs) [e]           = ppSwiz ixs e
   pp op args                     = ppOperator (apPrec,LAssoc)
                                      (text name <+> onRight (sep (map ppExp args)))
                                    where
                                      name | op == MkVec = "vec" ++ show (length args)
                                           | otherwise = opName op
   ppOp = ppOperator (opInfo op)


ppSwiz ixs e = ppOperator (apPrec,NAssoc) $
               (ppExp e <> text (showSwizzle False ixs))


ppSwiz' ixs e = ppOperator (apPrec,RAssoc) $
                text (swizF ixs) <+> onRight (ppExp e)

swizF ixs = "swiz" ++ map (assocJ (zip [0..] "XYZW")) ixs


-- ppSwiz ixs e = ppOperator (apPrec+1,LAssoc) $
--                ppExp e <> (text ('.' : ixls))
--  where
--    ixls = map (assocJ (zip [0..] "xyzw")) ixs

assocJ keyVals key = fromJust (lookup key keyVals)

ppBinOp :: String -> Exp -> Exp -> PPDoc
ppBinOp op e1 e2 = sep [ppExp e1, text op, onRight (ppExp e2)]


apPrec :: Int
apPrec  = 10

-- operator precedence levels
opPrec :: Op -> Maybe Int
opPrec Mul = Just 7
opPrec Dp3 = Just 7
opPrec Dp4 = Just 7
opPrec Add = Just 6
opPrec Slt = Just 4
opPrec Sge = Just 4
opPrec _   = Nothing

opAssoc :: Op -> Assoc
opAssoc op = LAssoc  -- Currently all left-associative

opInfo :: Op -> (Int, Assoc)
opInfo op = (fromMaybe 0 (opPrec op), opAssoc op)
