-- Register pools

module Pool where

import List (elemIndex,delete,(\\))
import Maybe (fromMaybe)

import Types

---- Register pool.

-- Each pool element is a temporary register and a mask saying which parts
-- are available.  (Recall that a mask is a layout in which the order is
-- unimportant.)
type Pool = [(Int,Mask)]

fullPool :: Int -> Pool
fullPool size = [(i,[0..3]) | i <- [0..size-1]]

-- Get a free register portion.  Yield the register number, layout, and
-- diminshed pool.  In the case of full-size (max == 4) allocation, we
-- guarantee to use the identity layout [0..3].  That guarantee helps in
-- generating code for non-simd and non-scalar-producing operations, which
-- at present all produce four-float results.

allocReg :: Int -> Pool -> (Int,Layout,Pool)

allocReg size = loop
 where
   loop [] = error "allocReg: no more registers"

   loop (free@(temp,mask):frees)
     | size == maskSize =
         (temp,mask,frees)  -- Just right
     | size >  maskSize = 
         let (temp',layout',frees') = loop frees in  -- Too small
           (temp',layout',free:frees')
     | otherwise =
         (temp, prefix, (temp,suffix):frees)
      
     where
       (prefix,suffix) = splitAt size mask
       maskSize = length mask


-- Temp hack for MkVec.  TODO: work out an elegant solution.
removeFromPool :: Int -> Mask -> Pool -> Pool
removeFromPool regNum remMask = filter (not.null.snd) . map trim
 where
   trim rm@(reg,mask) | regNum == reg = (reg, mask \\ remMask)
                      | otherwise     = rm


-- The swizzle necessary to re-arrange a value to fit a given layout For
-- instance, layoutSwizzle [2,1] = [arb,1,0], where "arb" are arbitrary.
layoutWriteSwizzle :: Layout -> Swizzle
layoutWriteSwizzle layout = map findIt [0 .. maximum layout]
 where
   findIt :: Index -> Index
   findIt i = fromMaybe arb (elemIndex i layout)
   -- Must be a legit index for the type.  Choosing the last available
   -- index leads to more succinct pretty-printed code when the layout is
   -- a prefix of a register.
   arb = length layout - 1


---- Monadic version

newtype Alloc a = Alloc (Pool -> (a,Pool))

instance Monad Alloc where
  return a = Alloc (\ pool -> (a,pool))
  Alloc f >>= h = Alloc $ \ pool ->
                  let (a, pool')  = f pool
                      Alloc g     = h a
                      (b, pool'') = g pool'
                  in
                      (b, pool'')

instance Functor Alloc where
  fmap f m = m >>= return.f

alloc :: Int -> Alloc (Int,Layout)
alloc size = Alloc $ \ pool ->
             let (regNum,layout,pool') = allocReg size pool in
             ((regNum,layout),pool')

runAlloc :: Pool -> Alloc a -> a
runAlloc pool (Alloc f) = fst (f pool)

-- TODO: revisit explicit uses of Pool & allocReg in CodeGen.
