-- Simple graph library.  Mostly derived from Oege de Moor's code in Pan.

module SimpleGraph where

import List ((\\))
import FiniteMap

{- BEGIN_GHC_ONLY
import IOExts (unsafePerformIO)
-- import GHC.Prim
import Stable
import DExp (Exp) -- see note below
   END_GHC_ONLY -}

import Types

-- The dependency on DExp is unfortunate here.  It's necessary only
-- because we need to define Node as StableName Exp in the GHC case.  Oh;
-- maybe I could simply unsafeCoerce the Exp to an Int.  In that case, I
-- could have kept using FGL.  Oh well.

{- BEGIN_GHC_ONLY
type Node = StableName Exp

ptrToNode :: Exp -> Node

ptrToNode e = e `seq` unsafePerformIO (makeStableName e)

-- According to the StableNames paper, an Ord instance is supposed to be
-- supplied, but it doesn't appear to be defined in Stable...so,

instance Ord (StableName a) where
   -- can go wrong should a GC strike between the eval of
   -- arg1 and arg2.
  compare x y = compare (hashStableName x) (hashStableName y)

instance Show (StableName a) where
  show x = show (hashStableName x)

   END_GHC_ONLY -}

{- BEGIN_HUGS_ONLY -}
type Node = Int

ptrToNode :: a -> Node

primitive unsafePtrToInt "unsafePtrToInt" :: a -> Int
ptrToNode e = e `seq` unsafePtrToInt e
{- END_HUGS_ONLY -}


type Path = [Node]


-- Node-labelled graphs

type Graph a = FiniteMap Node ([Node],a)

nodes :: Graph a -> [Node]
nodes = map fst . fmToList

children :: Graph a -> Node -> [Node]
children g n = case lookupFM g n of
                 Just (ns,_) -> ns
                 Nothing     -> []

edges :: Graph a -> [(Node,Node)]
edges g = [(a,b) | (a,(bs,_)) <- fmToList g, b<-bs]

addedge :: Graph a -> (Node,Node) -> Graph a
addedge g (a,b) = 
     addToFM_C (\ (x,a) (y,b) -> (y++x,a) )
               g a
	       ([b],error "undefined in Share.addedge")

labelsonly :: Graph a -> Graph a
labelsonly = mapFM (\ _ (_,b) -> ([],b))

converse :: Graph a -> Graph a
converse f =  foldl addedge (labelsonly f) (map swap (edges f))

topsort :: Graph a -> [Node]
topsort g = map fst (tsort [ (a,b,as) | (a,(as,b)) <- fmToList g])
 where
  tsort :: Eq a => [(a,b,[a])] -> [(a,b)]
  tsort [] = []
  tsort xs = (a,b):tsort [ (c,d,cs\\[a]) | (c,d,cs) <- xs, c /= a ]
               where (a,b,_) = head [ (c,d,cs) | (c,d,cs) <- xs, not (elem c (concat (map thd3 xs))) ]
                     thd3 (a,b,c) = c

-- Depth-first enumeration
depthFirst :: Graph a -> Node -> [Node]
depthFirst g node = -- ptrace ("depthFirst " ++ show node) $
                    search [node] []
 where
   search :: [Node] -> [Node] -> [Node]
   search []     seen                 = reverse seen
   search (n:ns) seen | n `elem` seen = search ns seen
                      | otherwise     = search (children g n ++ ns) (n:seen)

lab :: Graph a -> Node -> a
lab g n = label where Just (_,label) = lookupFM g n

subgraph :: (Node -> Bool) -> Graph a -> Graph a
subgraph pred g = foldl addNode emptyFM (nodes g)
 where
   addNode g node = if pred node then
                      addToFM g node (filter pred (children g node), lab g node)
                    else g

-- TODO: consider memoizing pred in subgraph since it gets applied
-- repeatedly.  Or just note that callers may want to memoize

-- subFM :: (Ord a) => FiniteMap a b -> (a -> Bool) -> FiniteMap a b
-- subFM m pred = addListToFM emptyFM (filter (pred . fst) (fmToList m))

invertFM :: (Ord a, Ord b) => FiniteMap a b -> FiniteMap b [a]
invertFM m = addListToFM emptyFM (gather (map swap (fmToList m)))

functionFM :: Ord a => [a] -> (a -> b) -> FiniteMap a b
functionFM xs f = addListToFM emptyFM (map (\ x -> (x, f x)) xs)

-------- Unused & untested



{-
mapNodes :: ([Node] -> a -> b) -> Graph a -> Graph b
mapNodes f = mapFM (\ _ (ns,a) -> (ns,f  ns a))

mapGraph :: (Node -> [Node] -> a -> b) -> Graph a -> Graph b
mapGraph f = mapFM (\ n (ns,a) -> (ns,f n ns a))

scan :: (Graph lab -> (lab,b) -> Node -> [a] -> a) -> Graph (lab,b) -> Graph ((lab,b),a)
scan f g = g'
           where g' = mapGraph (\ n ns a -> (a,f g'' a n (scans ns))) g
                 g'' = mapNodes (\ ns (a,b) -> a) g
                 scans = map (attr. label . lookupWithDefaultFM g' (error "undefd"))
                 label (n,l) = l
                 attr (l,a) = a 

sscan :: (lab -> [a] -> a) -> Graph lab -> Graph (lab,a)
sscan f g = g'
            where g' = mapGraph (\ n ns a -> (a,f a (scans ns))) g
                  scans = map (attr. label . lookupWithDefaultFM g' (error "undefd"))
                  label (n,l) = l
                  attr (l,a) = a 

-- tag the nodes with their max height (only makes sense for acyclic graphs)
labelHeight :: Graph a -> Graph (a,Int)
labelHeight g = g'
                where g'          = mapNodes f g
                      f ns a      = (a, 1+ foldr max (-1) (depths ns))
                      depths      = map (depth.label.lookupWithDefaultFM g' ([],(error "undefined in Share.labeldepth",-1)))
                      depth (a,d) = d
                      label (n,l) = l

labelTop :: Graph a -> Graph (a,Int)
labelTop g = mapFM (\ n (ns,a) -> (ns,(a, topnumber n)) ) g
             where topnumber = lookupWithDefaultFM top (error "undefined in Share.labelTop.topnumber") 
                   top = listToFM (zip (map node (topsort g)) [0..])
                   node (n,l) = n
-}
