module SmartDExp where

import Monad
import Maybe (maybe,isJust,fromJust,fromMaybe,listToMaybe)
import Int (fromInt)
import List (sort,transpose)
import Memo

{- BEGIN_HUGS_ONLY -}
import IOExts (unsafePtrEq)
{- END_HUGS_ONLY -}
{- BEGIN_GHC_ONLY
import GHC.Exts (reallyUnsafePtrEquality#)
   END_GHC_ONLY -}

import Vector
import Types
import DExp
import PPDExp


infix  4  ==*, /=*, <*, <=*, >=*, >*
infixr 3  &&*
infixr 2  ||*
infix  4  ==%

-- Sometimes it's useful to suppress optimization.  In that case, define
-- "optimize" as False.  We'll still do constant folding.  (Should there
-- be a separate flag for constant folding?)
optimize = True

{- BEGIN_GHC_ONLY
unsafePtrEq :: a -> a -> Bool
unsafePtrEq a b = case reallyUnsafePtrEquality# a b of
                    1# -> True
                    _  -> False
   END_GHC_ONLY -}

-- Equality test that first tries cheap pointer-equality
(==%) :: Exp -> Exp -> Bool
x ==% y = (x `unsafePtrEq` y) || x == y

equalExps :: [Exp] -> Bool
equalExps []     = True
equalExps (x:xs) = all (==% x) xs

-- Convert scalars to expressions by repetition
scalarTo :: AsFloat a => Type -> a -> Exp
scalarTo ty a = LitVec (replicate (type2size ty) (toFloat a))

-- Some favorites
zero, one :: Type -> Exp
one  ty = scalarTo ty (1 :: Float)
zero ty = scalarTo ty (0 :: Float)


type Op1 = Exp -> Exp
type Op2 = Exp -> Op1
type Op3 = Exp -> Op2
type Op4 = Exp -> Op3
type OpL = [Exp] -> Exp


type Try1 = Exp -> Maybe Exp
type Try2 = Exp -> Try1
type Try3 = Exp -> Try2
type Try4 = Exp -> Try3
type TryL = [Exp] -> Maybe Exp


-- Generalized constant folders

class CFold1 f where cfold1 :: f -> Try1
class CFold2 f where cfold2 :: f -> Try2
class CFold3 f where cfold3 :: f -> Try3
class CFold4 f where cfold4 :: f -> Try4
class CFoldL f where cfoldL :: f -> TryL


-- Trivial, always-failing constant folder.
data NoFold = NoFold

instance CFold1 NoFold where cfold1 _ _       = Nothing
instance CFold2 NoFold where cfold2 _ _ _     = Nothing
instance CFold3 NoFold where cfold3 _ _ _ _   = Nothing
instance CFold4 NoFold where cfold4 _ _ _ _ _ = Nothing
instance CFoldL NoFold where cfoldL _ _       = Nothing


-- Supported "literal" type.

isLit :: Exp -> Bool
isLit (LitVec _) = True
isLit _          = False

class Literal a where
  mkLit :: a -> Exp
  dsLit :: Exp -> Maybe a

-- Anything that can be encoded as a vector.
instance AsVector a => Literal a where
  mkLit a = LitVec (toVector a)
  dsLit (LitVec vec) = fromVector vec
  dsLit _            = Nothing



-- Scalar functions can be promoted to constant folders, simd style.

newtype Simd1 a b   = Simd1 (a -> b)
newtype Simd2 a b c = Simd2 (a -> b -> c)

wrapF1 :: (AsFloat a, AsFloat b) => (a -> b) -> (Float -> Float)
wrapF1 f = toFloat . f . fromFloat

wrapF2 :: (AsFloat a, AsFloat b, AsFloat c)
       => (a -> b -> c) -> (Float -> Float -> Float)
wrapF2 f a b = toFloat (f (fromFloat a) (fromFloat b))

instance (AsFloat a, AsFloat b) => CFold1 (Simd1 a b) where
  cfold1 (Simd1 f) e1 = liftM (\ xs -> mkLit (map (wrapF1 f) xs))
                              (dsLit e1)

instance (AsFloat a, AsFloat b, AsFloat c) => CFold2 (Simd2 a b c) where
  cfold2 (Simd2 f) e1 e2 =
    liftM2 (\ xs ys -> mkLit (zipWith (wrapF2 f) xs ys))
                             (dsLit e1) (dsLit e2)
                              

-- Literal->literal functions are constant folders

instance (Literal a, Literal b) => CFold1 (a -> b) where
  cfold1 f e1 = do a <- dsLit e1
                   return (mkLit (f a))

instance (Literal a, Literal b, Literal c) => CFold2 (a -> b -> c) where
  cfold2 f e1 e2 = do a <- dsLit e1
                      b <- dsLit e2
                      return (mkLit (f a b))

instance (Literal a, Literal b, Literal c, Literal d) =>
     CFold3 (a -> b -> c -> d) where
  cfold3 f e1 e2 e3 = do a <- dsLit e1
                         b <- dsLit e2
                         c <- dsLit e3
                         return (mkLit (f a b c))

instance (Literal a, Literal b, Literal c, Literal d, Literal e) =>
     CFold4 (a -> b -> c -> d -> e) where
  cfold4 f e1 e2 e3 e4 = do a <- dsLit e1
                            b <- dsLit e2
                            c <- dsLit e3
                            d <- dsLit e4
                            return (mkLit (f a b c d))


instance (Literal a, Literal b) => CFoldL ([a] -> b) where
  cfoldL f es = do as <- mapM dsLit es
                   return (mkLit (f as))


-- We can also pair up constant folders to make a constant folder that
-- tries the first and on failure the second.  Useful for overloaded
-- functions.  (See "==" and "+" below.)

instance (CFold1 f, CFold1 g) => CFold1 (f,g) where
  cfold1 (f,g) e1 = cfold1 f e1 `mplus` cfold1 g e1

instance (CFold2 f, CFold2 g) => CFold2 (f,g) where
  cfold2 (f,g) e1 e2 = cfold2 f e1 e2 `mplus` cfold2 g e1 e2

instance (CFold3 f, CFold3 g) => CFold3 (f,g) where
  cfold3 (f,g) e1 e2 e3 = cfold3 f e1 e2 e3 `mplus` cfold3 g e1 e2 e3

instance (CFold4 f, CFold4 g) => CFold4 (f,g) where
  cfold4 (f,g) e1 e2 e3 e4 =
    cfold4 f e1 e2 e3 e4 `mplus` cfold4 g e1 e2 e3 e4

instance (CFoldL f, CFoldL g) => CFoldL (f,g) where
  cfoldL (f,g) e = cfoldL f e `mplus` cfoldL g e


-- Optimizers: Try constant folding, then function-specific optimization.

opt1 :: CFold1 f => Op1 -> f -> Op1
opt1 opt f = h
 where
  h e1 = maybe (opt e1) id (cfold1 f e1)

opt2 :: CFold2 f => Op2 -> f -> Op2
opt2 opt f = h
 where
  h e1 e2 = maybe (opt e1 e2) id (cfold2 f e1 e2)

opt3 opt f = h
 where
  h e1 e2 e3 = maybe (opt e1 e2 e3) id (cfold3 f e1 e2 e3)

opt4 opt f = h
 where
  h e1 e2 e3 e4 = maybe (opt e1 e2 e3 e4) id (cfold4 f e1 e2 e3 e4)

optL :: CFoldL f => ([Exp] -> Exp) -> f -> ([Exp] -> Exp)
optL opt f = h
 where
  h es = maybe (opt es) id (cfoldL f es)


-- Make an optN function that either uses the given optimizer or the given
-- constructor.
mbOpt optN | optimize  = \ opt f con -> optN opt f
           | otherwise = \ opt f con -> optN con f

opt1C :: CFold1 f => Op1 -> f -> Op -> Op1
opt1C opt f op = mbOpt opt1 opt f (apply1 op)

opt2C :: CFold2 f => Op2 -> f -> Op -> Op2
opt2C opt f op = mbOpt opt2 opt f (apply2 op)

opt3C :: CFold3 f => Op3 -> f -> Op -> Op3
opt3C opt f op = mbOpt opt3 opt f (apply3 op)

opt4C :: CFold4 f => Op4 -> f -> Op -> Op4
opt4C opt f op = mbOpt opt4 opt f (apply4 op)

optLC :: CFoldL f => ([Exp] -> Exp) -> f -> Op -> ([Exp] -> Exp)
optLC opt f op = mbOpt optL opt f (Apply op)


------ Class- or Type-specific simplifiers

instance Ord Exp where
  min = opt2C minZ (Simd2 (min :: Binop Float)) Min
        where
          minZ e e' | e ==% e'  = e
                    | otherwise = apply2 Min e e'
  max = opt2C maxZ (Simd2 (max :: Binop Float)) Max
        where
          maxZ e e' | e ==% e'  = e
                    | otherwise = apply2 Max e e'
  compare = error "Exp: no compare method.  Use <*, etc."


type Unop  a = a -> a
type Binop a = a -> a -> a
type Cmp   a = a -> a -> Bool


--- Boolean

-- Booleans are represented by vectors of floats whose values are always
-- zero or one.  Since we're representing booleans as floats, we don't
-- need any boolean-specific optimizations.  Just expand and rely on
-- optimizations of the numeric expressions that result.

-- Strict (!) conditional.

ifE :: Op3
ifE cond a b = cond * a + notE cond * b

-- TODO: consider that we could optimize the above, assuming cond==1 in a
-- and cond==0 in b.

(==*), (/=*) :: Op2

-- There's no equality test, so we'll use the primitives >= and <
e1 ==* e2 = e1 >=* e2 &&* e2 >=* e1
e1 /=* e2 = e1 <*  e2 ||* e2 <*  e1

(<*), (<=*), (>=*), (>*) :: Op2

(<*) = opt2C (<#) (Simd2 ((<) :: Cmp Float)) Slt
 where
   -- Cannot commute literals to RHS this time, since "<" is not
   -- commutative, and ">" would just gets commuted back right away.
   -- c+e<c' or c*e<c' for literals c,c'
   -- (LitFloat x `Add` e) <# LitFloat x' = e <* LitFloat (x' - x)
   -- LitFloat x <# (LitFloat x' `Add` e') = LitFloat (x - x') <* e'
   Apply Add [e,e'] <# e'' | isLit e  && isLit e'' = e' <* e'' - e
                           | isLit e' && isLit e'' = e  <* e'' - e'
   e <# Apply Add [e',e''] | isLit e  && isLit e'  = e - e'  <* e''
                           | isLit e  && isLit e'' = e - e'' <* e'
   Apply Mul [e, e'@(LitVec xs)] <# e'' | all (>0) xs = e <* e''/e'
                                        | all (<0) xs = e >* e''/e'
   Apply Mul [e'@(LitVec xs), e] <# e'' | all (>0) xs = e <* e''/e'
                                        | all (<0) xs = e >* e''/e'
   e <# e' =  apply2 Slt e e'

(>=*) = opt2C (>=#) (Simd2 ((>=) :: Cmp Float)) Slt
 where
   -- TODO: optimizations
   e >=# e' =  apply2 Sge e e'


-- The others are derived:

a >*  b = b <* a
a <=* b = b >=* a

(&&*), (||*) :: Op2

-- e1 &&* e2 = e1 * e2
-- e1 ||* e2 = max (one (typeOf e1)) (e1 + e2)

-- TODO: is there a cheaper disjunction?  Yes!

(&&*) = min
(||*) = max

notE :: Op1
notE (Apply Slt [e1,e2]) = apply2 Sge e1 e2
notE (Apply Sge [e1,e2]) = apply2 Slt e1 e2
notE cond                = one (typeOf cond) - cond

-- TODO: use opt1C for notE, so we can use the "optimize" flag to turn
-- optimization off when desired.  Look for other non-uses.


-----  Numbers

-- Bogus instances needed for Num, etc.
instance Enum Exp
instance Real Exp

instance Num Exp where
  (+) = opt2C (+#) (Simd2 ((+) :: Binop Float)) Add
   where
     -- TODO: do more like *.  Don't move literals to right and don't
     -- re-associate.  Search for literals to combine.

     -- Move literals to right.  Since we've already done constant folding, 
     -- we won't make a loop here
     a@(LitVec _) +# b = b + a
     a +# LitVec vec | all (==0) vec = a

     -- Reassociate to the left.  Helps to expose constants for
     -- anti-distribution.  Also reduces parens when pretty-printed, since
     -- "+" is left-associative.
     e1 +# Apply Add [e2,e3]           = (e1 + e2) + e3

     Apply Negate [e1] +# Apply Negate [e2] = -(e1+e2)

     -- Cancellation.  Subsumed by factoring.
     -- e +# e' | e ==% negate e' = zero (typeOf e)

     -- Combine literals.  (Can only appear as right argument.)
     Apply Add [e,x@(LitVec _)] +# x'@(LitVec _) = e + (x + x')
     Apply Add [e,x@(LitVec _)] +# Apply Add [e',x'@(LitVec _)] =
       (e+e') + (x + x')
     -- Pull literals out of left addend
     Apply Add [e1,x2@(LitVec _)] +# e3 = (e1+e3) + x2
     -- Try factoring
     e +# e' | isJust mb = fromJust mb where mb = tryFactorSum e e'

     -- sin^2 u + cos^2 u --> 1 (modulo AC)
     e +# e' | isJust mb = fromJust mb where mb = trySum (sym2 sin2cos2) e e'

{-
     -- Combine literals.  (Can only appear as right argument.)
     (e `Add` e'@(LitInt    n)) +# e''@(LitInt    _) = e + (e' + e'')
     (e `Add` e'@(LitFloat _)) +# e''@(LitFloat _) = e + (e' + e'')
     (e `Add` e'@(PackF32 (LitFloat _) (LitFloat _) (LitFloat _) (LitFloat _)))
       +# e''@(PackF32 (LitFloat _) (LitFloat _) (LitFloat _) (LitFloat _)) = e + (e' + e'')
     -- Pull literals out of left summand
     (e `Add` e'@(LitInt    _)) +# e'' = (e + e'') + e'
     (e `Add` e'@(LitFloat _)) +# e'' = (e + e'') + e'
     (e `Add` e'@(PackF32 (LitFloat _) (LitFloat _)
                          (LitFloat _) (LitFloat _))) +# e'' = (e + e'') + e''
     -- Anti-distribution
     Mul a b +# Mul a' b' | a ==% a' = a * (b + b')
     Mul a b +# Mul a' b' | b ==% b' = (a + a') * b
     I2f a +# I2f b                 = i2f (a + b)
     I2f a +# LitFloat x | r == 0 = i2f (a + LitInt k)
       where (k,r) = properFraction x
     -- If nothing else works, try reassociating to the left.  Helps to
     -- expose constants for anti-distribution.  Also reduces parens when
     -- pretty-printed, since "+" is left-associative.
     e +# (e' `Add` e'')            = (e + e') + e''
-}
     e +# e'                        = apply2 Add e e'

  (*) = opt2C (*#) (Simd2 ((*) :: Binop Float)) Mul
   where
     -- Move literals to left.  Since we've already done constant folding, 
     -- we won't make a loop here
     -- a *# b@(LitVec _) = b * a
     -- Special literal multipliers
--      lit@(LitVec vec) *# b | all (==  1) vec = b
--                            | all (==  0) vec = lit
--                            | all (== -1) vec = negate b

     -- Combine literals.  (Can only appear as right argument.)
     -- Apply Mul [e,x@(LitVec _)] *# x'@(LitVec _) = e * (x * x')
     -- Apply Mul [e,x@(LitVec _)] *# Apply Mul [e',x'@(LitVec _)] =
     --  (e*e') * (x * x')
     e1 *# e2 | isJust mb = fromJust mb where mb = trySpecialLitProd e1 e2

     -- Pull literals out of multiplicands
     -- Apply Mul [e1,x2@(LitVec _)] *# e3 = (e1*e3) * x2
     e1 *# e2 | isJust mb = f * (e1'*e2')
              where mb          = tryFactorLiterals e1 e2
                    (f,e1',e2') = fromJust mb

     -- a^n * a^m == a^(n+m)
     e1 *# e2 | isJust mb = fromJust mb where mb = tryPowersProd e1 e2

     -- Restricted distributive rule, to combine literals.  We know e'*x
     -- will get folded, so still one add and one multiply, but then we
     -- may get more simplification in e*x.
     -- TODO: is this rule really helpful?
     Apply Add [e, e'@(LitVec _)] *# x@(LitVec _) = e*x + e'*x

     -- Combine recips.  TODO: search for terms, subsuming this rule and next.
     Apply Rcp [e1] *# Apply Rcp [e2] = recip (e1*e2)

     -- Set up for previous rule
     Apply Mul [e1, e2@(Apply Rcp _)] *# e3@(Apply Rcp _) = e1 * (e2*e3)

     -- Reassociate to the left.  Helps to expose constants for
     -- anti-distribution.  Also reduces parens when pretty-printed, since
     -- "*" is left-associative.  Remove for now, as it destroys too much
     -- sharing, especially with exponentiation, and because I now search
     -- for literals to combine.
     -- e1 *# Apply Mul [e2,e3] = (e1 * e2) * e3

     -- Float negations upward
     Apply Negate [e1] *# e2 = -(e1*e2)
     e1 *# Apply Negate [e2] = -(e1*e2)

     -- Inverses.  Note handles both "recip a * a" and "a * recip a",
     -- because of double recip rule.
     -- e *# e' | e ==% recip e' = one (typeOf e)
     e *# e' | isJust mb = fromJust mb where mb = tryRecipsProd e e'

     -- Product of reciprocal square roots.  TODO: search.  Hmm... it
     -- seems to find the pairs without an explicit search here.  Why?
     e1@(Apply Rsq [a1]) *# e2@(Apply Rsq [a2]) = rsqrt (a1*a2)
     -- e *# e' | isJust mb = fromJust mb where mb = trySqrtsProd e e'

{-
     -- TODO: resurrect these rules as I'm ready to test them.

     -- Introduce exponents.  Do before re-association in case e is a product.
     Pow e n *# e' | e ==% e' = e ** (n + one Float)
     e' *# Pow e n | e ==% e' = e ** (n + one Float)
     Pow e n *# Pow e' n' | e ==% e' = e ** (n + n')
     e *# e'       | e ==% e' = e ** LitFloat 2
-}
     e *# e'                 = apply2 Mul e e'

  negate = opt1C negateZ (Simd1 (negate :: Unop Float)) Negate
   where
     negateZ (Apply Negate [e]) = e
     -- negateZ (Apply Add [e,e']) = -e + -e'
     negateZ (Apply Mul [e,e']) | isLit e  = negate e * e'
                                | isLit e' = e * negate e'
     -- Don't do any more, since negation is free in almost all cases.
     negateZ e            = apply1 Negate e

  abs e = -- ifE (e <* 0) e (-e)
          max e (-e)   -- One machine instruction


  fromInteger = error "No fromInteger for DExp.Exp"

instance Fractional Exp where
  -- Use default for /:  a/b = a * recip b
  recip = opt1C recipZ (recip :: Unop Float) Rcp  -- No simd
    where
      recipZ (Apply Rcp [e])                 = e
      recipZ (Apply Mul [e, Apply Rcp [e']]) = e'/e
      recipZ (Apply Exp [e])                 = exp (-e)
      recipZ e                               = apply1 Rcp e
  
  fromRational = error "fromRational: no method for Exp"


-- floorE is defined in Exp.hs.

rsqrt :: Op1
rsqrt = opt1C rsqrtZ ((recip . sqrt) :: Unop Float) Rsq
 where
   -- rsqrtZ (Apply Mul [e,e']) | e ==% e' = recip e
   -- rsqrt(a*a*A) --> a * rsqrt(A)
   rsqrtZ e | isJust mb = fromJust mb
     where mb = do (r,e') <- tryExtractSqr e
                   return (recip r * rsqrt e')
   rsqrtZ e = apply1 Rsq e

instance Floating Exp where
  pi     = error "pi: no method for Exp"  -- because we need to know the type
  sin    = -- error "sin: no method yet for Exp"
           opt1C sinZ  (sin :: Unop Float) Sin  -- no simd
            where sinZ (Apply Negate [e]) = - sin e
                  sinZ e                  = apply1 Sin e
  cos    = -- error "cos: no method yet for Exp"
           opt1C cosZ  (cos :: Unop Float) Cos  -- no simd
            where cosZ (Apply Negate [e]) = cos e
                  cosZ e                  = apply1 Cos e
  sqrt x = recip (rsqrt x)
           -- x * rsqrt x  
  log    = opt1C logZ (log :: Unop Float) Log  -- no simd
            where logZ (Apply Exp [e]) = e
                  logZ e               = apply1 Log e
  exp    = opt1C expZ (exp :: Unop Float) Exp   -- no simd
            where expZ (Apply Log [e]) = e
                  expZ e               = apply1 Exp e

-- About sqrt x, using x * rsqrt x is supposedly more accurate than
-- recip (rsqrt x).  It does have the drawback of tying up more registers
-- sometimes, since it needs to preserve x longer.


---- Substitution.

class Substable a where
  subst :: Env Exp -> a -> a

instance Substable Exp where
  subst env = msub
   where
     msub                = memo sub
     sub e@(Var v _)     = fromMaybe e $
                           lookup v env
     sub e@(LitVec _)    = e
     sub (Apply op args) = applyOpt op (map msub args)

instance (Substable a, Substable b) => Substable (a,b) where
  subst env (a,b) = (subst env a, subst env b)

instance (Substable a, Substable b, Substable c)
  => Substable (a,b,c) where
  subst env (a,b,c) = (subst env a, subst env b, subst env c)

instance (Substable a, Substable b, Substable c, Substable d)
         => Substable (a,b,c,d) where
  subst env (a,b,c,d) = (subst env a, subst env b, subst env c, subst env d)


-- Like Apply, but uses smart constructors
applyOpt :: Op -> [Exp] -> Exp

applyOpt op = case op of
              Add        -> op2 (+)
              Max        -> op2 max
              Min        -> op2 min
              Mul        -> op2 (*)
              Sge        -> op2 (>=*)
              Slt        -> op2 (<*)
              Rcp        -> op1 recip
              Rsq        -> op1 rsqrt
              Log        -> op1 log
              Exp        -> op1 exp
              Dp3        -> op2 dp3
              Dp4        -> op2 dp4
           -- Expp       -> op1 expp
           -- Logp       -> op1 logp
           -- Frc        -> op1 frac
              Negate     -> op1 negate
              Swizzle sw -> op1 (swizzle sw)
           -- MkVec      -> op1 mkvec
              Cos        -> op1 cos
              Sin        -> op1 sin
              Frac       -> op1 frac
              _          -> Apply op
 where
   op1 f [e1]    = f e1 
   op2 f [e1,e2] = f e1 e2







---- Other misc

-- expp :: Op1
-- expp = apply1 Expp

frac :: Op1
frac = opt1C (apply1 Frac) fracF Frac
 where
   fracF :: Float -> Float
   fracF z = z - fromInt (floor z)


swizzle :: [Int] -> Op1
swizzle = \ swiz -> opt1C (swizZ swiz)
                          ((swizzleVals swiz) :: [Float] -> [Float])
                          (Swizzle swiz)
 where
   swizZ swB (Apply (Swizzle swA) [e])     = swizzle (compSwiz swB swA) e
   swizZ [i] (Apply MkVec args)            = args !! i
   swizZ sw e | sw == idSwizzle (typeOf e) = e
   swizZ sw e                              = apply1 (Swizzle sw) e

-- swizzle swiz (LitVec vec)      | optimize = LitVec (swizzleVals swiz vec)
-- swizzle [i] (Apply MkVec args) | optimize = args !! i
-- swizzle swiz e                            = apply1 (Swizzle swiz) e

-- TODO:

select :: Int -> Op1
select i = swizzle [i]

vecX = select 0
vecY = select 1
vecZ = select 2
vecW = select 3


-- Vector formation.

-- + If all args of vecN are equal, do a simple fan-out swizzle.
-- + If all (scalar) args are swizzles of the same expression, collect the
--   indices into a single swizzle.
-- + If all args share a top-level simd-able operator, then apply that
--   operator to vectors of corresponding arguments.

vecOpt :: [Exp] -> Exp
vecOpt args | not optimize = Apply MkVec args
vecOpt args =
  fromMaybe (Apply MkVec args) $
  tryFanout `mplus` trySwizzle `mplus` trySimd
 where
  tryFanout  = do whenF (equalExps args) $
                    swizzle (map (const 0) args) (head args)
  trySwizzle = do (indices,exps) <- fmap unzip (mapM asSelect args)
                  whenF (equalExps exps) $
                    swizzle indices (head exps)
  trySimd    = do (ops@(op:_),argss) <- fmap unzip (mapM asApply args)
                  whenF (equal ops && (op `elem` simdOps)) $
                    Apply op (map vecOpt (transpose argss))

-- TODO: do I need to invoke the *optimized* simd operators in trySimd?


asApply                 :: Exp -> Maybe (Op,[Exp])
asApply (Apply op args) = Just (op,args)
asApply _               = Nothing

asSelect                             :: Exp -> Maybe (Int,Exp)
asSelect (Apply (Swizzle [i]) [exp]) = Just (i,exp)
asSelect _                           = Nothing


vec2 :: Op2
vec2 = opt2C (appL2 vecOpt) ((,):: Float -> Float -> Vec2) MkVec

vec3 :: Op3
vec3 = opt3C (appL3 vecOpt)
             ((,,):: Float -> Float -> Float -> Vec3)
             MkVec

vec4 :: Op4
vec4 = opt4C (appL4 vecOpt)
             ((,,,):: Float -> Float -> Float -> Float -> Vec4)
             MkVec

vecL :: [Exp] -> Exp
vecL [a]       = a
vecL [a,b]     = vec2 a b
vecL [a,b,c]   = vec3 a b c
vecL [a,b,c,d] = vec4 a b c d
vecL l         = vecOpt l


-- Dot products

dp3 :: Op2
dp3 = opt2C dpZ ((<.>) :: Vec3 -> Vec3 -> Float) Dp3
 where
   -- Pull literals out of arguments
   e1 `dpZ` e2 | isJust mb = f * (e1' `dp3` e2')
               where mb          = tryFactorReplScalars e1 e2
                     (f,e1',e2') = fromJust mb
   e1 `dpZ` e2 = apply2 Dp3 e1 e2


dp4 :: Op2
dp4 = opt2C dpZ ((<.>) :: Vec4 -> Vec4 -> Float) Dp4
 where
   -- Pull literals out of multiplicands
   e1 `dpZ` e2 | isJust mb = f * (e1' `dp4` e2')
               where mb          = tryFactorReplScalars e1 e2
                     (f,e1',e2') = fromJust mb
   e1 `dpZ` e2 = apply2 Dp4 e1 e2

-- TODO: merge dp3 & dp4 into a Dot & separate out in time for code generation.



---- More optimizations 

-- Interpret an expression as application of a binary operator (usually AC)
type AsOp2 = Exp -> [(Exp,Exp)]

-- Yield the various sum forms of an expression e.  Each sum form is a
-- pair (p,e'), such that e==p+e', p is not a sum and not zero, and e' may
-- be a sum and may be zero.
asSum :: AsOp2
asSum (Apply Add [e1,e2]) = f e1 e2 (+) ++ f e2 e1 (flip (+))
 where
   f a b op = map (onSnd (op a)) (asSum b)
asSum e                   = [(e,zero (typeOf e))]

-- Similarly for products: pairs (f,e'), where e==f*e', f is a
-- non-product, and e' may be 1.
asProd :: AsOp2

asProd (Apply Mul [e1,e2]) =
  f e1 e2 (*) ++ f e2 e1 (flip (*))
 where
   f a b op = map (onSnd (op a)) (asProd b)

-- For negation, factor e, and try negation with either factor.
asProd (Apply Negate [e]) = f onFst ++ f onSnd
 where
   f onPart = map (onPart negate) asp
   asp = asProd e

asProd e                   = [(e,one (typeOf e))]


-- Promote a Try2 to one that works modulo AC
tryAC :: Op2 -> AsOp2 -> Try2 -> Try2
tryAC op asOp try e1 e2 = 
  msum [ fmap (`op` (e1' `op` e2')) (try p1 p2)
       | (p1,e1') <- asOp e1
       , (p2,e2') <- asOp e2
       ]

trySum, tryProd :: Try2 -> Try2
trySum  = tryAC (+) asSum
tryProd = tryAC (*) asProd

-- Try a rule both ways
sym2 :: Try2 -> Try2
sym2 try e1 e2 = try e1 e2 `mplus` try e2 e1

-- sin^2 u + cos^2 u --> 1.  Combine with trySum to apply modulo AC.
sin2cos2 :: Try2
sin2cos2 (Apply Mul [a@(Apply Sin [u]), a']) (Apply Mul [b@(Apply Cos [v]), b'])
  | a ==% a' && b ==% b' && u ==% v
  = return (one scalar)
sin2cos2 _ _ = fail ""


-- TODO: Use trySum & tryProd in some of the rules below.  May require
-- generalization in some cases.

-- TODO: replace names like "tryFactorSum" to names like "factorSumAC"
-- when they do search.

-- TODO: maybe rename "Try2" etc to "Rule2" etc.


-- Factor a sum.
tryFactorSum :: Try2
tryFactorSum e1 e2 = listToMaybe [ f1 * (p1' + p2') + e1' + e2' 
                                 | (p1,e1') <- asSum e1
                                 , (p2,e2') <- asSum e2
                                 , (f1,p1') <- asProd p1
                                 , (f2,p2') <- asProd p2
                                 , f1 ==% f2
                                 ]

-- Cancel reciprocal pairs in products
tryRecipsProd :: Try2
tryRecipsProd e1 e2 = listToMaybe [ e1' * e2'
                                  | (f1,e1') <- asProd e1
                                  , (f2,e2') <- asProd e2
                                  , f2 ==% recip f1
                                  ]

{-  -- 
-- Cancel square root pairs in products
trySqrtsProd :: Try2
trySqrtsProd e1 e2 = ...
-}

-- Given e, find (f,e') such that e=f^2*e'
tryExtractSqr :: Exp -> Maybe (Exp,Exp)
tryExtractSqr e = listToMaybe [ (f'',e'')
                              | (f', e' ) <- asProd e
                              , (f'',e'') <- asProd e'
                              , f' ==% f''
                              ]

{-
-- Find literals for combination
tryFactorLiterals :: Exp -> Exp -> Maybe (Exp,Exp,Exp)
tryFactorLiterals e1 e2 = listToMaybe [ (f1*f2, e1', e2')
                                      | (f1,e1') <- asLitProd e1
                                      , (f2,e2') <- asLitProd e2
                                      ]
 where
   asLitProd = filter (isLit . fst) . asProd
-}


-- Find literals for combination.  Take a Try1 to filter & transform the
-- literals.  Used with dot products, where we can only extract literals
-- that are repetitions of a single scalar, and we want the scalar back.
tryFactorLiterals' :: Try1 -> Exp -> Exp -> Maybe (Exp,Exp,Exp)
tryFactorLiterals' tryLit e1 e2 = 
  listToMaybe [ (f1*f2, e1', e2')
              | (Just f1, e1') <- asLitProd e1
              , (Just f2, e2') <- asLitProd e2
              ]
 where
   asLitProd = map (onFst tryLit) . filter (isLit . fst) . asProd

tryFactorLiterals :: Exp -> Exp -> Maybe (Exp,Exp,Exp)
tryFactorLiterals = tryFactorLiterals' Just

{-  
-- Works, but no longer used
tryFactorLiteralsScalar = tryFactorLiterals' try
 where
   try (LitVec xs) | equal xs  = Just (LitVec [(head xs)])
                   | otherwise = Nothing
-}

-- View an expression as a replicated scalar
asReplScalar :: Try1
asReplScalar (LitVec xs)
  | equal xs  = Just (LitVec [(head xs)])
asReplScalar (Apply (Swizzle ixs) [e])
  | equal ixs = Just (swizzle [head ixs] e)
asReplScalar _ = Nothing

asReplScalarProd :: Exp -> [(Exp,Exp)]
asReplScalarProd e = [ (f,e2)
                     | (e1,e2) <- asProd e
                     , Just f  <- [asReplScalar e1]
                     ] 

-- Similar to tryFactorLiteralsScalar, but don't require the scalar to be
-- a literal.
tryFactorReplScalars :: Exp -> Exp -> Maybe (Exp,Exp,Exp)
tryFactorReplScalars e1 e2 = 
  listToMaybe [ (f1*f2, e1', e2')
              | (f1, e1') <- asReplScalarProd e1
              , (f2, e2') <- asReplScalarProd e2
              ]

-- Look for factor of 0, 1, or -1
trySpecialLitProd :: Try2
trySpecialLitProd = sym2 try
 where
   try lit@(LitVec vec) b | all (==  0) vec = Just lit
                          | all (==  1) vec = Just b
                          | all (== -1) vec = Just (negate b)
                          | otherwise       = Nothing
   try _ _ = Nothing


-- Coalesce common factors into a power tree.
tryPowersProd :: Try2
tryPowersProd e1 e2 = listToMaybe [ f1^(n1+n2) * e'
                                  | (f1,n1,e1') <- asPowProd e1
                                  , (f2,n2,e2') <- asPowProd e2
                                  , f1 ==% f2
                                  , let e' = e1' * e2'
                                  , e' /= one (typeOf e1)
                                  ]

-- Factor e into f^n * e', assuming that like factors are already together
asPowProd :: Exp -> [(Exp, Int, Exp)]

asPowProd e@(LitVec _) = []  -- We deal with literals in tryFactorLiterals

asPowProd e | n > 1 = [(f, n, one (typeOf e))]
 where (f,n) = asPowTree e

asPowProd (Apply Mul [e1,e2]) = asp e1 e2 (*) ++ asp e2 e1 (flip (*))
 where
   asp a b op = [ (f, n, a' `op` b) | (f,n,a') <- asPowProd a ]

asPowProd (Apply Negate [e]) = [(f,n,-e') | (f,n,e') <- asPowProd e]

asPowProd e = [(e, 1, one (typeOf e))]

-- View an expression as e^n, where n is preferably /= 1.  (Recips?)
asPowTree :: Exp -> (Exp,Int)
asPowTree e@(Apply Mul [e1,e2]) | f1 ==% f2 = (f1, n1+n2)
                                | otherwise = (e,1)
 where
  (f1,n1) = asPowTree e1
  (f2,n2) = asPowTree e2
asPowTree e = (e,1)

-- TODO: Maybe it would be better to represent power trees more directly,
-- as "PowI a n", and then rewrite away in the Approx phase.  Seems much
-- more straightforward.  We could then also choose to rewrite very large
-- exponents to a**n.  Recip would be a special case with n=-1, and the
-- approximation phase introduces it also for small negative n.

-- Even better, instead of integer powers, use FloatE powers, and so
-- subsume a^n, recip a, and 2^a all into a Pow constructor, and then
-- remove that constructor during the approximation phase.  Think of this
-- phase as mapping from a representation that's convenient for algebraic
-- manipulation (including differentiation) into one that uses only
-- executable operators.

-- Handle products & powers in analogy to sums & products

-- Yes.  I think this is the way to go.

