---- Statically typed expressions
---- Just a wrapper, using phantom types.

module TExp where

import Int (fromInt)

import Function
import Vector
import Types
import PP
import qualified DExp as D
import DExp (Op(..))
import qualified SmartDExp as SD
import qualified PPDExp
import NameStruct
import qualified CSE
import qualified NumOps

infix  4  ==*, /=*, <*, <=*, >=*, >*
infixr 3  &&*
infixr 2  ||*


-- The representation
newtype Exp a = E D.Exp

type BoolE   = Exp Bool
type FloatE  = Exp Float
type Vec1E   = Exp Vec1   -- Same as FloatE
type Vec2E   = Exp Vec2
type Vec3E   = Exp Vec3
type Vec4E   = Exp Vec4
type OutputE = Exp Output

data Output  -- Dummy output type

unE :: Exp a -> D.Exp
unE (E e) = e

mkE :: D.Exp -> Exp a
mkE = E

undefinedE = mkE D.Undefined

instance Show (Exp a) where
  showsPrec p (E e) = PPDExp.showsPrecExp p e

ppExp :: Exp a -> PPDoc
ppExp (E e) = PPDExp.ppExp e

---- Convenient function wrappers

typ1 :: (D.Exp -> D.Exp) -> (Exp a -> Exp b)
typ2 :: (D.Exp -> D.Exp -> D.Exp) -> (Exp a -> Exp b -> Exp c)
typ3 :: (D.Exp -> D.Exp -> D.Exp -> D.Exp) -> (Exp a -> Exp b -> Exp c -> Exp d)
typ4 :: (D.Exp -> D.Exp -> D.Exp -> D.Exp -> D.Exp) -> (Exp a -> Exp b -> Exp c -> Exp d -> Exp e)
-- ...

typ1 f (E e1)                      = E (f e1)
typ2 f (E e1) (E e2)               = E (f e1 e2)
typ3 f (E e1) (E e2) (E e3)        = E (f e1 e2 e3)
typ4 f (E e1) (E e2) (E e3) (E e4) = E (f e1 e2 e3 e4)


-- Specialization to Apply formers
apply1 :: Op -> (Exp a -> Exp b)
apply2 :: Op -> (Exp a -> Exp b -> Exp c)
apply3 :: Op -> (Exp a -> Exp b -> Exp c -> Exp d)
apply4 :: Op -> (Exp a -> Exp b -> Exp c -> Exp d -> Exp e)

apply1 = typ1 . D.apply1
apply2 = typ2 . D.apply2
apply3 = typ3 . D.apply3
apply4 = typ4 . D.apply4

-- Convert to a literal of the right type
constE :: AsVector a => a -> Exp a
constE = mkE . D.LitVec . toVector


---- Variable introduction

var :: AsVector a => Id -> Exp a
var n = e
 where
   e = E (D.Var n (expType e)) 

expType :: AsVector a => Exp a -> Type
expType = Vec . expTypeSize

expTypeSize :: AsVector a => Exp a -> Int
expTypeSize e = length (toVector (eVal e))

eVal :: Exp a -> a
eVal = error "eVal bogus value"


---- Naming and staging

instance AsVector a => AsVar Id (Exp a) where
  asVar name e = ([(name,expType e)], var name)

staged :: (AsVar uN uniform, AsVar vN varying)
       => uN -> vN -> (uniform -> varying -> Exp o)
       -> (Exp o, TypeEnv, TypeEnv)
staged uN vN f = (withVarying, uniformEnv, varyingEnv)
 where
 (uniformEnv,withUniform) = unFunction uN f
 (varyingEnv,withVarying) = unFunction vN withUniform

---- Vector formation

vec1 :: Exp a -> Exp a
vec1 = id

vec2 :: Exp x -> Exp y -> Exp (x,y)
vec2 = typ2 SD.vec2

vec3 :: Exp x -> Exp y -> Exp z -> Exp (x,y,z)
vec3 = typ3 SD.vec3

vec4 :: Exp x -> Exp y -> Exp z -> Exp w -> Exp (x,y,z,w)
vec4 = typ4 SD.vec4


class FromScalar s v | v -> s where
  fromScalar :: s -> v

instance FromScalar FloatE FloatE where {fromScalar = id}
instance FromScalar BoolE  BoolE  where {fromScalar = id}

instance FromScalar (Exp a) (Exp (a,a))     where fromScalar s = vec2 s s
instance FromScalar (Exp a) (Exp (a,a,a))   where fromScalar s = vec3 s s s
instance FromScalar (Exp a) (Exp (a,a,a,a)) where fromScalar s = vec4 s s s s

instance FromScalar s v => FromScalar (a->s) (a->v) where
  fromScalar f = fromScalar . f   -- \ a -> fromScalar (f a)


instance VectorOf FloatE FloatE where {(*^) = (*); (<.>) = (*)}

instance VectorOf FloatE Vec2E where
  s *^ vec = fromScalar s * vec
  u <.> v  = ux * vx + uy * vy where (ux,uy) = coords u
                                     (vx,vy) = coords v

-- The <.> def above is faster than forming two vec3's and dotting.  It
-- will use only two instructions, compared with five, and tie up fewer
-- registers.  In the case of u == v, which is fairly common, we're still
-- a win at two vs three.

instance VectorOf FloatE Vec3E where
  s *^ vec = fromScalar s * vec
  (<.>)    = typ2 SD.dp3

instance VectorOf FloatE Vec4E where
  s *^ vec = fromScalar s * vec
  (<.>)    = typ2 SD.dp4


class VecX a x | a -> x where vecX :: Exp a -> Exp x
                              vecX = typ1 SD.vecX
class VecY a y | a -> y where vecY :: Exp a -> Exp y
                              vecY = typ1 SD.vecY
class VecZ a z | a -> z where vecZ :: Exp a -> Exp z
                              vecZ = typ1 SD.vecZ
class VecW a w | a -> w where vecW :: Exp a -> Exp w
                              vecW = typ1 SD.vecW

instance VecX Float Float
instance VecX Bool  Bool

instance VecX (x,y) x
instance VecY (x,y) y

instance VecX (x,y,z) x
instance VecY (x,y,z) y
instance VecZ (x,y,z) z

instance VecX (x,y,z,w) x
instance VecY (x,y,z,w) y
instance VecZ (x,y,z,w) z
instance VecW (x,y,z,w) w

-- Swizzlings

-- Selector function

type Sel a b = Exp a -> Exp b

swiz1 :: (Sel a x) -> Sel a x
swiz1 s1 vec = vec1 (s1 vec)

swiz2 :: (Sel a x, Sel a y) -> Sel a (x,y)
swiz2 (s1,s2) vec = vec2 (s1 vec) (s2 vec)

swiz3 :: (Sel a x, Sel a y, Sel a z) -> Sel a (x,y,z)
swiz3 (s1,s2,s3) vec = vec3 (s1 vec) (s2 vec) (s3 vec)

swiz4 :: (Sel a x, Sel a y, Sel a z, Sel a w) -> Sel a (x,y,z,w)
swiz4 (s1,s2,s3,s4) vec = vec4 (s1 vec) (s2 vec) (s3 vec) (s4 vec)

-- TODO: This use of selector functions sure seems the hard way.  Instead,
-- go directly to a swizzle form.  Try redoing the type classes VecX etc
-- to yield an index number.  Use defaults (X->0, Y->1, etc), and never
-- override them.  Use subclassing: HasW < HasZ < HasY < HasX.


-- Dissection into coordinates
coords2 :: Vec2E -> (FloatE,FloatE)
coords3 :: Vec3E -> (FloatE,FloatE,FloatE)
coords4 :: Vec4E -> (FloatE,FloatE,FloatE,FloatE)

coords2 vec = (vecX vec, vecY vec)
coords3 vec = (vecX vec, vecY vec, vecZ vec)
coords4 vec = (vecX vec, vecY vec, vecZ vec, vecW vec)

class Coords vec vals | vec -> vals, vals -> vec where
  coords     :: vec -> vals
  fromCoords :: vals -> vec

instance Coords Vec2E (FloatE,FloatE) where
  coords           = coords2
  fromCoords (x,y) = vec2 x y

instance Coords Vec3E (FloatE,FloatE,FloatE) where
  coords             = coords3
  fromCoords (x,y,z) = vec3 x y z

instance Coords Vec4E (FloatE,FloatE,FloatE,FloatE) where
  coords               = coords4
  fromCoords (x,y,z,w) = vec4 x y z w


instance Cross (FloatE,FloatE) (FloatE,FloatE) where
  cross (x,y) = (y,-x)

instance Cross Vec2E Vec2E where
  cross v = fromCoords (cross (coords v))

instance Cross (Vec3E,Vec3E) Vec3E where
  cross (v,v') = fromCoords (cross (coords v,coords v'))




-- We can give simple names for tuples.  In that case the name refers to a
-- vector and we yield a tuple of extractions, just as if from the coords
-- function.

instance (AsVector v, Coords (Exp v) vals) => AsVar String vals where
  asVar name _ = ([(name, expType vec)], coords vec) where vec = var name

-- Dually, we can give a tuple of names for a single vector.  In that
-- case, generate a vector formation.

instance (AsVar ns vals, Coords vec vals) => AsVar ns vec where
  asVar ns = onSnd fromCoords . asVar ns . coords


---- Overloadings

-- Now we define a bunch of methods.  The naming policy is to use the
-- same name as the unlifted function if overloaded and the types permit.
-- Sadly, type restrictions prevent us from defining some of the methods
-- on behaviors.  In such cases, add a suffix of "*" or "E" if the name is
-- a syntactic "operator" rather than an identifier.

noOverloadE   :: String -> String -> a
noOverloadEId :: String -> a
noOverloadEOp :: String -> a

noOverloadE suffix name =
 error ("No overloading of \"" ++ name ++ "\" for expressions.  Use \"" ++
        name ++ suffix ++"\" instead.")

noOverloadEId = noOverloadE "E"
noOverloadEOp = noOverloadE "*"

instance Eq (Exp a)  where
  -- Questionable.  Used here so that AA.optLPart works.  Note that this
  -- version of equality is conservative.  It may yield False when
  -- expressions are semantically equivalent.
  -- TODO: revisit.
  E e1 == E e2 = e1 SD.==% e2

--   (==) = noOverloadEOp "=="
--   (/=) = noOverloadEOp "/="

instance (AsVector a, Ord a) => Ord (Exp a)  where
  (<)  = noOverloadEOp "<"
  (<=) = noOverloadEOp "<="
  (>)  = noOverloadEOp ">"
  (>=) = noOverloadEOp ">="
  min  = typ2 min
  max  = typ2 max

class BoolTup bool

instance BoolTup Bool
instance BoolTup (Bool,Bool)
instance BoolTup (Bool,Bool,Bool)
instance BoolTup (Bool,Bool,Bool,Bool)

-- Boolean type corresponding to a given type.
class BoolTup bool => BoolType bool a | a -> bool

-- instance AsFloat a => BoolType Bool a

-- The line above is flagged inconsistent, although no tuples are declared
-- as AsFloat types.  Instead, list the AsFloat types here (redundantly).

-- If you add a BoolType type, make the corresponding addition for AsFloat.

instance BoolType Bool Float
instance BoolType Bool Bool

instance (AsFloat a, AsFloat b) => BoolType (Bool,Bool) (a,b)
instance (AsFloat a, AsFloat b, AsFloat c) => BoolType (Bool,Bool,Bool) (a,b,c)
instance (AsFloat a, AsFloat b, AsFloat c, AsFloat d) => BoolType (Bool,Bool,Bool,Bool) (a,b,c,d)

-- instance BoolType bool val => BoolType (Exp bool) (Exp val)

-- Generalized boolean-related types
class GEq val bool | val -> bool where
  (==*), (/=*) :: val -> val -> bool

instance (BoolType bool a, Eq a) => GEq (Exp a) (Exp bool) where
  (==*) = typ2 (SD.==*)
  (/=*) = typ2 (SD./=*)

class GOrd bool val | val -> bool where
  (<*), (<=*), (>=*), (>*) :: val -> val -> bool

instance (BoolType bool val, Ord val) => GOrd (Exp bool) (Exp val) where
  (<*)  = typ2 (SD.<*)
  (<=*) = typ2 (SD.<=*)
  (>=*) = typ2 (SD.>=*)
  (>*)  = typ2 (SD.>*)

instance GOrd bool val => GOrd (a -> bool) (a -> val) where
  (<*)  = lift2 (<*)
  (<=*) = lift2 (<=*)
  (>=*) = lift2 (>=*)
  (>*)  = lift2 (>*)

class GBoolType bool val

instance (BoolTup bool, BoolType bool val) => GBoolType (Exp bool) (Exp val)

instance GBoolType bool val => GBoolType (a -> bool) (a -> val)

class GBool bool where
  (&&*), (||*) :: bool -> bool -> bool
  notE :: bool -> bool
  -- ifE :: GBoolType bool a => bool -> a -> a -> a

instance BoolTup bool => GBool (Exp bool) where
  (&&*) = typ2 (SD.&&*)
  (||*) = typ2 (SD.||*)
  notE  = typ1 SD.notE

instance GBool bool => GBool (a -> bool) where
  (&&*) = lift2 (&&*)
  (||*) = lift2 (||*)
  notE = lift1 notE


class GBool bool => GCond bool val | val -> bool where
  -- Turn a boolean into 1 or 0 for True or False, respectively
  bsign :: bool -> val
  ifE   :: bool -> val -> val -> val

instance (GBool (Exp bool), BoolType bool val) => GCond (Exp bool) (Exp val) where
  bsign = typ1 id  -- Nothing to do, really.
  ifE   = typ3 SD.ifE

instance GCond bool val => GCond (a -> bool) (a -> val) where
  bsign = lift1 bsign
  ifE   = lift3 ifE


-- ifE :: BoolType bool a => Exp bool -> Exp a -> Exp a -> Exp a
-- ifE = typ3 SD.ifE

instance (AsVector a, Num a) => Num (Exp a) where
  (+)         = typ2 (+)
  (*)         = typ2 (*)
  negate      = typ1 negate
  abs         = typ1 abs
  fromInteger = constE . fromInteger

undef name = error $ name ++ ": no def for Exp"

-- Multiply-add operation.  Not really necessary.  Introduce the operator
-- in CNF.
-- mad :: Num a => a -> a -> a -> a
-- mad a b c = a*b+c

-- Some functions are not supported in simd style, so provide only scalar
-- version here.  We could tuple them up into separate scalar operations
-- here, but that would be deceptive, performance-wise.  On the other
-- hand, future chips may do these operations in simd style.

instance Fractional (Exp Float) where
  recip        = typ1 recip
  fromRational = constE . fromRational

instance (AsVector a, Fractional a) => Fractional (Exp a) where
  recip        = scalarOnly "recip"
  fromRational = constE . fromRational

instance Floating (Exp Float) where
  pi    = constE pi
  sin   = sinN . wrap (-pi,pi)
  cos   = cosN . wrap (-pi,pi)
  sqrt  = typ1 sqrt
  log   = typ1 log
  exp   = typ1 exp

-- These ones don't wrap their arguments.  Use them to improve performance
-- when you know your argument is small enough.  If Approx.hs uses
-- sinApproxDouble and cosApproxDouble, the argument may be somewhat outside of
-- [-pi,pi], but not all the way to [-2pi,2pi].  A cool thing is that the
-- angle wrapping does not complicate the derivatives, thanks to
-- simplification.
sinN, cosN :: FloatE -> FloatE
sinN = typ1 sin
cosN = typ1 cos

-- Approximations are now introduced after differentiation and before CSE.
-- Should we wrap?  The programmer often knows it to be unnecessary, which
-- argues for requiring explicit wrapping.  Fortunately wrapping does not
-- contaminate derivatives, since simplification removes the effects, as
-- one would hope.  Solution: sin & cos do the wrapping, 

sinApprox4 d = sinCoeffs4 <.> oddPowers4  d
cosApprox4 d = cosCoeffs4 <.> evenPowers4 d

cosCoeffs4 = constE (rfac 0,-rfac 2,rfac 4,-rfac 6)
sinCoeffs4 = constE (rfac 1,-rfac 3,rfac 5,-rfac 7)

evenPowers4 d = vec4 (d^0) (d^2) (d^4) (d^6)
oddPowers4  d = d *^ evenPowers4 d
               -- vec4 (d^1) (d^3) (d^5) (d^7)

-- oddPowers4 is defined in terms of evenPowers4, because sin & cos are
-- often used together with the same argument, so we want sharing.  For
-- instance, sin r + cos r takes 10 instructions with this version of
-- oddPowers4, and 14 instructions with the other version.  sin r by itself
-- takes eight instructions.

-- The results of the sin&cos approximations near pi & -pi are different
-- enough and wrong enough to look bad in these models.  So ...

-- Further reduce the range of sin & cos's arguments by using double angle
-- formulae:
-- 
--    sin 2x = 2 * sin x * cos x
--    cos 2x = (cos x)^2 - (sin x)^2
-- 
-- and hence, substituting w/2 for x:
-- 
-- sin w = 2 * sin (w/2) * cos (w/2)
-- cos w = (cos (w/2))^2 - (sin (w/2))^2

sinApprox = sinApprox4
cosApprox = cosApprox4

sinApproxDouble w = 2 * sinApprox (w/2) * cosApprox (w/2)
cosApproxDouble w = (cosApprox (w/2))^2 - (sinApprox (w/2))^2


-- Cheaper approximations.  May be useful sometimes.

sinApprox3 d = sinCoeffs3 <.> oddPowers3  d
cosApprox3 d = cosCoeffs3 <.> evenPowers3 d

cosCoeffs3 = constE (rfac 0,-rfac 2,rfac 4)
sinCoeffs3 = constE (rfac 1,-rfac 3,rfac 5)

evenPowers3 d = vec3 (d^0) (d^2) (d^3)
oddPowers3  d = d *^ evenPowers3 d


-- sinApproxDouble = sinApprox
-- cosApproxDouble = cosApprox

rfac :: Int -> Float
rfac = recip . fromIntegral . fac
 where
   fac n = product [1 .. n]


instance (AsVector a, Floating a, Fractional (Exp a)) => Floating (Exp a) where
  pi   = constE pi
  sin  = scalarOnly "sin"
  cos  = scalarOnly "cos"
  sqrt = scalarOnly "sqrt"
  log  = scalarOnly "log"
  exp  = scalarOnly "exp"

-- TODO: Figure out why Fractional (Exp a) couldn't be deduced.

scalarOnly name exp =
  error $ name ++ ": scalar (FloatE) only for for Exp.  "
          ++ "Applied to " ++ show exp


-- Expp is an strange instruction.  As a side-effect, it places the
-- fractional part of its argument in Y.  It also expects its argument in
-- W, so we'll swizzle it from X to W.  No, that will happen implicitly
-- when the register is used as a source with only one explicit component.
-- expp :: FloatE -> Vec4E
-- expp = typ1 SD.expp

frac :: FloatE -> FloatE
frac = typ1 SD.frac

-- Now done in Approx.hs

-- expp (E dexp) = E (D.Apply Expp [D.Apply (Swizzle [0,0,0,0]) [dexp]])

-- Extract fractional component (frac x = x - floor x).  Could use the
-- "frc" macro-instruction, but it's rather awkward, since it computes two
-- fracs at once.  It also takes three instructions, while vecY.expp uses
-- one (assuming the vecY swizzle gets absorbed).
-- fracApprox :: FloatE -> FloatE
-- fracApprox = vecY . typ1 D.expp

-- Wrap to an interval
wrap :: (FloatE, FloatE) -> FloatE -> FloatE
wrap (lo,hi) x = frac ((x-lo)/(hi-lo)) * (hi-lo) + lo


---- Colors

type ColorE  = Vec4E                      -- RGBA
type ColourE = ColorE

-- TODO: Determine whether these colors are pre-multiplied alpha

rgba = vec4
rgb r g b = rgba r g b 1

gray,grey :: FloatE -> ColorE
gray x = rgb x x x
grey   = gray

black, white :: ColorE
black = gray 0
white = gray 1

-- Some common colors:
red, orange, yellow, green, blue :: ColorE
red       = rgb 1 0 0
orange    = rgb 1 0.65 0
yellow    = rgb 1 1 0
green     = rgb 0 1 0
blue      = rgb 0 0 1


---- Transforms/matrices

infixl 7 &* -- transposed matrix * vector

type Mat4x4E = (Vec4E,Vec4E,Vec4E,Vec4E) -- transposed

type Mat3x4E = (Vec3E,Vec3E,Vec3E,Vec3E) -- transposed

class Transform mat vec vec'
     | mat vec -> vec' {-, mat vec' -> vec, vec vec' -> mat -} where
  (&*) :: mat -> vec -> vec'

instance Transform Mat4x4E Vec4E Vec4E where
  (m1,m2,m3,m4) &* vec = vec4 (dot m1) (dot m2) (dot m3) (dot m4)
    where dot mi = vec <.> mi

instance Transform Mat3x4E Vec3E Vec4E where
  (m1,m2,m3,m4) &* vec = vec4 (dot m1) (dot m2) (dot m3) (dot m4)
    where dot mi = vec <.> mi

-- Tack on w==1 if the source vector is too short.
instance Transform Mat4x4E Vec3E Vec4E where
  mat &* vec = mat &* (vec4 x y z 1 :: Vec4E) where (x,y,z) = coords vec

-- TODO: The choice of adding a one or a zero depends on whether we mean
-- a point or a vector.

instance AsVar String Mat4x4E where
  asVar name mat = ([(name, Matrix 4 4)], (r1,r2,r3,r4))
   where
     [r1,r2,r3,r4] = map (var . matVarName name) [1 .. 4]

instance AsVar String Mat3x4E where
  asVar name mat = ([(name, Matrix 3 4)], (r1,r2,r3,r4))
   where
     [r1,r2,r3,r4] = map (var . matVarName name) [1 .. 4]

matVarName :: String -> Int -> String
matVarName name i = name ++ "_" ++ show i

-- Hmm... still not clear what to do with matrices.


---- Outputs

-- Represent Output as a list of pairs of register number and DExp.

type OutVert = (Int, D.Exp)

outVert :: Int -> Exp a -> OutVert
outVert regNum (E dexp) = (regNum, dexp)

-- Friendly interface
oDiffuse ,oSpecular,oFog,oPos,oPointSize :: Exp a -> OutVert
[oDiffuse,oSpecular,oFog,oPos,oPointSize] = map outVert [0..4]

oTexCoord :: Int -> Exp a -> OutVert
oTexCoord n = outVert (n+5)

output :: [OutVert] -> OutputE
output outs = E (D.Apply (Out regs) dexps)
 where
   (regs,dexps) = unzip outs

-- TODO: respect requirement that texture registers are used in order.  At
-- least I think I read this requirement somewhere.


-- Convenience
wrapVecDiffuse :: Transform mat (Exp vec) (Exp vec')
             => (Vec3E -> a -> b -> (Exp vec, Exp col))
             -> ((mat,Vec3E,a) -> b -> OutputE)
wrapVecDiffuse f (viewProj,eye,a) b =
  output [oDiffuse col, oPos (viewProj &* vec)]
 where
   (vec,col) = f eye a b

-- Type specialization to avoid overloading ambiguities (Transform)
wrapVecDiffuse4 :: Transform Mat4x4E (Exp vec) (Exp vec')
                => (Vec3E -> a -> b -> (Exp vec, ColorE))
                -> ((Mat4x4E,Vec3E,a) -> b -> OutputE)
wrapVecDiffuse4 = wrapVecDiffuse

{- -- Not used yet

wrapVecDiffuse3 :: (a -> b -> (Vec3E, ColorE)) -> ((Mat3x4E,a) -> b -> OutputE)
wrapVecDiffuse3 = wrapVecDiffuse

-}

---- Animation
type TimeE  = FloatE
type Anim a = TimeE -> a

---- CSE

class CSE a where cse :: [Id] -> a -> a

instance CSE (D.Exp) where cse = CSE.cse

instance CSE (Exp a) where
  cse avoid = typ1 (cse avoid)

instance (CSE a, CSE b) => CSE (a,b) where
  cse vs (a,b) = (cse vs a, cse vs b)

instance (CSE a, CSE b, CSE c) => CSE (a,b,c) where
  cse vs (a,b,c) = (cse vs a, cse vs b, cse vs c)

instance (CSE a, CSE b, CSE c, CSE d) => CSE (a,b,c,d) where
  cse vs (a,b,c,d) = (cse vs a, cse vs b, cse vs c, cse vs d)

instance (CSE a) => CSE [a] where
  cse vs xs = map (cse vs) xs

---- Substitution

instance SD.Substable (Exp a) where
  subst env = typ1 (SD.subst env)

---- Complexity counting

countOps pred = NumOps.countOps pred . unE
analyzeOps    = NumOps.analyzeOps    . unE

{-
---- Factoring (temp for testing)

tryFactor        = typ1 SD.tryFactor
tryFactors (E e) = map E (SD.tryFactors e)
-}
