-- Basic expression types used for programmable graphics, plus a bit of
-- simple misc.  Note: Requires overlapping instances.

module Types where

import Maybe  (fromMaybe)
import List   (sortBy,groupBy)
import IOExts (trace)

data Type = Vec Int  -- n-dimensional vector of floats
          | Matrix Int Int -- # cols, # rows
          | Output
       -- | Bool | Int
  deriving Eq

-- The Matrix type is eliminated early on, so doesn't appear to rewriting
-- and code generation.

-- Could derive Show, but this version is a little prettier
instance Show Type where
  show (Vec 1) = "Float"
  show (Vec n) = "Vec" ++ show n
  show Output  = "Output"


scalar = Vec 1

size2type :: Int -> Type
size2type = Vec

type2size :: Type -> Int
type2size (Vec n) = n

type Id = String  -- Identifier

type Vec1 = Float
type Vec2 = (Float,Float)
type Vec3 = (Float,Float,Float)
type Vec4 = (Float,Float,Float,Float)


-- In our untyped representation, we'll represent n-tuples as lists of
-- floats.

class AsFloat a where
  toFloat   :: a -> Float
  fromFloat :: Float -> a

instance AsFloat Float where
  toFloat   = id
  fromFloat = id

instance AsFloat Bool where
  toFloat False = 0.0
  toFloat True  = 1.0
  fromFloat 0.0 = False
  fromFloat 1.0 = True
  -- Watch for failure here with approximated values

-- If you add an AsFloat type, make the corresponding addition for BoolType.

type Vector = [Float]

class AsVector a where
  toVector   :: a -> Vector
  fromVector :: Vector -> Maybe a

instance AsVector Vector where
  toVector   = id
  fromVector = Just

-- TODO: shouldn't fromFloat also return a Maybe?  For instance, some
-- floats do not represent booleans, and some do not represent integers.

-- The next two instances overlap, but it works out okay.
instance AsFloat a => AsVector a where
  toVector a = [toFloat a]
  fromVector [x] = Just (fromFloat x)
  fromVector _   = Nothing

instance (AsFloat a, AsFloat b) => AsVector (a,b) where
  toVector ~(a,b)  = [toFloat a, toFloat b]
  fromVector [x,y] = Just (fromFloat x, fromFloat y)
  fromVector _     = Nothing

-- Or, if fromFloat yields Maybe Float,
-- 
--   fromVector [x,y] = liftM2 (,) (fromFloat x) (fromFloat y)
--   fromVector _     = Nothing


instance (AsFloat a, AsFloat b, AsFloat c) => AsVector (a,b,c) where
  toVector ~(a,b,c) = [toFloat a, toFloat b, toFloat c]
  fromVector [x,y,z] = Just (fromFloat x, fromFloat y, fromFloat z)
  fromVector _       = Nothing

instance (AsFloat a, AsFloat b, AsFloat c, AsFloat d) => AsVector (a,b,c,d) where
  toVector ~(a,b,c,d) = [toFloat a, toFloat b, toFloat c, toFloat d]
  fromVector [x,y,z,w] = Just (fromFloat x, fromFloat y, fromFloat z, fromFloat w)
  fromVector _         = Nothing

-- For example, the following yield True
-- 
-- toVector (False,True,False,True) == [0.0,1.0,0.0,1.0]
-- fromVector [0.0,1.0,0.0,1.0] == (False,True,False,True)

-- The lazy patterns in toVector above are for a trick used in the
-- definition of TExp.expTypeSize.  The fromVector lazy patterns are just
-- there.


---- Swizzling

-- A swizzle is a mapping from a^m to a^n where each component of the
-- resulting tuple is one of the components of the argument tuple.
-- A swizzle is represented as a list of n values in [0..m-1], saying for
-- each position in the result, what position in the argument does it come
-- from.

type Index   = Int
type Swizzle = [Index]

idSwizzle :: Type -> Swizzle
idSwizzle ty = [0 .. type2size ty - 1]

swizzleVals :: Swizzle -> [a] -> [a]
swizzleVals swiz vals = map (vals !!) swiz

-- To compose swizB & swizA, swizzle swizA using swizB
compSwiz :: Swizzle -> Swizzle -> Swizzle
compSwiz = swizzleVals

showIndex :: Index -> Char
showIndex 0 = 'x'
showIndex 1 = 'y'
showIndex 2 = 'z'
showIndex 3 = 'w'

-- Show a swizzling, optionally trimming suffix repetitions.  For asm code
-- we'd like to trim, but for pretty-printing expressions, we'd rather not.

showSwizzle :: Bool -> Swizzle -> String
showSwizzle _ [0,1,2,3]    = ""
showSwizzle False ixs      = '.' : map showIndex ixs
showSwizzle True (xi:rest) = 
  -- Trim constant suffix, e.g., "xyyy" -> "xy"
  '.' : xc : trim xc (map showIndex rest)
 where
   xc = showIndex xi
   -- Use either of the next two clauses.  The second one trims, which is
   -- terser but perhaps less clear.
   -- trim _ [] = []  -- redundant with next case
   trim prev cs | all (== prev) cs = ""
   trim _ (c:cs) = c : trim c cs



-- For instance (using xyzw for 0123)   xz `compSwiz` yyx == yx

-- Related to Swizle

-- A logical value is held in one or more parts of a register, as
-- specified by a layout (sequence of unique component indices).  The
-- layout needn't be initial, contiguous, or even in increasing index order.

-- Sequence of unique component indices.
type Layout = [Index]

-- Destination mask.  Really an index set, since order is unimportant.
type Mask = Layout

showMask :: Mask -> String
showMask [0,1,2,3] = ""
showMask mask = '.' : showLayout mask

showLayout :: Layout -> String
showLayout = map showIndex

---- Possible negation combined with swizzling.

data NegSwiz = NegSwiz Bool Swizzle deriving (Eq,Show)

idNegSwiz :: Type -> NegSwiz
idNegSwiz ty = NegSwiz False (idSwizzle ty)

negateNS :: NegSwiz -> NegSwiz
negateNS (NegSwiz neg swizzle) = NegSwiz (not neg) swizzle

swizzleNS :: Swizzle -> NegSwiz -> NegSwiz
swizzleNS swiz ns = NegSwiz False swiz `composeNS` ns

composeNS :: NegSwiz -> NegSwiz -> NegSwiz
NegSwiz negB swizB `composeNS` NegSwiz negA swizA =
  NegSwiz (negB /= negA) (swizB `compSwiz` swizA)



---- Environments

-- Map variable names to whatever.  Using simple association lists, as
-- finite maps seems overkill.
type Env a = [(Id, a)]

-- Bindings in each added group should be distinct.  Otherwise, earlier
-- ones win.
extendEnv :: Env a -> [(Id, a)] -> Env a
extendEnv = flip (++)

lookupEnv :: Show a => Env a -> Id -> a
lookupEnv env v = fromMaybe (error err) (lookup v env)
 where
   err = "lookupEnv failed: " ++ v ++ " in " ++ show env

mkEnv :: [(Id,a)] -> Env a
mkEnv = id

type TypeEnv = Env Type  -- [(Id,Type)]

allIds :: [Id]
allIds = tail all'
 where
   all' = "" : [ ch:name | name <- all', ch <- ['a' .. 'z'] ]



---- Misc

-- For sections
pair = (,)

-- list-currying
appL1 :: ([a] -> b) -> a                -> b
appL2 :: ([a] -> b) -> a -> a           -> b
appL3 :: ([a] -> b) -> a -> a -> a      -> b
appL4 :: ([a] -> b) -> a -> a -> a -> a -> b

appL1 f a1          = f [a1]
appL2 f a1 a2       = f [a1,a2]
appL3 f a1 a2 a3    = f [a1,a2,a3]
appL4 f a1 a2 a3 a4 = f [a1,a2,a3,a4]


-- Optional function application
ifF :: Bool -> (a -> a) -> (a -> a)
ifF False f = id
ifF True  f = f

implies :: Bool -> Bool -> Bool
a `implies` b = not a || b

-- Sort using a given key-extractor
sortF :: Ord k => (a -> k) -> [a] -> [a]
sortF key = sortBy (\ a b -> key a `compare` key b)

-- Apply a function to one second half and leave the other
onFst :: (a -> b) -> (a,c) -> (b,c)
onFst f (a,c) = (f a, c)

onSnd :: (a -> b) -> (c,a) -> (c,b)
onSnd f (c,a) = (c, f a)

swap :: (a,b) -> (b,a)
swap    (a,b)  = (b,a)

whenF :: Monad m => Bool -> a -> m a
whenF cond a = if cond then return a else fail ""

equal :: Eq a => [a] -> Bool
equal [] = True
equal (a:as) = all (== a) as

gather :: Ord b => [(b,a)] -> [(b,[a])]
gather = map (\ bas@((b,_):_) -> (b, map snd bas)) .
         groupBy (\ (b,a) (b',a') -> b==b') .
         sortF fst

withF :: (a -> k) -> [a] -> [(k,a)]
withF f = map (\ x -> (f x, x))

ptrace :: Show a => String -> a -> a
ptrace msg a = seq (length show_a) $  -- fully evaluate a
               trace ("\n" ++ msg ++ "\n" ++ show_a ++ "\n") a
 where show_a = show a

ftrace :: Show a => String -> (b -> a) -> (b -> a)
ftrace msg f b = ptrace msg (f b)

ftrace' :: (Show a, Show b) => String -> (b -> a) -> (b -> a)
ftrace' msg f b = seq (length show_a) $  -- fully evaluate a
                  seq (length show_b) $  -- fully evaluate b
                  trace ("\n" ++ msg ++ "(" ++ show b ++ ") == " ++ show_a ++ "\n") $
                  a
 where 
       a      = f b
       show_a = show a
       show_b = show b
