-- Vector, Point, and Color types.  Also vector space and numeric
-- overloadings for tuples.

module Vector where
import Function

-- scalar*vector & vector/scalar, dot product.  The "^" means "vector."
infixl 7 ^/, <.>
infixl 7 *^, ^*

-- type Point3  = (Float,Float,Float)
-- type Vector3 = (Float,Float,Float)

---- Type classes.  Vector spaces and numeric overloadings for tuples.

class Floating s => VectorOf s v | v -> s where  -- s for scalar, v for vector
  (*^)  :: s -> v -> v  -- scalar product
  (<.>) :: v -> v -> s  -- dot product

(^*),(^/) :: VectorOf s v => v -> s -> v
(^*)   = flip (*^)
v ^/ s = recip s *^ v

cross3 :: (Num s, VectorOf s (s,s,s)) => (s,s,s) -> (s,s,s) -> (s,s,s)
(a1,b1,c1) `cross3` (a2,b2,c2) = (b1*c2-b2*c1, c1*a2-c2*a1, a1*b2-a2*b1)

-- Vector magnitude and normalization.
magnitudeSq, magnitude :: VectorOf s v => v -> s
magnitudeSq v = v <.> v
magnitude = sqrt . magnitudeSq

normalize :: VectorOf s v => v -> v
normalize v = v ^/ magnitude v

-- Generalized cross prduct
class Cross vecs vec | vecs -> vec, vec -> vecs where
  cross :: vecs -> vec

instance (Num s, VectorOf s (s,s,s))
   => Cross ((s,s,s),(s,s,s)) (s,s,s) where cross = uncurry cross3

instance Cross vecs vec => Cross (a -> vecs) (a -> vec) where
  cross = lift1 cross




-- We'd like to say:
-- 
--   instance (Fractional s) => VectorOf s s where
--     (*^)  = (*)
--     (<.>) = (*)
-- 
-- But then:
-- 
--   *** This instance    : VectorOf a (b,c)
--   *** Conflicts with   : VectorOf a a
--   *** For class        : VectorOf a b
--   *** Under dependency : b -> a
-- 
-- If I remove the functional type dependency in VectorOf then dot raises
-- too much ambiguity.  This seems a good compromise.
-- I don't understand the error message.  I think the problematic example
-- is VectorOf (a,b) (a,b), where a & b are Fractional.

instance VectorOf Float  Float  where {(*^) = (*); (<.>) = (*)}
instance VectorOf Double Double where {(*^) = (*); (<.>) = (*)}

-- Tuples of vectors are vectors
instance (Num s, VectorOf s v1, VectorOf s v2) => VectorOf s (v1,v2) where
  s *^ (a,b) = (s*^a,s*^b)
  (a1,b1) <.> (a2,b2) = a1 <.> a2 + b1 <.> b2

-- A less general, but possibly (?) less ambiguous alternative.
-- instance Num s => VectorOf s (s,s) where
--   s *^ (a,b) = (s*a,s*b)
--   (a1,b1) <.> (a2,b2) = a1*a2+b1*b2

instance (Num s, VectorOf s v1, VectorOf s v2, VectorOf s v3)
    => VectorOf s (v1,v2,v3) where
  s *^ (a,b,c) = (s*^a,s*^b,s*^c)
  (a1,b1,c1) <.> (a2,b2,c2) = a1<.>a2+b1<.>b2+c1<.>c2

instance (Num s, VectorOf s v1, VectorOf s v2, VectorOf s v3, VectorOf s v4)
    => VectorOf s (v1,v2,v3,v4) where
  s *^ (a,b,c,d) = (s*^a,s*^b,s*^c,s*^d)
  (a1,b1,c1,d1) <.> (a2,b2,c2,d2) = a1<.>a2+b1<.>b2+c1<.>c2+d1<.>d2

instance VectorOf s v => VectorOf (a->s) (a->v) where
  s *^ vec = \ a -> s a *^ vec a
  f <.> g  = \ a -> f a <.> g a



instance (Num a, Num b) => Num (a,b) where
  (a,b) + (a',b') = (a+a',b+b')
  (a,b) * (a',b') = (a*a',b*b')
  negate (a,b) = (negate a, negate b)
  fromInteger n = (fromInteger n, fromInteger n)

instance (Fractional a, Fractional b) => Fractional (a,b) where
  recip (a,b) = (recip a, recip b)
  fromRational x = (fromRational x, fromRational x)

instance (Floating a, Floating b) => Floating (a,b) where
  sqrt (a,b) = (sqrt a, sqrt b)
  -- and so on for abs, pi, log, exp, cos, ...

instance (Num a, Num b, Num c) => Num (a,b,c) where
  (a,b,c) + (a',b',c') = (a+a',b+b',c+c')
  (a,b,c) * (a',b',c') = (a*a',b*b',c*c')
  negate (a,b,c) = (negate a, negate b, negate c)
  fromInteger n = (fromInteger n, fromInteger n, fromInteger n)

instance (Fractional a, Fractional b, Fractional c)
    => Fractional (a,b,c) where
  recip (a,b,c) = (recip a, recip b, recip c)
  fromRational x = (fromRational x, fromRational x, fromRational x)

instance (Floating a, Floating b, Floating c)
    => Floating (a,b,c) where
  sqrt (a,b,c) = (sqrt a, sqrt b, sqrt c)

instance (Num a, Num b, Num c, Num d) => Num (a,b,c,d) where
  (a,b,c,d) + (a',b',c',d') = (a+a',b+b',c+c',d+d')
  (a,b,c,d) * (a',b',c',d') = (a*a',b*b',c*c',d*d')
  negate (a,b,c,d) = (negate a, negate b, negate c, negate d)
  fromInteger n = (fromInteger n, fromInteger n, fromInteger n, fromInteger n)

instance (Fractional a, Fractional b, Fractional c, Fractional d)
    => Fractional (a,b,c,d) where
  recip (a,b,c,d) = (recip a, recip b, recip c, recip d)
  fromRational x = (fromRational x, fromRational x, fromRational x, fromRational x)

instance (Floating a, Floating b, Floating c, Floating d)
    => Floating (a,b,c,d) where
  sqrt (a,b,c,d) = (sqrt a, sqrt b, sqrt c, sqrt d)

